import os
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog as tkfd
import tkinter.font as font
from ast import literal_eval
from base64 import b64decode
from numpy import array, linalg
from copy import deepcopy
from math import sqrt, pi, inf, sin, cos, tan, asin, acos, atan

# Big picture: The main process is started at the very bottom of this script. It mostly instantiates a MainWindow object, which contains the main GUI window and orchestrates the further procedure. Prompted by commands from the user, the MainWindow object takes in one or more .cif files, each containing information about the positions of atoms in a crystal structure and the symmetry of that structure. For each such file, Polynator instantiates a CifFile object, which reads the file and stores the information it contains. Depending on the crystal structure and the user input, each CifFile may host any number of CoordinationEnvironment objects. Each CoordinationEnvironment contains a fixed set of atoms and is home to any number of Polyhedron objects, which are built from blueprints fitting the coordination number of the respective CoordinationEnvironment object. Each Polyhedron has a belts_real and a belts_model attribute, which store the vertices. The vertices of a given polyhedron are structured as a list of lists. The main calculations are carried out within the Polyhedron objects. The actual fitting methods have names starting with adjust_ and are specific to the type of parameter being fitted. Central processes can be found under the functions / methods MainWindow.run_full(), which wraps run_blueprint(), which wraps Polyhedron.evaluate().
# recurring concepts:
# blueprint: the tuples found in DICT_BLUEPRINTS are referred to as blueprints. Each contains all the necessary information to construct and process one specific polyhedron (and sometimes its chiral counterpart). They have three entries, which are referred to as name, dict_info and belt_dicts.
# belt: The third and last entry of each blueprint is a tuple containing one or more dictionaries. Each of these dictionaries contains the information to construct and process a polygon (which is usually, but not always, regular and two-dimensional). These polygons are referred to as belts. Three-dimensional polyhedra are formed by stacking up one or more belt polygons in parallel, whereby the vertices of each become vertices of the polyhedron. The axis perpendicular to the belts (= the stacking direction) becomes the main axis (usually a high symmetry axis).
# main_axis: This refers to the axis around which a given polyhedron is constructed by the instructions in the respective blueprint, generally an axis of symmetry. The main axis is always equal to the z axis when the model prototype is first constructed, but is soon adjusted by the Polyhedron.find_initial_main_axis() method to fit the real vertices. The model then follows suit via the Polyhedron.match_real_vecs_to_model() method.
# parameter bundle: Some polyhedra, such as the pyritohedron, require multiple parameters to be modified in a synchronized manner. this is achieved by constraining them to a single variable and defining their dependence on that variable in mathematical terms. This is referred to as a parameter bundle. It is defined in the dict_info section of a given blueprint.

DICT_BLUEPRINTS = {
    2: [(">equidistant_linear_coordination", {"tags": ("#linear", "#rigid", "#molecule", "#essential", "#free", "#equidistant"), "point_group": "[inf]/mmm", "axis": "2/m", "parameter_count": 1}, ({"v": 2, "z_factor": 0.0, "len": "len1"},)), (">psi-1_planar_triangle", {"tags": ("#kinked", "#planar", "#rigid", "#molecule", "#constrained", "#equidistant"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.8660254037844386, "z_factor": 0.5, "len": "len1"},)), (">psi-2_tetrahedron", {"tags": ("#kinked", "#planar", "#rigid", "#molecule", "#constrained", "#equidistant"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.8164965809277261, "z_factor": 0.5773502691896258, "len": "len1"},)), (">generic_linear_coordination", {"tags": ("#linear", "#molecule", "#free"), "aristohedra": {">equidistant_linear_coordination": ("same",)}, "synonyms": (">linear_coordination",), "point_group": "[inf]mm", "axis": "m", "parameter_count": 2}, ({"v": 2, "z_factor": 0.0, "xy": "xy1", "xy_mod": "xy_mod1", "xy_mod_frq": 1},)), (">equidistant_kinked_coordination", {"tags": ("#kinked", "#molecule", "#free", "#equidistant"), "aristohedra": {">equidistant_linear_coordination": ("same",), ">psi-1_planar_triangle": (), ">psi-2_tetrahedron": ()}, "point_group": "mm2", "axis": "2", "parameter_count": 2}, ({"v": 2, "z": "z1", "xy": "xy1"},))], 
    3: [(">regular_triangle", {"tags": ("#regular", "#planar", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "synonyms": (">equilateral_triangle",), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "z_factor": 0.0, "len": "len1"},)), (">isosceles_triangle", {"tags": ("#planar", "#free"), "synonyms": (">digonal_pyramid",), "aristohedra": {">regular_triangle": ((0,), (1,), (2,))}, "point_group": "mm2", "axis": "2", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 2, "xy": "xy1", "z": "z*"})), (">psi-1_tetrahedron", {"tags": ("#rigid", "#molecule", "#constrained", "#equidistant"), "point_group": "3m", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.9428090415820634, "z_factor": 0.3333333333333333, "len": "len1"},)), (">psi-1_square", {"tags": ("#rigid", "#molecule", "#constrained", "#equidistant"), "synonyms": (">T_shape",), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "len": "len1"}, {"v": 2, "z_factor": 0.0, "len": "len1"})), (">psi-1_trigonal_pyramid", {"tags": ("#molecule", "#free", "#equidistant"), "aristohedra": {">psi-1_tetrahedron": ("same",)}, "point_group": "3m", "axis": "3", "parameter_count": 2}, ({"v": 3, "z": "z1", "xy": "xy1"},))], 
    4: [(">square", {"tags": ("#regular", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "point_group": "4/mmm", "axis": "4/m", "parameter_count": 1}, ({"v": 4, "z_factor": 0.0, "len": "len1"},)), (">tetrahedron", {"tags": ("#platonic", "#rigid", "#essential", "#free", "#equilateral", "#equidistant", "#deltahedron"), "synonyms": (">platonic_tetrahedron", ">equilateral_trigonal_pyramid", ">equilateral_disphenoid"), "point_group": "-43m", "axis": "3", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 3, "xy_factor": 0.9428090415820634, "z_factor": -0.3333333333333333, "len": "len1"})), (">rectangle", {"tags": ("#planar", "#free", "#equidistant"), "aristohedra": {">square": ("same",)}, "point_group": "mmm", "axis": "2/m", "parameter_count": 2}, ({"v": 4, "z_factor": 0.0, "xy": "xy1", "phi_mod": "phi_mod1"},)), (">rhombus", {"tags": ("#planar", "#free", "#equilateral"), "aristohedra": {">square": ("same",)}, "point_group": "mmm", "axis": "2/m", "parameter_count": 2}, ({"v": 4, "z_factor": 0.0, "xy": "xy1", "xy_mod": "xy_mod1"},)), (">disphenoid", {"tags": ("#antiprism", "#essential", "#free", "#equidistant"), "synonyms": (">bisphenoid", ">digonal_antiprism"), "aristohedra": {">tetrahedron": ((0, 1), (0, 2), (0, 3)), ">square": ()}, "point_group": "-42m", "axis": "-4", "parameter_count": 2}, ({"v": 2, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 2, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">trigonal_pyramid", {"tags": ("#pyramid", "#essential", "#free"), "aristohedra": {">tetrahedron": ((0,), (1,), (2,), (3,))}, "point_group": "3m", "axis": "3", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 3, "xy": "xy1", "z": "z*"})), (">heterodisphenoid", {"tags": ("#antifrustum", "#free"), "synonyms": (">heterobisphenoid", ">digonal_antifrustum"), "aristohedra": {">disphenoid": ((0, 1),), ">rhombus": ([[0, 2], [1, 3]],)}, "point_group": "mm2", "axis": "2", "parameter_count": 3}, ({"v": 2, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 2, "z_factor": -1.0, "xy": "xy2", "z": "z1", "phi_init": "s"})), (">twisted_disphenoid", {"tags": ("#twisted_prism", "#free", "#equidistant"), "synonyms": (">twisted_bisphenoid",), "aristohedra": {">disphenoid": ((0, 1), (0, 2), (0, 3)), ">rectangle": ()}, "point_group": "222", "axis": "2", "parameter_count": 3}, ({"v": 2, "z_factor": 1.0, "phi": "phi1", "xy": "xy1", "z": "z1"}, {"v": 2, "z_factor": -1.0, "phi": "phi*", "xy": "xy1", "z": "z1"})), (">parallelogram", {"tags": ("#planar", "#free"), "aristohedra": {">rhombus": ("same",), ">rectangle": ()}, "point_group": "2/m", "axis": "2/m", "parameter_count": 3}, ({"v": 4, "z_factor": 0.0, "xy": "xy1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1"},)), (">trapezoid", {"tags": ("#planar", "#free"), "synonyms": (">trapezium",), "aristohedra": {">rectangle": ((0, 1), (1, 2))}, "point_group": "mm2", "axis": "2", "parameter_count": 3}, ({"v": 2, "z_factor": 1, "xy": "xy1", "z": "z1"}, {"v": 2, "z_factor": -1, "xy": "xy2", "z": "z1"})), (">seesaw", {"tags": ("#molecule", "#constrained"), "synonyms": (">psi-1_trigonal_bipyramid",), "point_group": "mm2", "axis": "m", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 3, "erase": (2,), "z_factor": 0.0, "xy": "xy1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"}))], 
    5: [(">regular_pentagon", {"tags": ("#regular", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 1}, ({"v": 5, "z_factor": 0.0, "len": "len1"},)), (">equilateral_trigonal_bipyramid", {"tags": ("#johnson", "#rigid", "#bipyramid", "#equilateral", "#constrained", "#deltahedron"), "synonyms": (">face-sharing_tetrahedron_doublet",), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.2132034355964256, "len": "len1"}, {"v": 3, "xy_factor": 0.8578643762690494, "z_factor": 0.0, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.2132034355964256, "len": "len1"})), (">equilateral_tetragonal_pyramid", {"tags": ("#johnson", "#rigid", "#pyramid", "#equilateral", "#constrained"), "synonyms": (">equilateral_square_pyramid",), "point_group": "4mm", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.819803902718557, "len": "len1"}, {"v": 4, "xy_factor": 1.0247548783981961, "z_factor": -0.20495097567963924, "len": "len1"})), (">trigonal_bipyramid", {"tags": ("#bipyramid", "#free", "#essential"), "aristohedra": {">equilateral_trigonal_bipyramid": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 3, "z_factor": 0.0, "xy": "xy1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">psi-1_octahedron", {"tags": ("#pyramid", "#rigid", "#molecule", "#constrained", "#equidistant"), "point_group": "4mm", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 4, "z_factor": 0.0, "len": "len1"})), (">tetragonal_pyramid", {"tags": ("#pyramid", "#free", "#essential"), "synonyms": (">square_pyramid",), "aristohedra": {">equilateral_tetragonal_pyramid": ("same",)}, "point_group": "4mm", "axis": "4", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z*"})), (">rhombic_pyramid", {"tags": ("#pyramid", "#constrained"), "aristohedra": {">tetragonal_pyramid": ("same",)}, "point_group": "mm2", "axis": "2", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z*", "xy_mod": "xy_mod1"})), (">rectangular_pyramid", {"tags": ("#pyramid", "#free"), "aristohedra": {">tetragonal_pyramid": ("same",)}, "point_group": "mm2", "axis": "2", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z*", "phi_mod": "phi_mod1"})), (">trigonal_heterobipyramid", {"tags": ("#heterobipyramid", "#free"), "aristohedra": {">trigonal_bipyramid": ("same",)}, "point_group": "3m", "axis": "3", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 3, "xy": "xy1", "z": "z2"}, {"v": 1, "xy_factor": 0.0, "z": "z*"}))], 
    6: [(">regular_hexagon", {"tags": ("#regular", "#planar", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 1}, ({"v": 6, "z_factor": 0.0, "len": "len1"},)), ( ">octahedron", {"tags": ("#platonic", "#rigid", "#essential", "#free", "#equilateral", "#equidistant", "#deltahedron"), "synonyms": (">platonic_octahedron",), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 4, "z_factor": 0.0, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len1"})), (">equilateral_trigonal_prism", {"tags": ("#rigid", "#prism", "#constrained", "#equilateral", "#equidistant"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.7559289460184545, "z_factor": 0.6546536707079771, "len": "len1"}, {"v": 3, "xy_factor": 0.7559289460184545, "z_factor": -0.6546536707079771, "len": "len1"})), (">trigonal_antiprism", {"tags": ("#antiprism", "#essential", "#free", "#equidistant"), "aristohedra": {">octahedron": ((0, 1, 2), (0, 2, 3), (0, 3, 4), (0, 4, 1)), ">regular_hexagon": ([[0, 2, 4], [1, 3, 5]],)}, "point_group": "-3m", "axis": "-3", "parameter_count": 2}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">trigonal_prism", {"tags": ("#prism", "#free", "#essential", "#equidistant"), "aristohedra": {">equilateral_trigonal_prism": ("same",), ">trigonal_antiprism*": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 2}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">trigonal_frustum", {"tags": ("#frustum", "#free"), "aristohedra": {">trigonal_prism": ("same", (0, 1, 3), (0, 2, 3), (0, 3, 4), (0, 3, 5))}, "point_group": "3m", "axis": "3", "parameter_count": 3}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy2", "z": "z1"})), (">trigonal_antifrustum", {"tags": ("#antifrustum", "#free"), "aristohedra": {">trigonal_antiprism": ("same", (0, 1, 3), (0, 2, 5), (0, 3, 5))}, "point_group": "3m", "axis": "3", "parameter_count": 3}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy2", "z": "z1", "phi_init": "s"})), (">twisted_trigonal_prism", {"tags": ("#twisted_prism", "#free", "#equidistant"), "aristohedra": {">trigonal_antiprism": ("same",), ">trigonal_prism": ("same",)}, "point_group": "32", "axis": "3", "parameter_count": 3}, ({"v": 3, "z_factor": 1.0, "phi": "phi*", "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "phi": "phi1", "xy": "xy1", "z": "z1"})), (">tetragonal_bipyramid", {"tags": ("#bipyramid", "#essential", "#free"), "synonyms": (">square_bipyramid",), "aristohedra": {">octahedron": ((0,), (1,), (2,))}, "point_group": "4/mmm", "axis": "4/m", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">pentagonal_pyramid", {"tags": ("#pyramid", "#free", "#essential"), "point_group": "5m", "axis": "5", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 5, "xy": "xy1", "z": "z*"})), (">psi_1-pentagonal_bipyramid", {"tags": ("#pyramid", "#constrained", "#molecule"), "aristohedra": {">pentagonal_pyramid*": ("same",)}, "point_group": "5m", "axis": "5", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 5, "z_factor": 0.0, "xy": "xy1"})), (">didigonal_scalenohedron", {"tags": ("#scalenohedron", "#free"), "aristohedra": {">tetragonal_bipyramid": ("same", (1,), (2,))}, "point_group": "-42m", "axis": "-4", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy1", "z_mod": "z_mod1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">rectangular_bipyramid", {"tags": ("#bipyramid", "#free"), "aristohedra": {">tetragonal_bipyramid": ("same", (1,), (2,))}, "point_group": "mmm", "axis": "2/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy1", "phi_mod": "phi_mod1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">rhombic_bipyramid", {"tags": ("#bipyramid", "#free"), "aristohedra": {">tetragonal_bipyramid": ("same",)}, "point_group": "mmm", "axis": "2/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy1", "xy_mod": "xy_mod1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">tetragonal_heterobipyramid", {"tags": ("#heterobipyramid", "#free"), "synonyms": (">square_heterobipyramid",), "aristohedra": {">tetragonal_bipyramid": ("same", (1,), (2,))}, "point_group": "4mm", "axis": "4", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z2"}, {"v": 1, "xy_factor": 0.0, "z": "z*"})), (">parallelogrammic_bipyramid", {"tags": ("#bipyramid", "#free"), "aristohedra": {">rhombic_bipyramid": ("same",), ">rectangular_bipyramid": ("same",)}, "point_group": "2/m", "axis": "2/m", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">isosceles_wedge", {"tags": ("#cupola", "#free"), "synonyms": (">digonal_cupola",), "aristohedra": {">trigonal_prism": ((0, 3), (1, 4), (2, 5)), ">regular_hexagon": ()}, "point_group": "mm2", "axis": "2", "parameter_count": 4}, ({"v": 2, "xy": "xy1", "z": "z1"}, {"v": 4, "xy": "xy2", "z": "z*", "phi_mod": "phi_mod1", "phi_init": "s"}))], 
    7: [(">regular_heptagon", {"tags": ("#regular", "#planar", "#rigid", "#free", "#equilateral", "#equidistant"), "point_group": "-[14]m2", "axis": "-14", "parameter_count": 1}, ({"v": 7, "z_factor": 0.0, "len": "len1"},)), (">equilateral_eq-monocapped_trigonal_prism", {"tags": ("#johnson", "#rigid", "#equator-capped_prism", "#constrained", "#equilateral", "#essential"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0873996694922003, "len": "len1"}, {"v": 4, "xy_factor": 0.9008588587194559, "z_factor": 0.18654081077274445, "len": "len1"}, {"v": 2, "xy_factor": 0.6370034078925012, "z_factor": -0.9167814562915891, "len": "len1", "phi_init": "s"})), (">hexagonal_pyramid", {"tags": ("#pyramid", "#free"), "point_group": "6mm", "axis": "6", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 6, "xy": "xy1", "z": "z*"})), (">pentagonal_bipyramid", {"tags": ("#bipyramid", "#free", "#essential"), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 5, "z_factor": 0.0, "xy": "xy1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">monocapped_isosceles_wedge", {"tags": ("#capped_cupola", "#free"), "synonyms": (">monocapped_digonal_cupola", ">generic_monocapped_trigonal_prism"), "aristohedra": {">hexagonal_pyramid": (), ">equilateral_eq-monocapped_trigonal_prism": ()}, "point_group": "mm2", "axis": "2", "parameter_count": 5}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z2", "phi_mod": "phi_mod1"}, {"v": 2, "z_factor": 1.0, "xy": "xy2", "z": "z*", "phi_init": "s"})), (">monocapped_trigonal_frustum", {"tags": ("#axis-capped_frustum", "#free"), "point_group": "3m", "axis": "3", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 3, "xy": "xy1", "z": "z2"}, {"v": 3, "xy": "xy2", "z": "z*"})), (">monocapped_trigonal_antifrustum", {"tags": ("#axis-capped_antifrustum", "#free"), "aristohedra": {">hexagonal_pyramid": ()}, "point_group": "3m", "axis": "3", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 3, "xy": "xy1", "z": "z2"}, {"v": 3, "xy": "xy2", "z": "z*", "phi_init": "s"})), (">pentagonal_heterobipyramid", {"tags": ("#heterobipyramid", "#free"), "aristohedra": {">pentagonal_bipyramid": ("same",)}, "point_group": "5m", "axis": "5", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 5, "xy": "xy1", "z": "z2"}, {"v": 1, "xy_factor": 0.0, "z": "z*"}))], 
    8: [(">regular_octagon", {"tags": ("#regular", "#planar", "#rigid", "#free", "#equilateral", "#equidistant"), "point_group": "8/mmm", "axis": "8/m", "parameter_count": 1}, ({"v": 8, "z_factor": 0.0, "len": "len1"},)), (">cube", {"tags": ("#platonic", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "synonyms": (">platonic_hexahedron", ">equilateral_tetragonal_prism"), "point_group": "m-3m", "axis": "-3", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 3, "xy_factor": 0.9428090415820634, "z_factor": 0.3333333333333333, "len": "len1"}, {"v": 3, "xy_factor": 0.9428090415820634, "z_factor": -0.3333333333333333, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len1"})), (">trigonal_dodecahedron", {"tags": ("#johnson", "#rigid", "#constrained", "#equilateral", "#deltahedron"), "synonyms": (">snub_disphenoid", ">equilateral_bidisphenoid", ">equilateral_snub_disphenoid", ">dodecadeltahedron"), "point_group": "-42m", "axis": "-4", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.6225182127, "z_factor": 0.976022555377479, "len": "len1"}, {"v": 4, "xy_factor": 0.8025308991249777, "z_factor": 0.0, "z_mod_init": -0.25593163733226615, "len": "len1"}, {"v": 2, "xy_factor": 0.6225182127, "z_factor": -0.976022555377479, "len": "len1", "phi_init": "s"})), (">gyrobifastigium", {"tags": ("#johnson", "#rigid", "#gyrobicupola", "#constrained", "#equilateral"), "synonyms": (">equilateral_digonal_gyrobicupola",), "point_group": "-42m", "axis": "-4", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.585786437626905, "z_factor": 1.0146118723545765, "len": "len1"}, {"v": 4, "xy_factor": 0.8284271247461901, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 2, "xy_factor": 0.585786437626905, "z_factor": -1.0146118723545765, "len": "len1", "phi_init": "s"})), (">equilateral_eq-bicapped_trigonal_prism", {"tags": ("#johnson", "#rigid", "#equator-capped_prism", "#constrained", "#equilateral", "#essential"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.8377971300404524, "z_factor": 0.5081343311225655, "len": "len1", "xy_mod_init": 0.22284257863500623, "z_mod_init": -0.048863865134828766}, {"v": 4, "xy_factor": 0.8696770668406447, "z_factor": -0.5081343311225655, "len": "len1", "phi_init": "s"})), (">equilateral_tetragonal_antiprism", {"tags": ("#rigid", "#antiprism", "#constrained", "#equilateral", "#equidistant"), "synonyms": (">equilateral_square_antiprism",), "point_group": "-82m", "axis": "-8", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.859532503769496, "z_factor": 0.5110810845293938, "len": "len1"}, {"v": 4, "xy_factor": 0.859532503769496, "z_factor": -0.5110810845293938, "len": "len1", "phi_init": "s"})), (">rhombohedron", {"tags": ("#scalenohedron", "#constrained", "#essential", "#equilateral"), "synonyms": (">trigonal_trapezohedron",), "aristohedra": {">cube": ((0,), (1,), (2,), (3,))}, "point_group": "-3m", "axis": "-3", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 3, "z_factor": 0.3333333333333333, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -0.3333333333333333, "xy": "xy1", "z": "z1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">triakis_tetrahedron", {"tags": ("#capped_tetrahedron", "#free", "#essential"), "synonyms": (">tetracapped_tetrahedron",), "aristohedra": {">cube": ("same",)}, "point_group": "-43m", "axis": "3", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 3, "xy_factor": 0.9428090415820634, "z_factor": 0.3333333333333333, "len": "len2"}, {"v": 3, "xy_factor": 0.9428090415820634, "z_factor": -0.3333333333333333, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len2"})), (">hexagonal_bipyramid", {"tags": ("#bipyramid", "#essential", "#free"), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">tetragonal_prism", {"tags": ("#prism",), "synonyms": (">square_prism", "#essential", "#free", "#equidistant"), "aristohedra": {">cube": ((0, 1, 2, 4), (0, 2, 3, 5), (0, 3, 1, 6))}, "point_group": "4/mmm", "axis": "4/m", "parameter_count": 2}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">tetragonal_antiprism", {"tags": ("#antiprism", "#free", "#essential", "#equidistant"), "synonyms": (">square_antiprism",), "aristohedra": {">regular_octagon": ([[0, 2, 4, 6], [1, 3, 5, 7]],), ">equilateral_tetragonal_antiprism": ("same",), ">tetragonal_prism*": ("same",)}, "point_group": "-82m", "axis": "-8", "parameter_count": 2}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">tetragonal_frustum", {"tags": ("#frustum", "#free"), "synonyms": (">square_frustum",), "aristohedra": {">tetragonal_prism": ("same", (0, 1, 4, 5), (1, 2, 5, 6))}, "point_group": "4mm", "axis": "4", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy2", "z": "z1"})), (">tetragonal_antifrustum", {"tags": ("#antifrustum", "#free"), "synonyms": (">square_antifrustum",), "aristohedra": {">tetragonal_antiprism": ("same", (0, 1, 4, 5), (1, 2, 5, 6))}, "point_group": "4mm", "axis": "4", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy2", "z": "z1", "phi_init": "s"})), (">twisted_tetragonal_prism", {"tags": ("#twisted_prism", "#free", "#equidistant"), "aristohedra": {">tetragonal_prism": ("same",), ">tetragonal_antiprism": ("same",)}, "point_group": "422", "axis": "4", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "phi": "phi*", "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "phi": "phi1", "xy": "xy1", "z": "z1"})), (">rhombic_prism", {"tags": ("#prism", "#constrained"), "aristohedra": {">tetragonal_prism": ("same", (0, 1, 4, 5), (1, 2, 5, 6))}, "point_group": "mmm", "axis": "2/m", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "xy_mod": "xy_mod1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "xy_mod": "xy_mod1"})), (">cuboid", {"tags": ("#prism", "#free", "#equidistant"), "synonyms": (">rectangular_prism",), "aristohedra": {">tetragonal_prism": ("same",)}, "point_group": "mmm", "axis": "2/m", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1"})), (">anticuboid", {"tags": ("#gyroprism", "#free", "#equidistant"), "synonyms": (">rectangular_gyroprism",), "aristohedra": {">tetragonal_prism": ("same", (0, 1, 4, 5), (1, 2, 5, 6))}, "point_group": "-42m", "axis": "-4", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "phi_mod_off": 0.5}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "phi_mod_off": 1.5})), (">ax-bicapped_trigonal_prism", {"tags": ("#axis-capped_prism", "#free"), "point_group": "-6m2", "axis": "-6", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">ax-bicapped_trigonal_antiprism", {"tags": ("#scalenohedron", "#free"), "synonyms": (">ditrigonal_scalenohedron",), "aristohedra": {">hexagonal_bipyramid": ([[0], [1, 3, 5], [2, 4, 6], [7]],), ">rhombohedron": ("same",)}, "point_group": "-3m", "axis": "-3", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">bidisphenoid", {"tags": ("#bidisphenoid", "#free"), "aristohedra": {">tetragonal_prism": ([[0, 2], [1, 3, 4, 6], [5, 7]],), ">trigonal_dodecahedron": ("same",), ">triakis_tetrahedron": ()}, "point_group": "-42m", "axis": "-4", "parameter_count": 4}, ({"v": 2, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "xy": "xy2", "z_factor": 0.0, "z_mod": "z_mod1"}, {"v": 2, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">digonal_gyrobicupola", {"tags": ("#gyrobicupola", "#free"), "synonyms": (">generic_gyrobifastigium",), "aristohedra": {">gyrobifastigium": ("same",)}, "point_group": "-42m", "axis": "-4", "parameter_count": 3}, ({"v": 2, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "xy": "xy2", "z_factor": 0.0, "phi_init": "s"}, {"v": 2, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">digonal_orthobicupola", {"tags": ("#orthobicupola", "#free"), "aristohedra": {">rhombic_prism": ((0, 4),)}, "point_group": "mmm", "axis": "2/m", "parameter_count": 4}, ({"v": 2, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "xy": "xy2", "z_factor": 0.0, "phi_mod": "phi_mod1", "phi_init": "s"}, {"v": 2, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">biaugmented_isosceles_wedge", {"tags": ("#capped_cupola", "#free"), "synonyms": (">biaugmented_digonal_cupola",), "aristohedra": {">tetragonal_antifrustum": ("same",), ">equilateral_eq-bicapped_trigonal_prism": ((2, 3, 7, 4),)}, "point_group": "mm2", "axis": "2", "parameter_count": 6}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "xy_mod": "xy_mod1", "z_mod": "z_mod1"}, {"v": 4, "z_factor": -1.0, "xy": "xy2", "z": "z1", "phi_mod": "phi_mod1", "phi_init": "s"})), (">hexagonal_heterobipyramid", {"tags": ("#heterobipyramid", "#free"), "aristohedra": {">hexagonal_bipyramid": ()}, "point_group": "6mm", "axis": "6", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 6, "xy": "xy1", "z": "z2"}, {"v": 1, "xy_factor": 0.0, "z": "z*"})), (">parallelogrammic_prism", {"tags": ("#prism", "#constrained"), "aristohedra": {">rhombic_prism": ("same",), ">cuboid": ("same",)}, "point_group": "2/m", "axis": "2/m", "parameter_count": 4}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1"})), (">parallelepiped", {"tags": ("#axis_fixed", "#constrained", "#skewed_prism"), "aristohedra": {">parallelogrammic_prism": ("same",), ">rhombohedron": ()}, "point_group": "-1", "axis": "-1", "parameter_count": 6}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1", "z_mod": "z_mod1", "z_mod_off": 0.5, "z_mod_frq": 1, "z_mod+": "z_mod2", "z_mod+_off": 1.5, "z_mod+_frq": 1}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "xy_mod": "xy_mod1", "z_mod": "z_mod1", "z_mod_off": 0.5, "z_mod_frq": 1, "z_mod+": "z_mod2", "z_mod+_off": 1.5, "z_mod+_frq": 1}))], 
    9: [(">equilateral_tricapped_trigonal_prism", {"tags": ("#johnson", "#rigid", "#equator-capped_prism", "#equilateral", "#constrained", "#deltahedron"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "z_factor": 0.5944579523447199, "xy_factor": 0.6864209176162755, "len": "len1"}, {"v": 3, "z_factor": 0.0, "xy_factor": 1.1839009572745796, "len": "len1", "phi_init": "s"}, {"v": 3, "z_factor": -0.5944579523447199, "xy_factor": 0.6864209176162755, "len": "len1"})), (">tridiminished_icosahedron", {"tags": ("#johnson", "#diminished_icosahedron", "#rigid", "#equilateral", "#constrained"), "point_group": "3m", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.6087623734295325, "z_factor": 0.8595862653305368, "len": "len1"}, {"v": 3, "xy_factor": 0.9849982112810391, "z_factor": -0.1254119459505024, "len": "len1"}, {"v": 3, "xy_factor": 0.6087623734295325, "z_factor": -0.7341743193800344, "len": "len1", "phi_init": "s"})), (">equilateral_monocapped_cube", {"tags": ("#rigid", "#capped_cube", "#johnson", "#equilateral", "#constrained"), "synonyms": (">elongated_square_pyramid", ">equilateral_monocapped_tetragonal_prism"), "point_group": "4mm", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.1985975122858756, "len": "len1"}, {"v": 4, "xy_factor": 0.7898874377292153, "z_factor": 0.4087100745566605, "len": "len1"}, {"v": 4, "xy_factor": 0.7898874377292153, "z_factor": -0.7083594526281294, "len": "len1"})), (">equilateral_monocapped_tetragonal_antiprism", {"tags": ("#rigid", "#capped_antiprism", "#johnson", "#equilateral", "#constrained"), "synonyms": (">gyroelongated_tetragonal_pyramid", ">equilateral_monocapped_tetragonal_antiprism", ">equilateral_monocapped_square_antiprism"), "point_group": "4mm", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.1806757771193017, "len": "len1"}, {"v": 4, "xy_factor": 0.8329720845101534, "z_factor": 0.3477036926091483, "len": "len1"}, {"v": 4, "xy_factor": 0.8329720845101534, "z_factor": -0.6428726368889738, "len": "len1", "phi_init": "s"})), (">equilateral_trigonal_cupola", {"tags": ("#johnson", "#rigid", "#cupola", "#equilateral", "#constrained"), "point_group": "3m", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.6042930898773952, "z_factor": 0.5697329889019707, "len": "len1"}, {"v": 6, "xy_factor": 1.0466663343304345, "z_factor": -0.28486649445098533, "len": "len1", "phi_init": "s"})), (">tricapped_trigonal_prism", {"tags": ("#equator-capped_prism", "#free", "#essential"), "aristohedra": {">equilateral_tricapped_trigonal_prism": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 3}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">trigonal_cupola", {"tags": ("#cupola", "#free"), "aristohedra": {">equilateral_trigonal_cupola": ("same",)}, "point_group": "3m", "axis": "3", "parameter_count": 4}, ({"v": 3, "xy": "xy1", "z": "z1"}, {"v": 6, "xy": "xy2", "z": "z*", "phi_mod": "phi_mod1", "phi_mod_frq": 3, "phi_init": "s"})), (">monocapped_tetragonal_frustum", {"tags": ("#axis-capped_frustum", "#essential", "#free"), "synonyms": (">monocapped_square_frustum",), "aristohedra": {">equilateral_monocapped_cube": ("same",)}, "point_group": "4mm", "axis": "4", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z2"}, {"v": 4, "xy": "xy2", "z": "z*"})), (">monocapped_tetragonal_antifrustum", {"tags": ("#axis-capped_antifrustum", "#essential", "#free"), "synonyms": (">monocapped_square_antifrustum",), "aristohedra": {">equilateral_monocapped_tetragonal_antiprism": ("same",)}, "point_group": "4mm", "axis": "4", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 4, "xy": "xy1", "z": "z2"}, {"v": 4, "xy": "xy2", "z": "z*", "phi_init": "s"}))], 
    10: [(">adamantane_cage", {"tags": ("#rigid", "#essential", "#equilateral", "#constrained"), "synonyms": (">diamond_cage", ">sphalerite_cage"), "point_group": "-43m", "axis": "3", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.9150635094610966, "len": "len1"}, {"v": 3, "xy_factor": 0.862730150341736, "z_factor": 0.6100423396407311, "len": "len1"}, {"v": 3, "xy_factor": 0.862730150341736, "z_factor": -0.30502116982036553, "len": "len1"}, {"v": 3, "xy_factor": 0.862730150341736, "z_factor": -0.6100423396407311, "len": "len1", "phi_init": "s"})), (">sphenocorona", {"tags": ("#johnson", "#rigid", "#equilateral", "#constrained"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.5567825581019253, "z_factor": 0.7383402595590762, "len": "len1"}, {"v": 4, "xy_factor": 1.1007653070464318, "z_factor": 0.15666180556959383, "len": "len1", "phi_mod_init": -0.25507278868633265, "phi_init": "s"}, {"v": 2, "xy_factor": 0.8790790668171434, "z_factor": -0.32756415852730975, "len": "len1"}, {"v": 2, "xy_factor": 0.5567825581019253, "z_factor": -0.724099712170954, "len": "len1", "phi_init": "s"})), (">equilateral_bicapped_tetragonal_antiprism", {"tags": ("#rigid", "#axis-bicapped_antiprism", "#johnson", "#equilateral", "#constrained", "#deltahedron"), "synonyms": (">gyroelongated_square_bipyramid",  ">equilateral_bicapped_square_antiprism"), "point_group": "-82m", "axis": "-8", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.2760307232725558, "len": "len1"}, {"v": 4, "xy_factor": 0.8002181590965547, "z_factor": 0.47581256417600104, "len": "len1"}, {"v": 4, "xy_factor": 0.8002181590965547, "z_factor": -0.47581256417600104, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.2760307232725558, "len": "len1"})), (">equilateral_trans-bicapped_cube", {"tags": ("#rigid", "#capped_cube", "#johnson", "#equilateral", "#constrained"), "synonyms": (">elongated_square_bipyramid", ">equilateral_ax-bicapped_tetragonal_prism", ">equilateral_ax-bicapped_square_prism"), "point_group": "4/mmm", "axis": "4/m", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.292071214523848, "len": "len1"}, {"v": 4, "xy_factor": 0.7568777939161935, "z_factor": 0.5351934206076546, "len": "len1"}, {"v": 4, "xy_factor": 0.7568777939161935, "z_factor": -0.5351934206076546, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.292071214523848, "len": "len1"})), (">equilateral_cis-bicapped_cube", {"tags": ("#rigid", "#capped_cube", "#equilateral"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.685793684938651, "z_factor": 0.6427658234355845, "xy_mod_init": -0.11766359435582979, "len": "len1"}, {"v": 4, "xy_factor": 0.9840301821981547, "z_factor": -0.16069145585889613, "len": "len1", "phi_mod_init": -0.16991845472706102, "phi_init": "s"}, {"v": 2, "xy_factor": 0.5681300905828212, "z_factor": -0.964148735153377, "len": "len1"})), (">equilateral_pentagonal_prism", {"tags": ("#rigid", "#equilateral", "#prism", "#constrained", "#equidistant"), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.8621037223969755, "z_factor": 0.5067318539713864, "len": "len1"}, {"v": 5, "xy_factor": 0.8621037223969755, "z_factor": -0.5067318539713864, "len": "len1"})), (">equilateral_pentagonal_antiprism", {"tags": ("#rigid", "#equilateral", "#antiprism", "#constrained", "#equidistant"), "synonyms": (">para-bidiminished_icosahedron",), "point_group": "-5m", "axis": "-5", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.8991560577953898, "z_factor": 0.3581882985517303, "len": "len1"}, {"v": 5, "xy_factor": 0.8991560577953897, "z_factor": -0.5409677592436595, "len": "len1", "phi_init": "s"})), (">meta-bidiminished_icosahedron", {"tags": ("#johnson", "#diminished_icosahedron", "#rigid", "#equilateral", "#constrained"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.5296452721014835, "z_factor": 0.9629131066611835, "len": "len1"}, {"v": 4, "xy_factor": 1.0074451747140711, "z_factor": 0.10592905442029671, "len": "len1", "phi_mod_init": 0.23182380450040307, "phi_init": "s"}, {"v": 2, "xy_factor": 0.8569840522408868, "z_factor": -0.42371621768118684, "len": "len1", "phi_init": "s"}, {"v": 2, "xy_factor": 0.5296452721014835, "z_factor": -0.75105499782059, "len": "len1"})), (">tetracapped_octahedron", {"tags": ("#capped_octahedron", "#free"), "aristohedra": {">adamantane_cage": ((1,),)}, "point_group": "-43m", "axis": "-4", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "len": "len1"}, {"v": 2, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896257, "len": "len2"}, {"v": 4, "phi_init": "s", "z_factor": 0.0, "len": "len1"}, {"v": 2, "phi_init": "s", "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896257, "len": "len2"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len1"})), (">pentagonal_prism", {"tags": ("#prism", "#free", "#essential", "#equidistant"), "aristohedra": {">equilateral_pentagonal_prism": ()}, "point_group": "-[10]m2", "axis": "-10", "parameter_count": 2}, ({"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">pentagonal_antiprism", {"tags": ("#antiprism", "#free", "#essential", "#equidistant"), "aristohedra": {">equilateral_pentagonal_antiprism": ()}, "point_group": "-5m", "axis": "-5", "parameter_count": 2}, ({"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">ax-bicapped_tetragonal_prism", {"tags": ("#axis-capped_prism", "#free", "#essential"), "synonyms": (">elongated_tetragonal_bipyramid", ">ax-bicapped_square_prism"), "aristohedra": {">equilateral_trans-bicapped_cube": ("same",)}, "point_group": "4/mmm", "axis": "4/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">bicapped_tetragonal_antiprism", {"tags": ("#scalenohedron", "#free", "#essential"), "synonyms": (">ditetragonal_scalenohedron", ">bicapped_quare_antiprism"), "aristohedra": {">equilateral_bicapped_tetragonal_antiprism": ("same",), ">ax-bicapped_tetragonal_prism*": ("same",)}, "point_group": "-82m", "axis": "-8", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">monocapped_trigonal_cupola", {"tags": ("#capped_cupola", "#free"), "point_group": "3m", "axis": "3", "parameter_count": 5}, ({"v": 3, "xy": "xy1", "z": "z1"}, {"v": 6, "xy": "xy2", "z": "z2", "phi_mod": "phi_mod1", "phi_mod_frq": 3, "phi_init": "s"}, {"v": 1, "xy_factor": 0, "z": "z*"}))], 
    11: [(">cubic_11E_edshammar_polyhedron", {"tags": ("#edshammar", "#rigid", "#constrained"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.1805776390165883, "len": "len1"}, {"v": 3, "xy_factor": 0.8347944542658337, "z_factor": 0.5902888195082943, "len": "len1"}, {"v": 3, "xy_factor": 0.8347944542658337, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.8347944542658337, "z_factor": -0.5902888195082943, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.1805776390165883, "len": "len1"})), (">equilateral_fac-tricapped_cube", {"tags": ("#rigid", "#capped_cube", "#equilateral", "#constrained"), "point_group": "3m", "axis": "3", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.7986974958398153, "len": "len1"}, {"v": 6, "xy_factor": 0.9203568251371149, "z_factor": 0.3106045817154837, "z_mod_init": 0.16269763804144388, "z_mod_frq": 3, "len": "len1"}, {"v": 3, "xy_factor": 0.9203568251371149, "z_factor": -0.5028836084917356, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.1536741606575112, "len": "len1"})), (">equilateral_mer-tricapped_cube", {"tags": ("#rigid", "#capped_cube", "#equilateral", "#constrained"), "point_group": "mm2", "axis": "2", "parameter_count": 1}, ({"v": 1, "z_factor": 1.0130649142031731, "xy_factor": 0.0, "len": "len1"}, {"v": 4, "xy_factor": 0.787979577676755, "z_factor": 0.45587921139142795, "len": "len1"}, {"v": 2, "z_factor": -0.10130649142031732, "xy_factor": 1.1143714056234906, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.787979577676755, "z_factor": -0.6584921942320626, "len": "len1"})), (">monodiminished_icosahedron", {"tags": ("#rigid", "#johnson", "#axis-capped_antiprism", "#constrained", "#equilateral"), "synonyms": (">gyroelongated_pentagonal_pyramid", ">equilateral_monocapped_pentagonal_antiprism"), "point_group": "5m", "axis": "5", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.9138973034596459, "len": "len1"}, {"v": 5, "xy_factor": 0.8991560577953898, "z_factor": 0.3581882985517303, "len": "len1"}, {"v": 5, "xy_factor": 0.8991560577953897, "z_factor": -0.5409677592436595, "len": "len1", "phi_init": "s"})), (">equilateral_monocapped_pentagonal_prism", {"tags": ("#rigid", "#johnson", "#axis-capped_prism", "#constrained", "#equilateral"), "synonyms": (">elongated_pentagonal_pyramid",), "point_group": "5m", "axis": "5", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.9469100973808061, "len": "len1"}, {"v": 5, "xy_factor": 0.863811980822662, "z_factor": 0.4130449333430288, "len": "len1"}, {"v": 5, "xy_factor": 0.863811980822662, "z_factor": -0.60242695281919, "len": "len1"})), (">rhombohedral_11E_edshammar_polyhedron", {"tags": ("#edshammar", "#constrained"), "aristohedra": {">cubic_11E_edshammar_polyhedron": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 3, "z_factor": 0.5, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": 0.0, "xy": "xy1", "phi_init": "s"}, {"v": 3, "z_factor": -0.5, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"})), (">fully_capped_trigonal_prism", {"tags": ("#fully_capped_prism", "#free", "#essential"), "synonyms": (">pentacapped_trigonal_prism",), "aristohedra": {">rhombohedral_11E_edshammar_polyhedron": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">monocapped_pentagonal_frustum", {"tags": ("#axis-capped_frustum", "#essential", "#free"), "point_group": "5m", "axis": "5", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 5, "xy": "xy1", "z": "z2"}, {"v": 5, "xy": "xy2", "z": "z*"})), (">monocapped_pentagonal_antifrustum", {"tags": ("#axis-capped_antifrustum", "#essential", "#free"), "aristohedra": {">monodiminished_icosahedron": ("same",)}, "point_group": "5m", "axis": "5", "parameter_count": 4}, ({"v": 1, "xy_factor": 0, "z": "z1"}, {"v": 5, "xy": "xy1", "z": "z2"}, {"v": 5, "xy": "xy2", "z": "z*", "phi_init": "s"}))], 
    12: [(">platonic_icosahedron", {"tags": ("#platonic", "#frank-kasper", "#rigid", "#essential", "#free", "#equilateral", "#equidistant", "#deltahedron"), "synonyms": (">icosahedron", ">Z12", ">equilateral_snub_trigonal_antiprism"), "point_group": "-5m-3", "axis": "-3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.6070619982066864, "z_factor": 0.794654472291766, "len": "len1"}, {"v": 6, "xy_factor": 0.982246946376846, "z_factor": 0.0, "len": "len1", "z_mod_frq": 3, "z_mod_init": -0.18759247408508006}, {"v": 3, "xy_factor": 0.6070619982066864, "z_factor": -0.794654472291766, "len": "len1", "phi_init": "s"})), (">archimedean_cuboctahedron", {"tags": ("#archimedean", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "synonyms": (">cuboctahedron", ">equilateral_trigonal_gyrobicupola", ">equilateral_eq-tetracapped_tetragonal_prism", ">ccp-polyhedron"), "point_group": "m-3m", "axis": "-3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.5773502691896258, "z_factor": 0.816496580927726, "len": "len1"}, {"v": 6, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.5773502691896258, "z_factor": -0.8164965809277261, "len": "len1", "phi_init": "s"})), (">equilateral_bicapped_pentagonal_prism", {"tags": ("#rigid", "#johnson", "#axis-capped_prism", "#equilateral", "#constrained"), "synonyms": (">elongated_pentagonal_bipyramid",), "point_group": "5m", "axis": "5", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0327353142920117, "len": "len1"}, {"v": 5, "xy_factor": 0.8564594751359798, "z_factor": 0.503414248671081, "len": "len1"}, {"v": 5, "xy_factor": 0.8564594751359798, "z_factor": -0.503414248671081, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0327353142920117, "len": "len1"})), (">archimedean_truncated_tetrahedron", {"tags": ("#rigid", "#archimedean", "#essential", "#equilateral", "#constrained", "#equidistant"), "synonyms": (">equilateral_friauf_polyhedron",), "point_group": "-43m", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.49236596391733095, "z_factor": 0.8703882797784892, "len": "len1"}, {"v": 3, "xy_factor": 0.9847319278346619, "z_factor": 0.17407765595569785, "len": "len1"}, {"v": 6, "xy_factor": 0.8528028654224418, "z_factor": -0.5222329678670935, "len": "len1", "phi_init": "s"})), (">anticuboctahedron", {"tags": ("#orthobicupola", "#johnson", "#rigid", "#constrained", "#essential", "#equilateral", "#equidistant"), "synonyms": (">trigonal_orthobicupola", ">hcp-polyhedron"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.5773502691896258, "z_factor": 0.816496580927726, "len": "len1"}, {"v": 6, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.5773502691896258, "z_factor": -0.8164965809277261, "len": "len1"})), (">equilateral_eq-tetracapped_cube", {"tags": ("#rigid", "#capped_cube", "#equilateral"), "aristohedra": {">archimedean_cuboctahedron*": ((0, 1, 3, 4), (0, 2, 7, 8), (1, 2, 5, 6))}, "point_group": "4/mmm", "axis": "4/m", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.7764571353075622, "z_factor": 0.5490381056766579, "len": "len1"}, {"v": 4, "xy_factor": 1.0980762113533158, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7764571353075622, "z_factor": -0.5490381056766579, "len": "len1"})), (">equilateral_hexagonal_prism", {"tags": ("#rigid", "#equilateral", "#prism", "#constrained", "#equidistant"), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 1}, ({"v": 6, "z_factor": 0.4472135954999579, "xy_factor": 0.8944271909999159, "len": "len1"}, {"v": 6, "z_factor": -0.4472135954999579, "xy_factor": 0.8944271909999159, "len": "len1"})), (">equilateral_hexagonal_antiprism", {"tags": ("#rigid", "#equilateral", "#antiprism", "#constrained", "#equidistant"), "point_group": "-[12]2m", "axis": "-12", "parameter_count": 1}, ({"v": 6, "z_factor": 0.3933198931903286, "xy_factor": 0.9194016867619662, "len": "len1"}, {"v": 6, "z_factor": -0.3933198931903286, "xy_factor": 0.9194016867619662, "len": "len1", "phi_init": "s"})), (">pseudohexagonal_elongated_gyrobifastigium", {"tags": ("#rigid", "#constrained", "#elongated_bicupola", "#equidistant"), "point_group": "-42m", "axis": "-4", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.6546536707079771, "z_factor": 0.7559289460184544, "len": "len1"}, {"v": 4, "xy_factor": 0.9258200997725514, "z_factor": 0.3779644730092272, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.9258200997725514, "z_factor": -0.3779644730092272, "len": "len1", "phi_init": "s"}, {"v": 2, "xy_factor": 0.6546536707079771, "z_factor": -0.7559289460184544, "len": "len1", "phi_init": "s"})), (">trigonal_icosahedron", {"tags": ("#free",), "aristohedra": {">platonic_icosahedron": ()}, "point_group": "-3m", "axis": "-3", "parameter_count": 4}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy2", "z_mod": "z_mod1", "z_mod_frq": 3}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">trigonal_orthobicupola", {"tags": ("#orthobicupola", "#free"), "synonyms": (">unconstrained_anticuboctahedron",), "aristohedra": {">anticuboctahedron": ("same",)}, "point_group": "-6m2", "axis": "-6", "parameter_count": 4}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy2", "phi_mod": "phi_mod1", "phi_mod_frq": 3, "phi_init": "s"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">hexagonal_prism", {"tags": ("#prism", "#free", "#essential", "#equidistant"), "aristohedra": {">equilateral_hexagonal_prism": ()}, "point_group": "6/mmm", "axis": "6/m", "parameter_count": 2}, ({"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">hexagonal_antiprism", {"tags": ("#antiprism", "#free", "#essential", "#equidistant"), "aristohedra": {">equilateral_hexagonal_antiprism": ()}, "point_group": "-[12]2m", "axis": "-12", "parameter_count": 2}, ({"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">trigonal_gyrobicupola", {"tags": ("#gyrobicupola", "#free"), "synonyms": (">trigonal_cuboctahedron",), "aristohedra": {">archimedean_cuboctahedron": ("same", (0, 3, 8), (1, 4, 5), (2, 6, 7))}, "point_group": "-3m", "axis": "-3", "parameter_count": 3}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"})), (">eq-tetracapped_tetragonal_prism", {"tags": ("#equator-capped_prism", "#free"), "aristohedra": {">archimedean_cuboctahedron": ((0, 1, 3, 4), (0, 2, 7, 8), (1, 2, 5, 6)), ">equilateral_eq-tetracapped_cube": ()}, "synonyms": (">eq-tetracapped_square_prism", ">tetragonal_cuboctahedron"), "point_group": "4/mmm", "axis": "4/m", "parameter_count": 3}, ({"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">bicapped_pentagonal_prism", {"tags": ("#axis-capped_prism", "#free"), "synonyms": (">elongated_pentagonal_bipyramid",), "aristohedra": {">equilateral_bicapped_pentagonal_prism": ("same",)}, "point_group": "-[10]m2", "axis": "-10", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">pyritohedral_icosahedron", {"tags": ("#bundled", "#free", "#equidistant"), "synonyms": (">pyritohedral_cuboctahedron",), "aristohedra": {">platonic_icosahedron": ((0, 2, 6, 1), (0, 1, 5, 4), (0, 4, 10, 3), (0, 3, 9, 8), (0, 8, 7, 2)), ">archimedean_cuboctahedron": ([[0, 1, 4, 3], [2, 5, 9, 8], [7, 6, 10, 11]],)}, "parameter_bundles": {"var1": {"phi_mod1": "var1", "xyz": "0.5 * (sin(0.7853981633974483 - var1) + sin(0.7853981633974483 + var1))", "xyz_mod": "0.5 * (sin(0.7853981633974483  - var1) - sin(0.7853981633974483  + var1))"}}, "var_range": {"var1": (-0.7853981633974483, 0.7853981633974483, 0.0)}, "point_group": "m-3", "axis": "2/m", "parameter_count": 2}, ({"v": 4, "xy_factor": 0.7071067811865476, "z_factor": 0.7071067811865476, "xy": "xyz", "z": "xyz", "len": "len1", "xy_mod": "xyz_mod", "z_mod": "xyz_mod", "z_mod_off": 1}, {"v": 4, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1", "phi_mod": "phi_mod1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7071067811865476, "z_factor": -0.7071067811865476, "xy": "xyz", "z": "xyz", "len": "len1", "xy_mod": "xyz_mod", "z_mod": "xyz_mod"})), (">tetrahedral_cuboctahedron", {"tags": ("#bundled", "#free", "#equidistant"), "synonyms": (">elpasolite_cuboctahedron",), "aristohedra": {">archimedean_cuboctahedron": ([[0, 1, 4, 3], [2, 5, 9, 8], [7, 6, 10, 11]],)}, "parameter_bundles": {"var1": {"phi_mod1": "atan(sqrt(2) * tan(var1))", "xy1": "sqrt(1 - 0.5 * cos(var1) ** 2)", "xy2": "cos(var1)", "z1": "sqrt(0.5) * cos(var1)", "z_mod1": "sin(var1)"}}, "var_range": {"var1": (-0.6154797086703874, 0.6154797086703874, 0.0)}, "point_group": "-43m", "axis": "-4", "parameter_count": 2}, ({"v": 4, "xy_factor": 0.7071067811865475, "z_factor": 0.7071067811865475, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 1.5}, {"v": 4, "z_factor": 0.0, "xy": "xy2", "len": "len1", "z_mod": "z_mod1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7071067811865475, "z_factor": -0.7071067811865475, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 0.5})), (">generic_truncated_tetrahedron", {"tags": ("#bundled", "#free", "#equidistant"), "aristohedra": {">archimedean_truncated_tetrahedron": ([[0, 3], [1, 2, 6, 11], [4, 5, 7, 10], [8, 9]],)}, "parameter_bundles": {"var1": {"xy1": "sin(var1)", "xy2": "sqrt(1 - 0.5 * sin(var1) ** 2)", "z1": "cos(var1)", "z2": "0.7071067811865476 * sin(var1)", "phi_mod1": "asin(1 / sqrt(1 + 2 / tan(var1) ** 2)) - 0.3217505543966421"}}, "var_range": {"var1": (0, 0.9553166181245093, 0.4405106630046985)}, "point_group": "-43m", "axis": "-4", "parameter_count": 2}, ({"v": 2, "xy_factor": 0.4264014327112209, "z_factor": 0.9045340337332908, "xy": "xy1", "z": "z1", "len": "len1"}, {"v": 4, "xy_factor": 0.9534625892455924, "z_factor": 0.30151134457776363, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 0.5, "phi_mod_init": 0.3217505543966421, "phi_init": "s"}, {"v": 4, "xy_factor": 0.9534625892455924, "z_factor": -0.30151134457776363, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 1.5, "phi_mod_init": 0.3217505543966421, "phi_init": "s"}, {"v": 2, "xy_factor": 0.4264014327112209, "z_factor": -0.9045340337332908, "xy": "xy1", "z": "z1", "len": "len1", "phi_init": "s"})), (">bidisphenoidal_cuboctahedron", {"tags": ("#distorted_cuboctahedron", "#free"), "aristohedra": {">eq-tetracapped_tetragonal_prism": ("same", (0, 4, 7, 8), (1, 5, 6, 9))}, "point_group": "-42m", "axis": "-4", "parameter_count": 5}, ({"v": 4, "xy": "xy1", "z": "z1", "xy_mod": "xy_mod1", "z_mod": "z_mod1"}, {"v": 4, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 4, "xy": "xy1", "z": "z1", "z_factor": -1.0, "xy_mod": "xy_mod1", "xy_mod_off": 1.0, "z_mod": "z_mod1"})), (">anticuboidal_cuboctahedron", {"tags": ("#free", "#distorted_cuboctahedron"), "aristohedra": {">equilateral_eq-tetracapped_prism": (), ">tetrahedral_cuboctahedron": ("same", (0, 4, 7, 8), (1, 5, 6, 9))}, "point_group": "-42m", "axis": "-4", "parameter_count": 5}, ({"v": 4, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1"}, {"v": 4, "phi_init": "s", "xy": "xy2", "z_factor": 0.0, "z_mod": "z_mod1"}, {"v": 4, "xy": "xy1", "z": "z1", "z_factor": -1.0, "phi_mod": "phi_mod1", "phi_mod_off": 1.5})), (">bicapped_pentagonal_antiprism", {"tags": ("#scalenohedron", "#free"), "synonyms": (">dipentagonal_scalenohedron",), "aristohedra": {">platonic_icosahedron": ()}, "point_group": "-5m", "axis": "-5", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"}))], 
    13: [(">equilateral_pentacapped_cube", {"tags": ("#rigid", "#capped_cube", "#equilateral", "#constrained", "#essential"), "synonyms": (">monodiminished_rhombic_dodecahedron",), "point_group": "4mm", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0099870112916585, "len": "len1"}, {"v": 4, "xy_factor": 0.7736827199775547, "z_factor": 0.46291071350867674, "len": "len1"}, {"v": 4, "xy_factor": 1.0941525955659632, "z_factor": -0.08416558427430487, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7736827199775547, "z_factor": -0.6312418820572864, "len": "len1"})), (">monocapped_hexagonal_frustum", {"tags": ("#axis-capped_frustum", "#free", "#essential"), "point_group": "6mm", "axis": "6", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 6, "xy": "xy1", "z": "z2"}, {"v": 6, "xy": "xy2", "z": "z*"})), (">monocapped_hexagonal_antifrustum", {"tags": ("#axis-capped_antifrustum", "#free", "#essential"), "point_group": "6mm", "axis": "6", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z": "z1"}, {"v": 6, "xy": "xy1", "z": "z2"}, {"v": 6, "xy": "xy2", "z": "z*", "phi_init": "s"}))], 
    14: [(">rhombic_dodecahedron", {"tags": ("#catalan", "#rigid", "#constrained", "#essential", "#equilateral"), "synonyms": (">bcc-polyhedron",), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.082903768654761, "len": "len1"}, {"v": 4, "xy_factor": 0.7657285981882497, "z_factor": 0.5414518843273805, "len": "len1"}, {"v": 4, "xy_factor": 1.082903768654761, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7657285981882497, "z_factor": -0.5414518843273805, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.082903768654761, "len": "len1"})), (">rigid_bicapped_hexagonal_antiprism", {"tags": ("#axis-bicapped_antiprism", "#frank-kasper", "#constrained", "#rigid"), "point_group": "-[12]2m", "axis": "-12", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 0.9081233528976269, "len": "len1"}, {"v": 6, "xy_factor": 0.9081233528976269, "z_factor": 0.45406167644881346, "len": "len1"}, {"v": 6, "xy_factor": 0.9081233528976269, "z_factor": -0.45406167644881346, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -0.9081233528976269, "len": "len1"})), (">fully_capped_cube", {"tags": ("#capped_cube", "#free", "#essential"), "synonyms": (">hexacapped_cube", ">generic_tetrakis_hexahedron", ">generic_triakis_octahedron"), "aristohedra": {">rhombic_dodecahedron": ("same",)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len2"}, {"v": 4, "xy_factor": 0.8164965809277261, "z_factor": 0.5773502691896258, "len": "len1"}, {"v": 4, "z_factor": 0.0, "len": "len2", "phi_init": "s"}, {"v": 4, "xy_factor": 0.8164965809277261, "z_factor": -0.5773502691896258, "len": "len1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len2"})), (">bicapped_hexagonal_prism", {"tags": ("#frank-kasper", "#axis-bicapped_prism", "#free", "#essential"), "synonyms": (">elongated_hexagonal_bipyramid",), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">bicapped_hexagonal_antiprism", {"tags": ("#axis-bicapped_antiprism", "#essential", "#free"), "synonyms": (">dihexagonal_scalenohedron", ">Z14"), "aristohedra": {">rigid_bicapped_hexagonal_antiprism": ("same",)}, "point_group": "-[12]2m", "axis": "-12", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">fully_capped_tetragonal_prism", {"tags": ("#fully_capped_prism", "#free"), "synonyms": (">hexacapped_tetragonal_prism", ">fully_capped_square_prism"), "aristohedra": {">fully_capped_cube": ((0,), (5,), (6,))}, "point_group": "4/mmm", "axis": "4/m", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 4, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 4, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 4, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"}))], 
    15: [(">rigid_tricapped_trigonal_biprism", {"tags": ("#frank-kasper", "#constrained", "#rigid", "#essential", "#capped_biprism"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.456859994078892, "z_factor": 0.913719988157784, "len": "len1"}, {"v": 3, "xy_factor": 0.913719988157784, "z_factor": 0.45685999407889205, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.913719988157784, "z_factor": 0, "len": "len1"}, {"v": 3, "xy_factor": 0.913719988157784, "z_factor": -0.45685999407889194, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.456859994078892, "z_factor": -0.913719988157784, "len": "len1"})), (">eq-tricapped_ditrigonal_prism", {"tags": ("#equator-capped_prism", "#free"), "point_group": "-6m2", "axis": "-6", "parameter_count": 4}, ({"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "phi_mod_frq": 3}, {"v": 3, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_mod": "phi_mod1", "phi_mod_frq": 3})), (">tricapped_trigonal_biprism", {"tags": ("#frank-kasper", "#free", "#capped_biprism"), "aristohedra": {">rigid_tricapped_trigonal_biprism": ("same",)}, "synonyms": (">Z15",), "point_group": "-6m2", "axis": "-6", "parameter_count": 5}, ({"v": 3, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 3, "z_factor": 1.0, "xy": "xy2", "z": "z2", "phi_init": "s"}, {"v": 3, "z_factor": 0.0, "xy": "xy3"}, {"v": 3, "z_factor": -1.0, "xy": "xy2", "z": "z2", "phi_init": "s"}, {"v": 3, "z_factor": -1.0, "xy": "xy1", "z": "z1"})), (">pentacapped_pentagonal_prism", {"tags": ("#equator-capped_prism", "#free"), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 3}, ({"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1"}))], 
    16: [(">disphenocingulum", {"tags": ("#johnson", "#rigid", "#constrained", "#equilateral"), "point_group": "-42m", "axis": "-4", "parameter_count": 1}, ({"v": 2, "xy_factor": 0.4507644650993652, "z_factor": 0.9956827563941761, "len": "len1"}, {"v": 6, "xy_factor": 0.8888670460452481, "z_factor": 0.37590696713708943, "len": "len1", "phi_mod_init": -0.0623839541592264, "phi_mod_off": 0.0, "xy_mod_init": 0.12669010042772172, "z_mod_init": -0.08290738191757648}, {"v": 6, "xy_factor": 0.8888670460452481, "z_factor": -0.37590696713708943, "len": "len1", "phi_mod_init": -0.0623839541592264, "phi_mod_off": 1.0, "xy_mod_init": 0.12669010042772172, "xy_mod_off": 1.0, "z_mod_init": 0.08290738191757648, "z_mod_off": 1.0, "phi_init": "s"}, {"v": 2, "xy_factor": 0.4507644650993652, "z_factor": -0.9956827563941761, "len": "len1", "phi_init": "s"})), (">equilateral_snub_tetragonal_antiprism", {"tags": ("#johnson", "#rigid", "#constrained", "#equilateral"), "synonyms": (">snub_square_antiprism",), "point_group": "-82m", "axis": "-8", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.6046110190680494, "z_factor": 0.5787558115816654, "len": "len1"}, {"v": 8, "xy_factor": 1.0373503139007878, "z_factor": 0.0, "len": "len1", "z_mod_frq": 4, "z_mod_init": -0.1587031170812634}, {"v": 4, "xy_factor": 0.6046110190680494, "z_factor": -0.5787558115816654, "len": "len1", "phi_init": "s"})), (">rigid_tetracapped_truncated_tetrahedron", {"tags": ("#rigid", "#constrained", "#frank-kasper", "#essential"), "synonyms": (">rigid_tetracapped_friauf_polyhedron", ">rigid_hexakis_truncated_tetrahedron"), "point_group": "-43m", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.4869527613691743, "z_factor": 0.8608189992041445, "len": "len1"}, {"v": 6, "xy_factor": 0.9739055227383485, "z_factor": 0.25824569976124334, "len": "len1", "z_mod_frq": 3, "z_mod_init": -0.08608189992041443}, {"v": 6, "xy_factor": 0.8434269235773731, "z_factor": -0.5164913995224867, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0, "z_factor": -1.0329827990449734, "len": "len1"})), (">generic_tetracapped_truncated_tetrahedron", {"tags": ("#bundled", "#free"), "synonyms": (">generic_tetracapped_friauf_polyhedron", ">generic_hexakis_truncated_tetrahedron", ">Z16"), "aristohedra": {">rigid_tetracapped_truncated_tetrahedron": ([[0, 3], [4, 8], [1, 2, 9, 14], [5, 7, 10, 13], [6, 15], [11, 12]],)}, "parameter_bundles": {"var1": {"xy1": "sin(var1)", "xy2": "sqrt(1 - 0.5 * sin(var1) ** 2)", "z1": "cos(var1)", "z2": "0.7071067811865476 * sin(var1)", "phi_mod1": "asin(1 / sqrt(1 + 2 / tan(var1) ** 2)) - 0.3217505543966421"}}, "var_range": {"var1": (0, 0.9553166181245093, 0.4405106630046985)}, "point_group": "-43m", "axis": "-4", "parameter_count": 3}, ({"v": 2, "xy_factor": 0.4264014327112209, "z_factor": 0.9045340337332908, "xy": "xy1", "z": "z1", "len": "len1"}, {"v": 2, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896257, "len": "len2", "phi_init": "s"}, {"v": 4, "xy_factor": 0.9534625892455924, "z_factor": 0.30151134457776363, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 0.5, "phi_mod_init": 0.3217505543966421, "phi_init": "s"}, {"v": 4, "xy_factor": 0.9534625892455924, "z_factor": -0.30151134457776363, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_off": 1.5, "phi_mod_init": 0.3217505543966421, "phi_init": "s"}, {"v": 2, "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896257, "len": "len2"}, {"v": 2, "xy_factor": 0.4264014327112209, "z_factor": -0.9045340337332908, "xy": "xy1", "z": "z1", "len": "len1", "phi_init": "s"}))], 
    17: [(">fully_capped_pentagonal_prism", {"tags": ("#fully_capped_prism", "#free"), "synonyms": (">heptacapped_pentagonal_prism",), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 5, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 5, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 5, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"})), (">pentacapped_ditrigonal_prism", {"tags": ("#capped_prism", "#free"), "point_group": "-6m2", "axis": "-6", "parameter_count": 5}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z1"}, {"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z2", "phi_mod": "phi_mod1", "phi_mod_frq": 3}, {"v": 3, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z2", "phi_mod": "phi_mod1", "phi_mod_frq": 3}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z1"}))], 
    18: [(">equilateral_elongated_trigonal_orthobicupola", {"tags": ("#elongated_bicupola", "#johnson", "#rigid", "#constrained", "#equilateral"), "point_group": "-6m2", "axis": "-6", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.4714860333348884, "z_factor": 1.0751008252092626, "len": "len1"}, {"v": 6, "xy_factor": 0.8166377647951402, "z_factor": 0.4083188823975701, "len": "len1", "phi_init": "s"}, {"v": 6, "xy_factor": 0.8166377647951402, "z_factor": -0.4083188823975701, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.4714860333348884, "z_factor": -1.0751008252092626, "len": "len1"})), (">equilateral_elongated_trigonal_gyrobicupola", {"tags": ("#elongated_bicupola", "#johnson", "#rigid", "#constrained", "#equilateral"), "point_group": "-3m", "axis": "-3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.4714860333348884, "z_factor": 1.0751008252092626, "len": "len1"}, {"v": 6, "xy_factor": 0.8166377647951402, "z_factor": 0.4083188823975701, "len": "len1", "phi_init": "s"}, {"v": 6, "xy_factor": 0.8166377647951402, "z_factor": -0.4083188823975701, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.4714860333348884, "z_factor": -1.0751008252092626, "len": "len1", "phi_init": "s"})), (">equilateral_gyroelongated_trigonal_bicupola", {"tags": ("#elongated_bicupola", "#johnson", "#chiral", "#rigid", "#constrained", "#equilateral"), "point_group": "32", "axis": "3", "parameter_count": 1}, ({"v": 3, "xy_factor": 0.48830816652259756, "z_factor": 1.052394249465024, "len": "len1"}, {"v": 6, "xy_factor": 0.845774554167943, "z_factor": 0.3618222177512266, "len": "len1", "phi_init": "s"}, {"v": 6, "xy_factor": 0.845774554167943, "z_factor": -0.3618222177512266, "len": "len1"}, {"v": 3, "xy_factor": 0.48830816652259756, "z_factor": -1.052394249465024, "len": "len1", "phi_init": 0.5235987755982988})), (">hexacapped_cuboctahedron", {"tags": ("#capped_cuboctahedron", "#essential", "#free"), "synonyms": (">tetrakis_cuboctahedron",), "point_group": "m-3m", "axis": "-3", "parameter_count": 2}, ({"v": 3, "xy_factor": 0.5773502691896256, "z_factor": 0.816496580927726, "len": "len1"}, {"v": 3, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896256, "len": "len2", "phi_init": "s"}, {"v": 6, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 3, "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896256, "len": "len2"}, {"v": 3, "xy_factor": 0.5773502691896256, "z_factor": -0.816496580927726, "len": "len1", "phi_init": "s"})), (">eq-hexacapped_hexagonal_prism", {"tags": ("#equator-capped_prism", "#free", "#essential"), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 3}, ({"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1"}))], 
    20: [(">platonic_dodecahedron", {"tags": ("#truncated_trapezohedron", "#fullerene", "#platonic", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "synonyms": (">dodecahedron", ">pentagonal_dodecahedron", ">F20"), "point_group": "-5m-3", "axis": "-5", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.6070619982066862, "z_factor": 0.7946544722917661, "len": "len1"}, {"v": 10, "xy_factor": 0.982246946376846, "z_factor": 0.0, "z_mod_init": 0.1875924740850799, "z_mod_frq": 5, "len": "len1"}, {"v": 5, "xy_factor": 0.6070619982066862, "z_factor": -0.7946544722917661, "len": "len1", "phi_init": "s"})), (">pyritohedron", {"tags": ("#fullerene", "#bundled", "#constrained"), "aristohedra": {">platonic_dodecahedron": ((0, 1), (1, 2), (2, 3), (3, 4), (4, 0))}, "parameter_bundles": {"var1": {"xy_ico": "(1 - var1 ** 2) * 5 / (sqrt(12) + 3 * sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))", "z_ico": "(1 + var1) * 5 / (sqrt(12) + 3 * sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))", "len_ico": "sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2) * 5 / (sqrt(12) + 3 * sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))", "xy_cube": "sqrt(50) / (sqrt(12) + 3 * sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))", "z_cube": "5 / (sqrt(12) + 3 * sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))", "phimo": "asin(0.7071067811865476 * (var1 ** 2 + var1) / (sqrt(var1 ** 4 - var1 ** 2 + 2 * var1 + 2))) - 0.42053433528396517"}}, "var_range": {"var1": (0, 1, 0.6180339887498949)}, "point_group": "m-3", "axis": "2/m", "parameter_count": 2}, ({"v": 2, "xy_factor": 0.35682208977308993, "z_factor": 0.9341723589627158, "xy": "xy_ico", "z": "z_ico", "len": "len1"}, {"v": 4, "xy_factor": 0.8164965809277261, "z_factor": 0.5773502691896258, "xy": "xy_cube", "z": "z_cube", "len": "len1", "phi_init": "s"}, {"v": 2, "xy_factor": 0.9341723589627158, "z_factor": 0.35682208977308993, "xy": "z_ico", "z": "xy_ico", "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 1.0, "z_factor": 0.0, "xy": "len_ico", "len": "len1", "phi_mod": "phimo", "phi_mod_init": 0.42053433528396517, "phi_init": "s"}, {"v": 2, "xy_factor": 0.9341723589627158, "z_factor": -0.35682208977308993, "xy": "z_ico", "z": "xy_ico", "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.8164965809277261, "z_factor": -0.5773502691896258, "xy": "xy_cube", "z": "z_cube", "len": "len1", "phi_init": "s"}, {"v": 2, "xy_factor": 0.35682208977308993, "z_factor": -0.9341723589627158, "xy": "xy_ico", "z": "z_ico", "len": "len1"})), (">faceted_pyritohedron", {"tags": ("#pyritohedron", "#bundled", "#free"), "aristohedra": {">pyritohedron": ("same",)}, "parameter_bundles": {"var1": {"xyz1": "cos((0.7853981633974483) * (1 + var1))", "xyz2": "sin((0.7853981633974483) * (1 + var1))", "phi_mod1": "(var1 - 0.5) * 0.7853981633974483"}}, "var_range": {"var1": (0, 1, 0.5)}, "point_group": "m-3", "axis": "2/m", "parameter_count": 3}, ({"v": 2, "xy_factor": 0.38268343236508984, "z_factor": 0.9238795325112867, "xy": "xyz1", "z": "xyz2", "len": "len1"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896257, "len": "len2", "phi_init": "s"}, {"v": 2, "xy_factor": 0.9238795325112867, "z_factor": 0.38268343236508984, "xy": "xyz2", "z": "xyz1", "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1", "phi_mod": "phi_mod1", "phi_mod_init": 0.39269908169872414, "phi_init": "s"}, {"v": 2, "xy_factor": 0.9238795325112867, "z_factor": -0.38268343236508984, "xy": "xyz2", "z": "xyz1", "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896257, "len": "len2", "phi_init": "s"}, {"v": 2, "xy_factor": 0.38268343236508984, "z_factor": -0.9238795325112867, "xy": "xyz1", "z": "xyz2", "len": "len1"})), (">fully_capped_hexagonal_prism", {"tags": ("#fully_capped_prism", "#free"), "synonyms": (">octacapped_hexagonal_prism",), "point_group": "6/mmm", "axis": "6/m", "parameter_count": 4}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "z": "z2"}, {"v": 6, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 6, "z_factor": 0.0, "xy": "xy2", "phi_init": "s"}, {"v": 6, "z_factor": -1.0, "xy": "xy1", "z": "z1"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "z": "z2"}))], 
    24: [(">archimedean_truncated_octahedron", {"tags": ("#archimedean", "#rigid", "#essential", "#constrained", "#equilateral", "#equidistant"), "synonyms": (">sodalite_cage",), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.4472135954999579, "z_factor": 0.8944271909999159, "len": "len1"}, {"v": 4, "xy_factor": 0.8944271909999159, "z_factor": 0.4472135954999579, "len": "len1"}, {"v": 8, "z_factor": 0.0, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.07094852730208198, "phi_init": "s"}, {"v": 4, "xy_factor": 0.8944271909999159, "z_factor": -0.4472135954999579, "len": "len1"}, {"v": 4, "xy_factor": 0.4472135954999579, "z_factor": -0.8944271909999159, "len": "len1"})), (">archimedean_rhombicuboctahedron", {"tags": ("#archimedean", "#rigid", "#constrained", "#equilateral", "#essential", "#equidistant"), "synonyms": (">interpenetrating_truncated_tetrahedra",), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.5054494651244236, "z_factor": 0.8628562094610168, "len": "len1"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": 0.35740674433659325, "len": "len1", "phi_init": "s"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": -0.35740674433659325, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.5054494651244236, "z_factor": -0.8628562094610168, "len": "len1"})), (">archimedean_truncated_cube", {"tags": ("#rigid", "#archimedean", "#constrained", "#equilateral", "#equidistant"), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 8, "xy_factor": 0.7345095552677554, "z_factor": 0.6785983445458469, "len": "len1"}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": 0.28108463771482023, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": -0.28108463771482023, "len": "len1", "phi_init": "s"}, {"v": 8, "xy_factor": 0.7345095552677554, "z_factor": -0.6785983445458469, "len": "len1"})), (">archimedean_snub_cube", {"tags": ("#archimedean", "#rigid", "#chiral", "#constrained", "#essential", "#equilateral", "#equidistant"), "synonyms": (">snub_cuboctahedron", ">snub_cube"), "point_group": "432", "axis": "4", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.5262333433625158, "z_factor": 0.8503402074073109, "len": "len1"}, {"v": 8, "xy_factor": 0.9273034226662824, "z_factor": 0.3568396367515094, "len": "len1", "xy_mod_frq": 4, "xy_mod_init": 0.04059059593245873, "z_mod_frq": 4, "z_mod_init": -0.10548099106614693}, {"v": 8, "xy_factor": 0.9273034226662824, "z_factor": -0.3568396367515094, "len": "len1", "xy_mod_frq": 4, "xy_mod_init": 0.04059059593245873, "z_mod_frq": 4, "z_mod_init": 0.10548099106614693, "phi_init": 0.5748262975155555}, {"v": 4, "xy_factor": 0.5262333433625158, "z_factor": -0.8503402074073109, "len": "len1", "phi_init": 0.5748262975155555})), (">equilateral_elongated_tetragonal_gyrobicupola", {"tags": ("#johnson", "#rigid", "#constrained", "#equilateral", "#equidistant"), "synonyms": (">equilateral_elongated_square_gyrobicupola",), "point_group": "-82m", "axis": "-8", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.5054494651244236, "z_factor": 0.8628562094610168, "len": "len1"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": 0.35740674433659325, "len": "len1", "phi_init": "s"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": -0.35740674433659325, "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.5054494651244236, "z_factor": -0.8628562094610168, "len": "len1", "phi_init": "s"})), (">generic_rhombicuboctahedron", {"tags": ("#bundled", "#free", "#equidistant"), "aristohedra": {">archimedean_rhombicuboctahedron": ("same",)}, "parameter_bundles": {"var1": {"xy1": "sin(var1)", "xy2": "sqrt(1 - 0.5 * sin(var1) ** 2)", "z1": "cos(var1)", "z2": "0.7071067811865476 * sin(var1)", "phi_mod1": "asin(1 / sqrt(1 + 2 / tan(var1) ** 2)) - 0.39269908169872414"}}, "var_range": {"var1": (0, 0.9553166181245093, 0.5299027897489266)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 2}, ({"v": 4, "xy_factor": 0.5054494651244236, "z_factor": 0.8628562094610168, "xy": "xy1", "z": "z1", "len": "len1"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": 0.35740674433659325, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4, "phi_init": "s"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": -0.35740674433659325, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4, "phi_init": "s"}, {"v": 4, "xy_factor": 0.5054494651244236, "z_factor": -0.8628562094610168, "xy": "xy1", "z": "z1", "len": "len1"})), (">generic_truncated_octahedron", {"tags": ("#bundled", "#free", "#equidistant"), "synonyms": (">generic_sodalite_cage",), "aristohedra": {">archimedean_truncated_octahedron": ("same",)}, "parameter_bundles": {"var1": {"xyz1": "sin(var1)", "xyz2": "cos(var1)", "phi_mod1": "0.4636476090008061 - var1"}}, "var_range": {"var1": (0, 0.7853981633974483, 0.4636476090008061)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 2}, ({"v": 4, "xy_factor": 0.4472135954999579, "z_factor": 0.8944271909999159, "xy": "xyz1", "z": "xyz2", "len": "len1"}, {"v": 4, "xy_factor": 0.8944271909999159, "z_factor": 0.4472135954999579, "xy": "xyz2", "z": "xyz1", "len": "len1"}, {"v": 8, "z_factor": 0.0, "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4, "phi_mod_init": -0.07094852730208198, "phi_init": "s"}, {"v": 4, "xy_factor": 0.8944271909999159, "z_factor": -0.4472135954999579, "xy": "xyz2", "z": "xyz1", "len": "len1"}, {"v": 4, "xy_factor": 0.4472135954999579, "z_factor": -0.8944271909999159, "xy": "xyz1", "z": "xyz2", "len": "len1"})), (">generic_truncated_cube", {"tags": ("#bundled", "#free", "#equidistant"), "aristohedra": {">archimedean_truncated_cube": ("same",)}, "parameter_bundles": {"var1": {"xy1": "sqrt(1 - 0.5 * cos(var1) ** 2)", "z1": "0.7071067811865476 * cos(var1)", "xy2": "cos(var1)", "z2": "sin(var1)", "phi_mod1": "0.39269908169872414 - atan(1.4142135623730951 * tan(var1))"}}, "var_range": {"var1": (0.0, 0.6154797086703874, 0.28492412662206235)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 2}, ({"v": 8, "xy_factor": 0.7345095552677554, "z_factor": 0.6785983445458469, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": 0.28108463771482023, "xy": "xy2", "z": "z2", "len": "len1", "phi_init": "s"}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": -0.28108463771482023, "xy": "xy2", "z": "z2", "len": "len1", "phi_init": "s"}, {"v": 8, "xy_factor": 0.7345095552677554, "z_factor": -0.6785983445458469, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4})), (">ax-truncated_hexagonal_trapezohedron", {"tags": ("#truncated_trapezohedron", "#fullerene", "#constrained"), "synonyms": (">F24",), "parameter_bundles": {"var1": {"z_mod1": "var1", "xy1": "var3 * (1 - (var2 - var1) / (var1 * 12.928203230275503))"}, "var2": {"z1": "var2", "xy1": "var3 * (1 - (var2 - var1) / (var1 * 12.928203230275503))"}, "var3": {"xy2": "var3", "xy1": "var3 * (1 - (var2 - var1) / (var1 * 12.928203230275503))"}}, "var_range": {"var1": (0.000000001, 3.0, 1.0), "var2": (0.0, 10.0, 1.0), "var3": (0.0, 15.0, 1.0)}, "point_group": "-[12]2m", "axis": "-12", "parameter_count": 3}, ({"v": 6, "xy_factor": 1.0, "z_factor": 1.0, "xy": "xy1", "z": "z1"}, {"v": 12, "z_factor": 0.0, "xy": "xy2", "z_mod": "z_mod1", "z_mod_init": 1.0, "z_mod_frq": 6}, {"v": 6, "xy_factor": 1.0, "z_factor": -1.0, "xy": "xy1", "z": "z1", "phi_init": "s"}))], 
    26: [(">deltoidal_icositetrahedron", {"tags": ("#rigid", "#catalan", "#essential", "#constrained"), "synonyms": (">strombic_icositetrahedron",), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0164858231293283, "len": "len1"}, {"v": 8, "xy_factor": 0.7524871042249965, "z_factor": 0.6373492876279603, "len": "len1", "xy_mod_frq": 4, "xy_mod_init": 0.03372308571025889, "z_mod_frq": 4, "z_mod_init": -0.08141473088677748}, {"v": 8, "xy_factor": 1.0164858231293283, "z_factor": 0.0, "len": "len1"}, {"v": 8, "xy_factor": 0.7524871042249965, "z_factor": -0.6373492876279603, "len": "len1", "xy_mod_frq": 4, "xy_mod_init": 0.03372308571025889, "z_mod_frq": 4, "z_mod_init": 0.08141473088677748}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0164858231293283, "len": "len1"})), (">fully_capped_cuboctahedron", {"tags": ("#capped_cuboctahedron", "#essential", "#free"), "aristohedra": {">deltoidal_icositetrahedron": ([[0], [1, 3, 5, 7], [2, 4, 6, 8], [9, 11, 13, 15], [10, 12, 14, 16], [18, 20, 22, 24], [17, 19, 21, 23], [25]], [[0], [2, 4, 6, 8], [1, 3, 5, 7], [10, 12, 14, 16], [9, 11, 13, 15], [17, 19, 21, 23], [18, 20, 22, 24], [25]])}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 4, "xy_factor": 0.7071067811865476, "z_factor": 0.7071067811865476, "len": "len2"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896257, "len": "len3", "phi_init": "s"}, {"v": 4, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1"}, {"v": 4, "xy_factor": 1.0, "z_factor": 0.0, "len": "len2", "phi_init": "s"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896257, "len": "len3", "phi_init": "s"}, {"v": 4, "xy_factor": 0.7071067811865476, "z_factor": -0.7071067811865476, "len": "len2"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len1"}))], 
    30: [(">archimedean_icosidodecahedron", {"tags": ("#archimedean", "#rigid", "#essential", "#free", "#equilateral", "#equidistant"), "synonyms": (">pentagonal_gyrobirotunda", ">icosidodecahedron"), "point_group": "-5m-3", "axis": "-5", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.5257311121191336, "z_factor": 0.85065080835204, "len": "len1"}, {"v": 5, "xy_factor": 0.85065080835204, "z_factor": 0.5257311121191336, "len": "len1", "phi_init": "s"}, {"v": 10, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 5, "xy_factor": 0.85065080835204, "z_factor": -0.5257311121191336, "len": "len1"}, {"v": 5, "xy_factor": 0.5257311121191336, "z_factor": -0.85065080835204, "len": "len1", "phi_init": "s"})), (">equilateral_pentagonal_orthobirotunda", {"tags": ("#johnson", "#rigid", "#constrained", "#equilateral", "#equidistant"), "point_group": "-[10]m2", "axis": "-10", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.5257311121191336, "z_factor": 0.85065080835204, "len": "len1"}, {"v": 5, "xy_factor": 0.85065080835204, "z_factor": 0.5257311121191336, "len": "len1", "phi_init": "s"}, {"v": 10, "xy_factor": 1.0, "z_factor": 0.0, "len": "len1", "phi_init": "s"}, {"v": 5, "xy_factor": 0.85065080835204, "z_factor": -0.5257311121191336, "len": "len1", "phi_init": "s"}, {"v": 5, "xy_factor": 0.5257311121191336, "z_factor": -0.85065080835204, "len": "len1"})), (">hexacapped_generic_truncated_cube", {"tags": ("#capped_truncated_cube", "#bundled", "#free"), "parameter_bundles": {"var1": {"xy1": "sqrt(1 - 0.5 * cos(var1) ** 2)", "z1": "0.7071067811865476 * cos(var1)", "xy2": "cos(var1)", "z2": "sin(var1)", "phi_mod1": "0.39269908169872414 - atan(1.4142135623730951 * tan(var1))"}}, "var_range": {"var1": (0.0, 0.6154797086703874, 0.28492412662206235)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 3}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len2"}, {"v": 8, "xy_factor": 0.7345095552677554, "z_factor": 0.6785983445458469, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": 0.28108463771482023, "xy": "xy2", "z": "z2", "len": "len1", "phi_init": "s"}, {"v": 4, "z_factor": 0.0, "len": "len2", "phi_init": 1.1780972450961724}, {"v": 4, "xy_factor": 0.9596829822606672, "z_factor": -0.28108463771482023, "xy": "xy2", "z": "z2", "len": "len1", "phi_init": "s"}, {"v": 8, "xy_factor": 0.7345095552677554, "z_factor": -0.6785983445458469, "xy": "xy1", "z": "z1", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len2"}))], 
    32: [(">equilateral_chamfered_cube", {"tags": ("#rigid", "#essential", "#constrained", "#equilateral", "#chamfered_cube"), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.38935525265229765, "z_factor": 0.9111302045547451, "len": "len1"}, {"v": 4, "xy_factor": 0.8389439725106952, "z_factor": 0.5932229719978931, "len": "len1"}, {"v": 8, "xy_factor": 0.9518177378237596, "z_factor": 0.27531573944104115, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.09925312964308414, "phi_init": "s"}, {"v": 8, "xy_factor": 0.9518177378237596, "z_factor": -0.27531573944104115, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.09925312964308414, "phi_init": "s"}, {"v": 4, "xy_factor": 0.8389439725106952, "z_factor": -0.5932229719978931, "len": "len1"}, {"v": 4, "xy_factor": 0.38935525265229765, "z_factor": -0.9111302045547451, "len": "len1"})), (">rhombic_triacontahedron", {"tags": ("#rigid", "#constrained", "#catalan", "#equilateral", "#essential"), "synonyms": (">rhombic_triacontahedron",), "point_group": "-5m-3", "axis": "-5", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0591866921461843, "len": "len1"}, {"v": 5, "xy_factor": 0.5855040032457821, "z_factor": 0.7664346905232933, "len": "len1"}, {"v": 5, "xy_factor": 0.9473653778008043, "z_factor": 0.47368268890040216, "len": "len1", "phi_init": "s"}, {"v": 10, "xy_factor": 0.9473653778008043, "z_factor": 0.0, "z_mod_init": 0.18093068727751108, "z_mod_frq": 5, "len": "len1"}, {"v": 5, "xy_factor": 0.9473653778008043, "z_factor": -0.47368268890040216, "len": "len1"}, {"v": 5, "xy_factor": 0.5855040032457821, "z_factor": -0.7664346905232933, "len": "len1", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0591866921461843, "len": "len1"})), (">chamfered_cube", {"tags": ("#bundled", "#free", "#chamfered_cube"), "aristohedra": {">equilateral_chamfered_cube": ("same",)}, "parameter_bundles": {"var1": {"xy1": "sin(var1)", "xy2": "sqrt(1 - 0.5 * sin(var1) ** 2)", "z1": "cos(var1)", "z2": "0.7071067811865476 * sin(var1)", "phi_mod1": "asin(1 / sqrt(1 + 2 / tan(var1) ** 2)) - 0.39269908169872414"}}, "var_range": {"var1": (0, 0.9553166181245093, 0.5299027897489266)}, "point_group": "m-3m", "axis": "4/m", "parameter_count": 3}, ({"v": 4, "xy_factor": 0.5054494651244236, "z_factor": 0.8628562094610168, "xy": "xy1", "z": "z1", "len": "len1"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": 0.5773502691896257, "len": "len2"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": 0.35740674433659325, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4, "phi_init": "s"}, {"v": 8, "xy_factor": 0.9339488310944647, "z_factor": -0.35740674433659325, "xy": "xy2", "z": "z2", "len": "len1", "phi_mod": "phi_mod1", "phi_mod_frq": 4, "phi_init": "s"}, {"v": 4, "xy_factor": 0.816496580927726, "z_factor": -0.5773502691896257, "len": "len2"}, {"v": 4, "xy_factor": 0.5054494651244236, "z_factor": -0.8628562094610168, "xy": "xy1", "z": "z1", "len": "len1"})), (">fully_capped_platonic_dodecahedron", {"tags": ("#free", "#capped_platonic_dodecahedron", "#essential"), "synonyms": (">fully_capped_platonic_icosahedron", ">fully_capped_dodecahedron"), "aristohedra": {">rhombic_triacontahedron": ("same",)}, "point_group": "-5m-3", "axis": "-5", "parameter_count": 2}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0, "len": "len1"}, {"v": 5, "xy_factor": 0.6070619982066862, "z_factor": 0.7946544722917661, "len": "len2"}, {"v": 5, "xy_factor": 0.8944271909999159, "z_factor": 0.4472135954999579, "len": "len1", "phi_init": "s"}, {"v": 10, "xy_factor": 0.982246946376846, "z_factor": 0.0, "z_mod_init": 0.1875924740850799, "z_mod_frq": 5, "len": "len2"}, {"v": 5, "xy_factor": 0.8944271909999157, "z_factor": -0.4472135954999579, "len": "len1"}, {"v": 5, "xy_factor": 0.6070619982066862, "z_factor": -0.7946544722917661, "len": "len2", "phi_init": "s"}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0, "len": "len1"}))], 
    38: [(">pentagonal_icositetrahedron", {"tags": ("#catalan", "#rigid", "#chiral", "#constrained", "#essential"), "synonyms": (">snub_cube_dual", ">pentagonal_icositetrahedron"), "point_group": "432", "axis": "4", "parameter_count": 1}, ({"v": 1, "xy_factor": 0.0, "z_factor": 1.0516322721686224, "len": "len1"}, {"v": 4, "xy_factor": 0.5962175254980832, "z_factor": 0.7907314860145546, "len": "len1"}, {"v": 8, "xy_factor": 0.808592035860624, "z_factor": 0.5717609117704834, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.10528593294094657, "phi_init": "s"}, {"v": 12, "xy_factor": 1.001070996205704, "z_factor": 0.0, "len": "len1", "phi_mod_off": 0.0, "phi_mod_frq": 4, "phi_mod_init": 0.11829764019264381, "xy_mod_frq": 4, "xy_mod_init": 0.050561275962918296, "z_mod_off": 0.75, "z_mod_frq": 4, "z_mod_init": 0.19515736379459842, "phi_init": -0.2874131487577779}, {"v": 8, "xy_factor": 0.808592035860624, "z_factor": -0.5717609117704834, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": 0.10528593294094657, "phi_init": 0.6032709475806171}, {"v": 4, "xy_factor": 0.5962175254980832, "z_factor": -0.7907314860145546, "len": "len1", "phi_init": 0.9959700292793412}, {"v": 1, "xy_factor": 0.0, "z_factor": -1.0516322721686224, "len": "len1"}))], 
    48: [(">archimedean_truncated_cuboctahedron", {"tags": ("#archimedean", "#rigid", "#essential", "#constrained", "#equilateral", "#equidistant"), "synonyms": (">zeolite_type_a_supercage", ">truncated_cuboctahedron"), "point_group": "m-3m", "axis": "4/m", "parameter_count": 1}, ({"v": 8, "xy_factor": 0.5637542339863973, "z_factor": 0.8259425910209562, "len": "len1"}, {"v": 8, "xy_factor": 0.8536537088595362, "z_factor": 0.5208409981466113, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.13720370805020238}, {"v": 8, "xy_factor": 0.9764509762465133, "z_factor": 0.21573940527226648, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": 0.16991845472706094}, {"v": 8, "xy_factor": 0.9764509762465133, "z_factor": -0.21573940527226648, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": 0.16991845472706094}, {"v": 8, "xy_factor": 0.8536537088595362, "z_factor": -0.5208409981466113, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": -0.13720370805020238}, {"v": 8, "xy_factor": 0.5637542339863973, "z_factor": -0.8259425910209562, "len": "len1"})), (">faujasite_supercage", {"tags": ("#rigid", "#constrained", "#essential", "#equilateral"), "point_group": "-43m", "axis": "-4", "parameter_count": 1}, ({"v": 4, "xy_factor": 0.2771005006468468, "z_factor": 1.0067042239967168, "len": "len1"}, {"v": 4, "xy_factor": 0.5518633347961824, "z_factor": 0.7804526126452052, "len": "len1", "phi_mod_init": 0.4224295147918075}, {"v": 4, "xy_factor": 0.8125860398898582, "z_factor": 0.5033521119983584, "len": "len1", "phi_mod_init": 0.2821632846979857}, {"v": 12, "erase": [1, 2, 7, 8], "xy_factor": 0.9546963612240688, "z_factor": 0.2432012411166233, "len": "len1", "phi_mod_off": 0, "phi_mod_frq": 4, "phi_mod_init": 0.05683140072325557, "xy_mod_frq": 4, "xy_mod_init": 0.05200786277264791, "z_mod_frq": 4, "z_mod_init": 0.033899259530223554}, {"v": 8, "xy_factor": 1.0441446652986188, "z_factor": 0.0, "len": "len1", "phi_mod_frq": 4, "phi_mod_init": 0.12409569523135061, "phi_init": "s"}, {"v": 12, "erase": [4, 5, 10, 11], "xy_factor": 0.9546963612240688, "z_factor": -0.2432012411166233, "len": "len1", "phi_mod_off": 0, "phi_mod_frq": 4, "phi_mod_init": 0.05683140072325557, "xy_mod_frq": 4, "xy_mod_init": 0.05200786277264791, "z_mod_frq": 4, "z_mod_init": -0.033899259530223554}, {"v": 4, "xy_factor": 0.8125860398898582, "z_factor": -0.5033521119983584, "len": "len1", "phi_mod_init": -0.2821632846979857}, {"v": 4, "xy_factor": 0.5518633347961824, "z_factor": -0.7804526126452052, "len": "len1", "phi_mod_init": -0.4224295147918075}, {"v": 4, "xy_factor": 0.2771005006468468, "z_factor": -1.0067042239967168, "len": "len1"}))], 
    60: [(">archimedean_truncated_icosahedron", {"tags": ("#archimedean", "#rigid", "#essential", "#fullerene", "#constrained", "#equilateral", "#equidistant"), "synonyms": (">C60", ">buckminsterfullerene", ">buckyball", ">truncated_icosahedron"), "point_group": "-5m-3", "axis": "-5", "parameter_count": 1}, ({"v": 5, "xy_factor": 0.34327861303195656, "z_factor": 0.9392336204772783, "len": "len1"}, {"v": 5, "xy_factor": 0.6865572260639131, "z_factor": 0.7270757700126065, "len": "len1"}, {"v": 10, "xy_factor": 0.857239485866363, "z_factor": 0.5149179195479349, "len": "len1", "phi_mod_frq": 5, "phi_mod_init": -0.07655323998507184, "phi_init": "s"}, {"v": 10, "xy_factor": 0.9851598593419821, "z_factor": 0.17163930651597828, "len": "len1", "phi_mod_frq": 5, "phi_mod_init": 0.10788603132564961, "phi_init": "s"}, {"v": 10, "xy_factor": 0.9851598593419821, "z_factor": -0.17163930651597828, "len": "len1", "phi_mod_frq": 5, "phi_mod_init": -0.10788603132564961, "phi_init": "s"}, {"v": 10, "xy_factor": 0.857239485866363, "z_factor": -0.5149179195479349, "len": "len1", "phi_mod_frq": 5, "phi_mod_init": 0.07655323998507184, "phi_init": "s"}, {"v": 5, "xy_factor": 0.6865572260639131, "z_factor": -0.7270757700126065, "len": "len1", "phi_init": "s"}, {"v": 5, "xy_factor": 0.34327861303195656, "z_factor": -0.9392336204772783, "len": "len1", "phi_init": "s"}))]}
DICT_INPUT_DEFAULT = {"input_dir": "unknown_directory", "cif_filepaths": [], "central_atom_criteria": {"and": {}, "or": {}, "not": {}}, "ligand_criteria": {"and": {}, "or": {}, "not": {}}, "central_coordinates": [], "ligand_coordinates": [], "poly_criteria": {"and": {}, "or": {}, "not": {}}, "d_min": 0.1, "d_max": 3.5, "cn_max": 60, "find_coordination_spheres": (), "preserve_radial_avg": False, "custom_axis": None, "axis_fixed": False, "var_tax": 3, "constraints_as_symmetry": True, "var_max": inf, "csm_cap": 50, "piggyback_cap": 35, "center_method": "default", "include_central_atom_in_fit": False, "skip_redundant_polyhedra": True, "output_details": True, "output_minicif_cart": True, "output_minicif_fract": False, "output_table": True, "output_log": True, "output_only_best_polyhedra": False, "overwrite_output": True, "automatic_output": False, "show_comments": True}
DICT_ABBREVIATIONS = {"trigonal": "trig", "tetragonal": "tetr", "pentagonal": "pent", "hexagonal": "hex", "elongated": "elong", "equilateral": "equilat", "frank-kasper": "fk", "truncated": "trunc"}
CONFIG_FILENAMES = ("custom_polyhedra", "blocked_polyhedra", "settings")
CONFIGURABLES = ("input_dir", "cif_filepaths", "preserve_radial_avg", "var_tax", "csm_cap", "center_method", "include_central_atom_in_fit", "output_details", "output_minicif_cart", "output_minicif_fract", "output_table", "output_only_best_polyhedra", "overwrite_output", "automatic_output", "show_comments", "piggyback_cap")
DICT_BELT_DEFAULTS = {"xy_factor": 1.0, "z_factor": 1.0, "phi_mod_init": 0.0, "phi_mod_off": 0.5, "phi_mod_frq": 2, "xy_mod_init": 0.0, "xy_mod_off": 0.0, "xy_mod_frq": 2, "z_mod_init": 0.0, "z_mod_off": 0.0, "z_mod_frq": 2}
DICT_SUFFIXES = {"phi_mod": ("_init", "_frq", "_off"), "xy_mod": ("_init", "_frq", "_off"), "z_mod": ("_init", "_frq", "_off")}
ALL_PARAMETER_TYPES = ("phi", "xy", "z", "len", "phi_mod", "xy_mod", "z_mod")
CHARS_FLOAT = "0123456789./+-"
DICT_POINT_GROUP_CONVERSION = {"Oh": "m-3m", "Th": "m-3", "Td": "-43m", "O": "432", "T": "23", "D4h": "4/mmm", "C4h": "4/m", "C4v": "4mm", "D4": "422", "D2d": "-42m", "C4": "4", "S2": "-4", "D6h": "6/mmm", "C6h": "6/m", "C6v": "6mm", "D6": "622", "D3h": "-6m2", "C6": "6", "C3h": "-6", "D3d": "-3m", "C3v": "3m", "D3": "32", "S3": "-3", "C3": "3", "D2h": "mmm", "C2v": "mm2", "D2": "222", "C2h": "2/m", "Cs": "m", "C2": "2", "Ci": "-1", "C1": "1", "Ih": "-5m-3", "I": "523", "D8h": "8/mmm", "D7h": "-[14]m2", "D6d": "-[12]2m", "D4d": "-82m", "D5d": "-[10]m2", "D5h": "-5m", "C[inf]v": "[inf]mm", "D[inf]h": "[inf]/mmm", "m-3m": "Oh", "m-3": "Th", "-43m": "Td", "432": "O", "23": "T", "4/mmm": "D4h", "4/m": "C4h", "4mm": "C4v", "422": "D4", "-42m": "D2d", "4": "C4", "-4": "S2", "6/mmm": "D6h", "6/m": "C6h", "6mm": "C6v", "622": "D6", "-6m2": "D3h", "6": "C6", "-6": "C3h", "-3m": "D3d", "3m": "C3v", "32": "D3", "-3": "S3", "3": "C3", "mmm": "D2h", "mm2": "C2v", "222": "D2", "2/m": "C2h", "m": "Cs", "2": "C2", "-1": "Ci", "1": "C1", "-5m-3": "Ih", "523": "I", "-[14]m2": "D7h", "-[12]2m": "D6d", "-[10]m2": "D5d", "8/mmm": "D8h", "-82m": "D4d", "-5m": "D5h", "5m": "C5v", "[inf]mm": "C[inf]v", "[inf]/mmm": "D[inf]h"}
ATOM_KEYWORDS = ("*M", "*E", "*X", "*Ln", "*An", "*RE", "*Lt", "*Gr1", "*Gr2", "*Gr3", "*Gr4", "*Gr5", "*Gr6", "*Gr7", "*Gr8", "*Gr9", "*Gr10", "*Gr11", "*Gr12", "*Gr13", "*Gr14", "*Gr15", "*Gr16", "*Gr17", "*Gr18", "H", "D", "T", "Li", "Na", "K", "Rb", "Cs", "Fr", "Be", "Mg", "Ca", "Sr", "Ba", "Ra", "B", "Al", "Ga", "In", "Tl", "C", "Si", "Ge", "Sn", "Pb", "N", "P", "As", "Sb", "Bi", "O", "S", "Se", "Te", "Po", "F", "Cl", "Br", "I", "At", "He", "Ne", "Ar", "Kr", "Xe", "Rn", "Sc", "Y", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Ti", "Zr", "Hf", "V", "Nb", "Ta", "Cr", "Mo", "W", "Mn", "Tc", "Re", "Fe", "Ru", "Os", "Co", "Rh", "Ir", "Ni", "Pd", "Pt", "Cu", "Ag", "Au", "Zn", "Cd", "Hg")
POLY_KEYWORDS = ("#antifrustum", "#antiprism", "#archimedean", "#axis-bicapped_antiprism", "#axis-bicapped_prism", "#axis-capped_antifrustum", "#axis-capped_antiprism", "#axis-capped_frustum", "#axis-capped_prism", "#axis_fixed", "#bidisphenoid", "#bipyramid", "#bundled", "#capped_antiprism", "#capped_biprism", "#capped_cube", "#capped_cuboctahedron", "#capped_cupola", "#capped_octahedron", "#capped_platonic_dodecahedron", "#capped_prism", "#capped_tetrahedron", "#capped_truncated_cube", "#catalan", "#chamfered_cube", "#chiral", "#constrained", "#cupola", "#deltahedron", "#diminished_icosahedron", "#distorted_cuboctahedron", "#edshammar", "#elongated_bicupola", "#equator-capped_prism", "#equidistant", "#equilateral", "#essential", "#frank-kasper", "#free", "#frustum", "#fullerene", "#fully_capped_prism", "#gyrobicupola", "#gyroprism", "#heterobipyramid", "#johnson", "#kinked", "#linear", "#molecule", "#orthobicupola", "#planar", "#platonic", "#prism", "#pyramid", "#pyritohedron", "#regular", "#rigid", "#scalenohedron", "#skewed_prism", "#truncated_trapezohedron", "#twisted_prism", "-1", "-3m", "-42m", "-43m", "-5m", "-5m-3", "-6m2", "-82m", "-[10]m2", "-[12]2m", "-[14]m2", "2/m", "222", "32", "3m", "4/mmm", "422", "432", "4mm", "5m", "6/mmm", "6mm", "8/mmm", ">C60", ">F20", ">F24", ">T_shape", ">Z12", ">Z14", ">Z15", ">Z16", ">adamantane_cage", ">anticuboctahedron", ">anticuboid", ">anticuboidal_cuboctahedron", ">archimedean_cuboctahedron", ">archimedean_icosidodecahedron", ">archimedean_rhombicuboctahedron", ">archimedean_snub_cube", ">archimedean_truncated_cube", ">archimedean_truncated_cuboctahedron", ">archimedean_truncated_icosahedron", ">archimedean_truncated_octahedron", ">archimedean_truncated_tetrahedron", ">ax-bicapped_square_prism", ">ax-bicapped_tetragonal_prism", ">ax-bicapped_trigonal_antiprism", ">ax-bicapped_trigonal_prism", ">ax-truncated_hexagonal_trapezohedron", ">bcc-polyhedron", ">biaugmented_digonal_cupola", ">biaugmented_isosceles_wedge", ">bicapped_hexagonal_antiprism", ">bicapped_hexagonal_prism", ">bicapped_pentagonal_antiprism", ">bicapped_pentagonal_prism", ">bicapped_quare_antiprism", ">bicapped_tetragonal_antiprism", ">bidisphenoid", ">bidisphenoidal_cuboctahedron", ">bisphenoid", ">buckminsterfullerene", ">buckyball", ">ccp-polyhedron", ">chamfered_cube", ">cube", ">cubic_11E_edshammar_polyhedron", ">cuboctahedron", ">cuboid", ">deltoidal_icositetrahedron", ">diamond_cage", ">didigonal_scalenohedron", ">digonal_antifrustum", ">digonal_antiprism", ">digonal_cupola", ">digonal_gyrobicupola", ">digonal_orthobicupola", ">digonal_pyramid", ">dihexagonal_scalenohedron", ">dipentagonal_scalenohedron", ">disphenocingulum", ">disphenoid", ">ditetragonal_scalenohedron", ">ditrigonal_scalenohedron", ">dodecadeltahedron", ">dodecahedron", ">elongated_hexagonal_bipyramid", ">elongated_pentagonal_bipyramid", ">elongated_pentagonal_pyramid", ">elongated_square_bipyramid", ">elongated_square_pyramid", ">elongated_tetragonal_bipyramid", ">elpasolite_cuboctahedron", ">eq-hexacapped_hexagonal_prism", ">eq-tetracapped_square_prism", ">eq-tetracapped_tetragonal_prism", ">eq-tricapped_ditrigonal_prism", ">equidistant_kinked_coordination", ">equidistant_linear_coordination", ">equilateral_ax-bicapped_square_prism", ">equilateral_ax-bicapped_tetragonal_prism", ">equilateral_bicapped_pentagonal_prism", ">equilateral_bicapped_square_antiprism", ">equilateral_bicapped_tetragonal_antiprism", ">equilateral_bidisphenoid", ">equilateral_chamfered_cube", ">equilateral_cis-bicapped_cube", ">equilateral_digonal_gyrobicupola", ">equilateral_disphenoid", ">equilateral_elongated_square_gyrobicupola", ">equilateral_elongated_tetragonal_gyrobicupola", ">equilateral_elongated_trigonal_gyrobicupola", ">equilateral_elongated_trigonal_orthobicupola", ">equilateral_eq-bicapped_trigonal_prism", ">equilateral_eq-monocapped_trigonal_prism", ">equilateral_eq-tetracapped_cube", ">equilateral_eq-tetracapped_tetragonal_prism", ">equilateral_fac-tricapped_cube", ">equilateral_friauf_polyhedron", ">equilateral_gyroelongated_trigonal_bicupola", ">equilateral_hexagonal_antiprism", ">equilateral_hexagonal_prism", ">equilateral_mer-tricapped_cube", ">equilateral_monocapped_cube", ">equilateral_monocapped_pentagonal_antiprism", ">equilateral_monocapped_pentagonal_prism", ">equilateral_monocapped_square_antiprism", ">equilateral_monocapped_tetragonal_antiprism", ">equilateral_monocapped_tetragonal_prism", ">equilateral_pentacapped_cube", ">equilateral_pentagonal_antiprism", ">equilateral_pentagonal_orthobirotunda", ">equilateral_pentagonal_prism", ">equilateral_snub_disphenoid", ">equilateral_snub_tetragonal_antiprism", ">equilateral_snub_trigonal_antiprism", ">equilateral_square_antiprism", ">equilateral_square_pyramid", ">equilateral_tetragonal_antiprism", ">equilateral_tetragonal_prism", ">equilateral_tetragonal_pyramid", ">equilateral_trans-bicapped_cube", ">equilateral_triangle", ">equilateral_tricapped_trigonal_prism", ">equilateral_trigonal_bipyramid", ">equilateral_trigonal_cupola", ">equilateral_trigonal_gyrobicupola", ">equilateral_trigonal_prism", ">equilateral_trigonal_pyramid", ">face-sharing_tetrahedron_doublet", ">faceted_pyritohedron", ">faujasite_supercage", ">fully_capped_cube", ">fully_capped_cuboctahedron", ">fully_capped_dodecahedron", ">fully_capped_hexagonal_prism", ">fully_capped_pentagonal_prism", ">fully_capped_platonic_dodecahedron", ">fully_capped_platonic_icosahedron", ">fully_capped_square_prism", ">fully_capped_tetragonal_prism", ">fully_capped_trigonal_prism", ">generic_gyrobifastigium", ">generic_hexakis_truncated_tetrahedron", ">generic_linear_coordination", ">generic_monocapped_trigonal_prism", ">generic_rhombicuboctahedron", ">generic_sodalite_cage", ">generic_tetracapped_friauf_polyhedron", ">generic_tetracapped_truncated_tetrahedron", ">generic_tetrakis_hexahedron", ">generic_triakis_octahedron", ">generic_truncated_cube", ">generic_truncated_octahedron", ">generic_truncated_tetrahedron", ">gyrobifastigium", ">gyroelongated_pentagonal_pyramid", ">gyroelongated_square_bipyramid", ">gyroelongated_tetragonal_pyramid", ">hcp-polyhedron", ">heptacapped_pentagonal_prism", ">heterobisphenoid", ">heterodisphenoid", ">hexacapped_cube", ">hexacapped_cuboctahedron", ">hexacapped_generic_truncated_cube", ">hexacapped_tetragonal_prism", ">hexagonal_antiprism", ">hexagonal_bipyramid", ">hexagonal_heterobipyramid", ">hexagonal_prism", ">hexagonal_pyramid", ">icosahedron", ">icosidodecahedron", ">interpenetrating_truncated_tetrahedra", ">isosceles_triangle", ">isosceles_wedge", ">linear_coordination", ">meta-bidiminished_icosahedron", ">monocapped_digonal_cupola", ">monocapped_hexagonal_antifrustum", ">monocapped_hexagonal_frustum", ">monocapped_isosceles_wedge", ">monocapped_pentagonal_antifrustum", ">monocapped_pentagonal_frustum", ">monocapped_square_antifrustum", ">monocapped_square_frustum", ">monocapped_tetragonal_antifrustum", ">monocapped_tetragonal_frustum", ">monocapped_trigonal_antifrustum", ">monocapped_trigonal_cupola", ">monocapped_trigonal_frustum", ">monodiminished_icosahedron", ">monodiminished_rhombic_dodecahedron", ">octacapped_hexagonal_prism", ">octahedron", ">para-bidiminished_icosahedron", ">parallelepiped", ">parallelogram", ">parallelogrammic_bipyramid", ">parallelogrammic_prism", ">pentacapped_ditrigonal_prism", ">pentacapped_pentagonal_prism", ">pentacapped_trigonal_prism", ">pentagonal_antiprism", ">pentagonal_bipyramid", ">pentagonal_dodecahedron", ">pentagonal_gyrobirotunda", ">pentagonal_heterobipyramid", ">pentagonal_icositetrahedron", ">pentagonal_prism", ">pentagonal_pyramid", ">platonic_dodecahedron", ">platonic_hexahedron", ">platonic_icosahedron", ">platonic_octahedron", ">platonic_tetrahedron", ">pseudohexagonal_elongated_gyrobifastigium", ">psi-1_octahedron", ">psi-1_planar_triangle", ">psi-1_square", ">psi-1_tetrahedron", ">psi-1_trigonal_bipyramid", ">psi-1_trigonal_pyramid", ">psi-2_tetrahedron", ">psi_1-pentagonal_bipyramid", ">pyritohedral_cuboctahedron", ">pyritohedral_icosahedron", ">pyritohedron", ">rectangle", ">rectangular_bipyramid", ">rectangular_gyroprism", ">rectangular_prism", ">rectangular_pyramid", ">regular_heptagon", ">regular_hexagon", ">regular_octagon", ">regular_pentagon", ">regular_triangle", ">rhombic_bipyramid", ">rhombic_dodecahedron", ">rhombic_prism", ">rhombic_pyramid", ">rhombic_triacontahedron", ">rhombohedral_11E_edshammar_polyhedron", ">rhombohedron", ">rhombus", ">rigid_bicapped_hexagonal_antiprism", ">rigid_hexakis_truncated_tetrahedron", ">rigid_tetracapped_friauf_polyhedron", ">rigid_tetracapped_truncated_tetrahedron", ">rigid_tricapped_trigonal_biprism", ">seesaw", ">snub_cube", ">snub_cube_dual", ">snub_cuboctahedron", ">snub_disphenoid", ">snub_square_antiprism", ">sodalite_cage", ">sphalerite_cage", ">sphenocorona", ">square", ">square_antifrustum", ">square_antiprism", ">square_bipyramid", ">square_frustum", ">square_heterobipyramid", ">square_prism", ">square_pyramid", ">strombic_icositetrahedron", ">tetracapped_octahedron", ">tetracapped_tetrahedron", ">tetragonal_antifrustum", ">tetragonal_antiprism", ">tetragonal_bipyramid", ">tetragonal_cuboctahedron", ">tetragonal_frustum", ">tetragonal_heterobipyramid", ">tetragonal_prism", ">tetragonal_pyramid", ">tetrahedral_cuboctahedron", ">tetrahedron", ">tetrakis_cuboctahedron", ">trapezium", ">trapezoid", ">triakis_tetrahedron", ">tricapped_trigonal_biprism", ">tricapped_trigonal_prism", ">tridiminished_icosahedron", ">trigonal_antifrustum", ">trigonal_antiprism", ">trigonal_bipyramid", ">trigonal_cuboctahedron", ">trigonal_cupola", ">trigonal_dodecahedron", ">trigonal_frustum", ">trigonal_gyrobicupola", ">trigonal_heterobipyramid", ">trigonal_icosahedron", ">trigonal_orthobicupola", ">trigonal_prism", ">trigonal_pyramid", ">trigonal_trapezohedron", ">truncated_cuboctahedron", ">truncated_icosahedron", ">twisted_bisphenoid", ">twisted_disphenoid", ">twisted_tetragonal_prism", ">twisted_trigonal_prism", ">unconstrained_anticuboctahedron", ">zeolite_type_a_supercage", "[inf]/mmm", "[inf]mm", "m-3", "m-3m", "mm2", "mmm")
ICON_STRING = "AAABAAEAAAAQAAEABADUCgAAFgAAAIlQTkcNChoKAAAADUlIRFIAAAEAAAABAAgGAAAAXHKoZgAA CptJREFUeJzt3U172jwQhWHxXl2x4f//TDZs8y5Sp8axsSTrY2bOc2/aJC2xxzMHCQi5vZ7px/2R IGZ7/dcf45z3mbnNPgDMxcBf4z0A/pt9AADmIQAAYQQAIIwAAIT9mX0AUbye6Wv5+/3Bg6udfK3+ To0bYAXQwHr4l4+3n8MlXyn9qif1bYAVwAVnQ758nRVBtbMhX75OfSsRABVK790JgmKl9+4EQSW2 AIWuLO3ZFmS5UiPqW4gVQKZWw8tq4FCr4WU1UIAAONHrXpsg+NHrXpsgyMAW4MCoR/LFtwUjzn3v GQT8RQDsGD2Ugk8bzhhKpfpmYwuwMnsIBbYFs4eQbcEGAZDmD/5WwCAwVd9EEPyQ3gJYX3pbPrYC ls9B/vEB2QDwMlzWQ+oDT8Pl5Tibk9sCOB0mT9sCl/VNotsCmQDwOvhbhoMgRH2TWBCE3wI4XkJ/ ZOycLB1LK562MNVCB4CxIWnOQLgpDEno8wu5zIk++EdqtwUV7wwsWd+0mRfv7wicUrAAUB38rY5B QH2TycdfqoU5EYb/XYcQoL4rUULA/Ukw+J81CALq+4H3IHB78Ax+mdJGpb5lvAaBu4OmMevlNik1 ructCEI/DYh3OYPN8GshAABhBADQ0Ovp6zcuEwBAB15CgAAAOvGwGiAAgM4sBwEBAAxiMQgIAGAw SyEw4g1BeF7ZFq5HR/dH8/p2fWFRrwCgyYA21rPUPAxaBwCDD/TT/O3KWgUAgw+M0ywIWjwIyPAD c1yevasBwPADc12awSsBwPADNlTPIq8DAITVBgD3/oAtVTNZEwAMP2BT8WyyBQCEEQCAsNIAYPkP 2FY0o6wAgGBKfuyYAACCygkBAgAI7Gw1UBIA7P8BB/bek+AoCFgBAMIIAEAYAQAIIwAAYQQAIIwA AIQRAIAwAgAQRgAAwggAQBgBAAgjAABhBAAgjAAAhBEAgDACABBGAADCCABAGAEACCMAAGEEACDs z+wDKHV/fP+Z+4sP8G2pW+6/o75lcutrjdsVgNeCj3Z/1NWq9v8p8lwntwGQEk16pkVtqO+xCP3n bguwh2Xru9ZNSX3feR/6tRABsFBv1N6NSX1nH0F7rrcARyJeqE9GL0UjLH1LRT3fkAGQkk6TzjxH lfpGPk/TW4BlqXnlAkRdtlppSup7rEX/9nYr+LfDfjvwXjO1KqL3RrXcTClR30XPHs78/ruzvT0G cwFw1kAtk9kT64O/pVrjEf2beRy+AqC0YZSCwNvwL5TqO6N/P3ETAFeaJPq2wOvgb0Wvr4Ue3jIf AC2bwsKFbCnK4G9Fq6/FHl7kBsCUpwFbN8Lr2eY2LQyehWPoxcK5tVru9+jhGYauAEadpMfHBywM x0ge6+usf+1sAWakm8Vl3h61wd/yUl9vPWwiACzs+awmv/rgb1mtsdcenh4AFgq3ZqkZGP59lurr vX+nBYC1wm3NbA4GP8/M+kbp3+EBYL1wa6P3hQx+ndH1jdTDQwPAU+HWet9jMPht9K5xxP4dEgBe C7fVo4kY/rZ61Ddy/3YNgCiF22JoNSj0b5dXAvZ4BZQlkc8N3yJf45r5zF4BvJ7j3g/AAlYDsUQe /AO/Znuvp08DQG3wtwgC3wQHf+uW0nEfHwaA+uCvEQI+Mfz/3B/7s/7rkwz+MYLABwb/2DYI3j5g +PMQBDYx+HnWIXBLicGvRRDYwODXuT/S7cbwX0MIzMXwXxP2F4MAOEcAAMIIAEAYAQAIIwAAYQQA IIwAAIQRAIAwAgAQRgAAwggAQBgBAAgjAABhBAAgjAAAhBEAgDACABBGAADCCABAGAEACCMAAGEE ACCMAACEEQCAMAIAEEYAAMIIAEAYAQAIIwAAYQQAIIwAAIQRAIAwAgAQRgAAwggAQBgBAAgjAABh BAAgjAAAhBEAgDACABBGAADCCABAGAEACCMAAGG3lFJ6PdPX7APx6P6YfQRIKaXXc/YR+HR/pNtt /QmCIA+DbxNBkOf+SD9zf9t+kRA4xuD7QBAcWw9/SjsBsCAI/mHwfSII/tkO/uJTACykg4Dh9009 BI4Gf7H7xYOiSQUBgx+LWhCcDf5i7zGAoxv8Uigigx+bUA9nBUDR6wDuj9gDEvnc8C3yNa6Zzz+1 3yilOGnaoim2tYjcaDO0rC/9+0/RFmDv856L2GPwW98++tfYaw+fnHv7xwA+3ZCnIrYazNxzJgjq jK5vsB4eGwBn/9+KmfcYBEGemfUN1L9zAuDsdmax1BwEwT5L9Q3Qv3MD4NNtjWS1GQiBd1Zr7LiH 5wfA2W325GVfqB4EXurrsIftBMDZbbfmMfnVgsBjfT31b7IYAGff4ypPF/lI9CCIUF/L/bvS/pWA rbRu8lavUJzdnFaOoRcL59biGHq8InZW8E9ZAeR8vxye93c5oqwGotfXQg/vsLsF2FNaxCj3+Dm8 BoFSfWf07wlfAXD2/VfH0f17WOQtBFRrPKJ/M/kMgKNjsLBcs8B6EFDfbz17OJPfAFgsx6J6r/+J tSCgvr+17N8K/gOghWiNuTU7CKivWXafBhzh9YzfnCnNPUeV+kY+z5ABEPmC7RndpNGHYk/U8616 RyCrol6kXL33nNT3+0/H24JfQgSAemNutW5U6vsuUhC43gIoLkVLtKgN9T0Wof/cBoD3wo9S26QR mnsUz3VytwXwXOyZcpet1LeO122B2xUAgOsIAEAYAQAIIwAAYQQAIIwAAIQRAIAwAgAQRgAAwggA QBgBAAgjAABhBAAgjAAAhBEAgDACABBGAADCCABAGAEACCMAAGEEACDsVwDcH/vvbPp65v2yQQDT Zc/q4QrgKAgAxHG6BSAEgLiyfjEIIQDEVPogII8DALYVzSjPAgDCCABAWE0AsA0AbCqezdoVACEA 2FI1k2wBAGFXAoBVAGBD9SxeXQEQAsBcl2awxRaAEADmuDx7Wa8EzLAcyFej2wNwrNmdbqsAWBAE QD/NV9utA2CxPlDCAKjXdYvdKwDWsk7g9cy+PQLlmrPrQX2vcfWYmJnXAfATh8B4ZgIgJd6EBBjN VAAsCAJgDJMBsCAIgL5MB8CCEAD6cBEAKbEaAHpwEwAA2iMAhNwf589R5/wbxOH2Yr+evGClROlg U98yXoPT5UGv0ainbinVP35CfT/zOvgL1we/RqP+snttCYI2vA/+IsRJLGjSHx+vKyFwTZThTylI AOz8IJFqoxZdT4KgTKTBX7g/oZOfIlRp1OrreOW1FSpBEHHwF9GfBgx74f66pYnneH+kW+ThSCn2 8KcUPwBSmjwkHZk5p4hDohBuKY15QxArorxdmcmmXIbF+7ZAYejXlAJg4TUIXDSm1yBQG/yFwhbg iJcL7nIL42kJ7eU4e1AOgJTsD5flY8tiebg8hVQviluAPda2BaGa0tq2QH3o1wiAd7ODIHRjzg4C Bv839S3AkdGNYn0r0tSMpTfDv48AODZqKGUbc8RQss//jC3AuV7bApoy9dsWMPR5WAHka9VQUsv9 XC3vqRn+fARAmavDS2OeuDK8LPfLsQWoU7otoCkLlG4LGPp6BMA1Z0FAY15wFgQM/nVsAdrYNiL7 /Ib2lvYMfxusANqhITtj6NtjBQAIIwAAYQQAIIwAAIQRAICw/wHbG07WxXbspgAAAABJRU5ErkJg gg==" # to change the icon, acquire a .ico file, run it through a base64 encoder (e. g. https://www.motobit.com/util/base64-decoder-encoder.asp), then paste the generated string here.

# General utility classes:

class Vec: # Fully immutable. This is reinventing the wheel, but grandfathered in at this point. I was unaware of numpy and other modules when I started writing it.
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
    
    def __str__(self):
        return f"Vec({round(self.x, 4)}, {round(self.y, 4)}, {round(self.z, 4)})"
    
    def __bool__(self):
        return self.x != 0 or self.y != 0 or self.z != 0
    
    def __eq__(self, other_vec):
        return self.x == other_vec.x and self.y == other_vec.y and self.z == other_vec.z
    
    def __add__(self, other_vec):
        return Vec(self.x + other_vec.x, self.y + other_vec.y, self.z + other_vec.z)
    
    def __sub__(self, other_vec):
        return Vec(self.x - other_vec.x, self.y - other_vec.y, self.z - other_vec.z)
    
    def __round__(self, decimal_places):
        return Vec(round(self.x, decimal_places), round(self.y, decimal_places), round(self.z, decimal_places))
    
    def dot(self, vec_2):
        return self.x * vec_2.x + self.y * vec_2.y + self.z * vec_2.z
    
    def cross(self, vec_2):
        return Vec(vec_2.y * self.z - vec_2.z * self.y, vec_2.z * self.x - vec_2.x * self.z, vec_2.x * self.y - vec_2.y * self.x)
    
    def scale(self, factor):
        return Vec(factor * self.x, factor * self.y, factor * self.z)
    
    def scale_along_axis(self, factor, axis_unitvec):
        z_comp = axis_unitvec.scale(self.dot(axis_unitvec))
        return self - z_comp + z_comp.scale(factor)
    
    def length(self):
        return sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)
    
    def squared_length(self):
        return self.x ** 2 + self.y ** 2 + self.z ** 2

    def unit(self):
        length = sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)
        if not length:
            raise ZeroDivisionError
        return Vec(self.x / length, self.y / length, self.z / length)

    def distance_points(self, point_2):
        return sqrt((point_2.x - self.x) ** 2 + (point_2.y - self.y) ** 2 + (point_2.z - self.z) ** 2)
    
    def distance_points_squared(self, point_2):
        return (point_2.x - self.x) ** 2 + (point_2.y - self.y) ** 2 + (point_2.z - self.z) ** 2

    def distance_point_from_axis(self, axis_through_origin):
        axis_unit = axis_through_origin.unit()
        axis_scaled = axis_unit.scale(self.dot(axis_unit))
        return axis_scaled.distance_points(self)
    
    def distance_point_from_plane(self, point_in_plane, normal_unit_vec_plane):
        return abs(normal_unit_vec_plane.dot(self) - normal_unit_vec_plane.dot(point_in_plane))

    def angle_vecs(self, vec_2): #hinge = origin
        d_12 = sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)
        d_13 = sqrt(vec_2.x ** 2 + vec_2.y ** 2 + vec_2.z ** 2)
        try:
            return acos((self.x * vec_2.x + self.y * vec_2.y + self.z * vec_2.z) / (d_12 * d_13))
        except ZeroDivisionError:
            return 0
        except ValueError:
            if (self.x * vec_2.x + self.y * vec_2.y + self.z * vec_2.z) / (d_12 * d_13) > 0:
                return 0
            else: # if (self.x * vec_2.x + self.y * vec_2.y + self.z * vec_2.z) / (d_12 * d_13) < 0:
                return pi

    def dihedral_angle(self, major_axis_unitvec, minor_axis_unitvec):
        third_base = major_axis_unitvec.cross(minor_axis_unitvec)
        dotprod_minor = self.dot(minor_axis_unitvec)
        dotprod_third = self.dot(third_base)
        if dotprod_minor == dotprod_third == 0:
            return 0
        elif dotprod_third >= 0:
            return acos(dotprod_minor / sqrt(dotprod_minor ** 2 + dotprod_third ** 2))
        else:
            return 2*pi - acos(dotprod_minor / sqrt(dotprod_minor ** 2 + dotprod_third ** 2))

    def rotate(self, rotation_axis_unit, rotation_angle):
        vec_rotated = self.scale(cos(rotation_angle)) + rotation_axis_unit.cross(self).scale(sin(-rotation_angle)) + rotation_axis_unit.scale((1-cos(rotation_angle))*rotation_axis_unit.dot(self))
        return vec_rotated

    def rotate_x(self, alpha):
        return Vec(self.x, self.y * cos(alpha) - self.z * sin(alpha), self.y * sin(alpha) + self.z * cos(alpha))

    def rotate_y(self, beta):
        return Vec(self.x * cos(beta) - self.z * sin(beta), self.y, self.x * sin(beta) + self.z * cos(beta))

    def rotate_z(self, gamma):
        return Vec(self.x * cos(gamma) - self.y * sin(gamma), self.x * sin(gamma) + self.y * cos(gamma), self.z)
    
    def multiply_matrix(self, three_by_three_matrix):
        x = three_by_three_matrix[0][0] * self.x + three_by_three_matrix[0][1] * self.y + three_by_three_matrix[0][2] * self.z
        y = three_by_three_matrix[1][0] * self.x + three_by_three_matrix[1][1] * self.y + three_by_three_matrix[1][2] * self.z
        z = three_by_three_matrix[2][0] * self.x + three_by_three_matrix[2][1] * self.y + three_by_three_matrix[2][2] * self.z
        return Vec(x, y, z)

    def coordinate_transformation(self, transformation_matrix): # optimized matrix multiplication for transformations from and to fractional coordinates.
        x = transformation_matrix[0][0] * self.x + transformation_matrix[0][1] * self.y + transformation_matrix[0][2] * self.z
        y = transformation_matrix[1][1] * self.y + transformation_matrix[1][2] * self.z
        z = transformation_matrix[2][2] * self.z
        return Vec(x, y, z)
    
    def copy(self):
        return Vec(self.x, self.y, self.z)

class TreeNode: # This data structure is used for keeping track of symmetry relations, typically along with a dictionary to access every node cheaply. Tree Nodes may have any number of children AND any number of parents (though both are typically few.). Things get messy...
    def __init__(self, content, name, list_children):
        self.content = content
        self.name = name
        self.children = list_children
    
    def get_branches(self):
        list_branches = [self]
        if self.children:
            for item_child in self.children:
                twigs = item_child.get_branches()
                list_branches += twigs
        return list_branches
    
    def get_branch_depth_tuples(self, depth = 0):
        list_branches = [(self, depth)]
        if self.children:
            for item_child in self.children:
                twigs = item_child.get_branch_depth_tuples(depth = depth + 1)
                list_branches += twigs
        return list_branches

    def get_branches_sorted(self, depth = 0):
        list_branches = [(self, depth)]
        if self.children:
            self.children.sort(key = lambda j: min([i.content.csm for i in j.get_branches()]))
            for item_child in self.children:
                twigs = item_child.get_branches_sorted(depth = depth + 1)
                list_branches += twigs
        return list_branches
    
    def copy(self):
        return TreeNode(self.content, self.name, self.children)

# GUI utility classes:

class Buttoon:
    def __init__(self, parent, text = "text", command = lambda: print("this does nothing."), width = 6, pady = 2):
        self.frame = tk.Frame(parent, bg = "#a0a0a0", padx = 1, pady = 1)
        self.butt = tk.Button(self.frame, text = text, command = command, width = width, relief = tk.FLAT, pady = pady, bg = "#e0e0e0", bd = 0)
        self.frame.columnconfigure(0, weight = 1), self.frame.rowconfigure(0, weight = 1)
        self.butt.grid(sticky = "nsew")

class CritSearch:
    def __init__(self, parent, frame, keywords, dict_search_init, crit_type = "trash", height = 7, pad_top = 2):
        self.LOGIC_TYPES = ("and", "or", "not")
        LOGIC_COMMENTS = ("Every criterion marked with 'and' needs to be met for a given candidate item to be valid.", "If there are criteria marked with 'or', at least one of them needs to be met for a given candidate item to be valid.", "A candidate item is invalid if it meets a criterion marked with 'not'.")
        self.crit_type = crit_type
        self.parent = parent
        self.frame = frame
        self.logic = "or"
        self.buttons = []
        self.keywords = keywords
        self.dict_search_init = dict_search_init
        self.current_entry = tk.StringVar()
        self.current_entry.trace("w", self.autofill)
        frame_combox = tk.Frame(self.frame)
        self.combox = ttk.Combobox(frame_combox, values = self.keywords, textvariable = self.current_entry, height = 12, width = 32)
        frame_listbox = tk.Frame(self.frame, bg = "#a0a0a0", padx = 1, pady = 1)
        self.listbox = tk.Listbox(frame_listbox, height = height, width = 30, highlightthickness = 0, selectmode = "extended", relief = tk.FLAT)
        frame_top = tk.Frame(self.frame, bg = "#ffffff", padx = 1, pady = 1)
        for count, item_logic in enumerate(self.LOGIC_TYPES):
            button = Buttoon(frame_top, text = item_logic, width = 9, command = lambda i = item_logic: self.change_logic(i))
            if item_logic == self.logic:
                button.butt.configure(bg = "#aaccff")
            button.frame.grid(row = 0, column = count, padx = (4 if count != 0 else 3, 4), pady = (0, 5), sticky = "nsew")
            give_hint(button.butt, self.parent, LOGIC_COMMENTS[count])
            self.buttons.append(button)
        frame_top.grid(row = 0, column = 0, pady = (pad_top, 0), sticky = "nsew"), frame_top.columnconfigure((0, 1, 2), weight = 1)
        self.combox.bind("<Return>", lambda event: self.add_search_term(self.logic))
        self.combox.grid(row = 0, column = 0, sticky = "nsew")
        self.listbox.grid(row = 0, column = 0, sticky = "nsew")
        frame_combox.grid(row = 1, column = 0, columnspan = 3, sticky = "nsew", padx = 4)
        frame_listbox.grid(row = 2, column = 0, columnspan = 3, padx = 4, pady = 5, sticky = "nsew")
        frame_combox.columnconfigure(0, weight = 1), frame_listbox.columnconfigure(0, weight = 1), frame_listbox.rowconfigure(0, weight = 1)
        self.frame.rowconfigure(2, weight = 1), self.frame.columnconfigure(0, weight = 1)
    
    def delete_search_term(self):
        index_tuple = self.listbox.curselection()[::-1]
        if self.listbox.size() and not index_tuple:
            self.listbox.delete(tk.END)
        for item_index in index_tuple:
            self.listbox.delete(item_index)
    
    def autofill(self, *args):
        dynamic_values = [i for i in self.keywords if self.current_entry.get() in i]
        self.combox.config(values = dynamic_values)
    
    def add_search_term(self, logic_type):
        entry_string = self.current_entry.get().replace("'", "").replace(",", " ")
        if self.crit_type == "poly" and entry_string.replace(" ", "").startswith("*"):
            try:
                var_max_string = str(int(entry_string.replace(" ", "")[1:]))
                self.listbox.insert(tk.END, f"variable max = {var_max_string}")
                return
            except ValueError:
                pass
        entry_split = entry_string.split()
        if len(entry_split) == 3 and self.crit_type in ("ligand_coordinates", "central_coordinates"):
            try:
                [parse_math(i.strip("()")) for i in entry_split]
                self.listbox.insert(tk.END, f"add dummy ({entry_split[0].lstrip('(')}, {entry_split[1]}, {entry_split[2].rstrip(')')})")
                return
            except ValueError:
                pass
        for item_entry in entry_split:
            self.listbox.insert(tk.END, f"{logic_type} {item_entry}")
    
    def change_logic(self, logic_type):
        self.logic = logic_type
        for count, item in enumerate(self.buttons):
            if self.LOGIC_TYPES[count] == logic_type:
                item.butt.config(bg = "#aaccff")
            else:
                item.butt.config(bg = "#e0e0e0")
    
    def read(self):
        dict_search = deepcopy(self.dict_search_init)
        tuple_entries = self.listbox.get(0, tk.END)
        for item_entrystring in tuple_entries:
            split_entry = item_entrystring.split(" ")
            if split_entry[0] == "add":
                coordinate_string_list = "".join(split_entry[2:])[1:-1].replace(" ", "").split(",")
                try:
                    floats = [parse_math(i) for i in coordinate_string_list]
                    if len(floats) == 3:
                        dict_search.get(self.crit_type).append(floats)
                except ValueError:
                    pass
            elif split_entry[0] == "variable":
                dict_search["var_max"] = int(split_entry[3])
            else:
                value = split_entry[4][:-1] if len(split_entry) > 2 else None
                dict_search.get(next(iter(dict_search))).get(split_entry[0])[split_entry[1]] = value
        return dict_search

class CheckList:
    def __init__(self, parent, frame, checkstuff, arrangement = "vertical"):
        self.parent = parent
        self.checkstuff = checkstuff
        self.frame = frame
        self.arrangement = arrangement
        self.variables = {}
        for count_check, item_check in enumerate(self.checkstuff):
            box = ttk.Checkbutton(self.frame, text = item_check[1])
            box.state(["!alternate", "selected" if item_check[2] else "!selected"])
            if self.arrangement == "horizontal":
                box.grid(row = 0, column = count_check, padx = 5, sticky = "w")
            else:
                box.grid(row = count_check, column = 0, padx = 5, sticky = "w")
            if len(item_check) > 3:
                give_hint(box, self.parent, item_check[3])
            self.variables[item_check[0]] = box
    
    def read(self):
        dict_readvars = {}
        for item in self.variables:
            dict_readvars[item] = "selected" in self.variables.get(item).state()
        return dict_readvars

class RadioList:
    def __init__(self, parent, frame, dict_radiostuff):
        self.parent = parent
        self.variables = {}
        self.frame = frame
        self.radiostuff = dict_radiostuff
        count_overall = 0
        for item_key in self.radiostuff:
            tk.Label(self.frame, text = item_key[1], bg = "#ffffff").grid(row = count_overall, column = 0, sticky = "w")
            var = tk.StringVar()
            self.variables[item_key[0]] = var
            var.set(self.radiostuff.get(item_key)[0][0])
            for count_tuple, item_tuple in enumerate(self.radiostuff.get(item_key)):
                count_overall += 1
                radbutt = ttk.Radiobutton(self.frame, text = item_tuple[1], variable = var, value = item_tuple[0], style = "TRadiobutton")
                if len(item_tuple) > 2:
                    give_hint(radbutt, self.parent, item_tuple[2])
                radbutt.grid(row = count_overall, column = 0, pady = (0, 2), padx = 5, sticky = "w")
                if count_tuple == 0:
                    radbutt.state(["selected"])
            count_overall += 1
    
    def read(self):
        dict_readvars = {}
        for item in self.variables:
            dict_readvars[item] = self.variables.get(item).get()
        return dict_readvars

class EntryList:
    def __init__(self, parent, frame, entrystuff, arrangement = "parallel", n_columns = 2, fieldwidth = 12):
        # entrystuff = tuple(variable_name, label, initial_text, hint, type)
        self.parent = parent
        self.frame = frame
        self.entrystuff = entrystuff
        self.variables = []
        self.fields = []
        for count_field, item_field in enumerate(self.entrystuff):
            stringvar = tk.StringVar()
            field = ttk.Entry(self.frame, width = fieldwidth, textvariable = stringvar)
            field.insert(0, item_field[2])
            give_hint(field, self.parent, item_field[3])
            if arrangement == "parallel":
                tk.Label(self.frame, text = item_field[1], bg = "#ffffff").grid(row = 2 * (count_field // n_columns), column = count_field % n_columns, sticky = "w", padx = 4)
                field.grid(row = 2 * (count_field // n_columns) + 1, column = count_field % n_columns, sticky = "w", padx = 6, pady = (0, 5))
            else:
                tk.Label(self.frame, text = item_field[1], bg = "#ffffff").grid(row = (2 * count_field) // n_columns, column = (2 * count_field) % n_columns, sticky = "w", padx = 4)
                field.grid(row = (2 * count_field) // n_columns, column = (2 * count_field + 1) % n_columns, sticky = "w", padx = 6, pady = (0, 5))
            self.fields.append(field)
            self.variables.append(stringvar)
    
    def read(self):
        dict_readvars = {}
        for count, item in enumerate(self.entrystuff):
            try:
                dict_readvars[item[0]] = item[4](self.fields[count].get())
            except ValueError:
                pass
        return dict_readvars

class VisualPolyhedron: # Visualization of polyhedra, coordination environments and both combined.
    def __init__(self, parent, names = [], elements = [], belt_dicts = (), dict_bundles = {}, dict_var_range = {}, model_vecs = [], real_vecs = [], excentricity = Vec(0, 0, 0), is_molecule = False, parameter_count = 0, model_labels = [], axis = Vec(0, 0, 1)):
        DICT_ELEMENT_COLORS = {"Cen": "#ffffdd", "Lig": "#ffffdd", "H": "#eeeeff", "D": "#ffeeee", "T": "#eeffee", "Li": "#bbbbbb", "Na": "#aaaaaa", "K": "#aa99aa", "Rb": "#aaaa99", "Cs": "#aaaa88", "Fr": "#999988", "Be": "#bbbbcc", "Mg": "#cccccc", "Ca": "#bbbbbb", "Sr": "#aaaa99", "Ba": "#666655", "Ra": "#555544", "B": "#666622", "C": "#333333", "N": "#77dd55", "O": "#ff4444", "F": "#bbdd88", "Al": "#bbbbcc", "Si": "#22dddd", "P": "#bb2288", "S": "#eedd33", "Cl": "#cceeaa", "Ga": "#779988", "Ge": "#119999", "As": "#557755", "Se": "#bb9955", "Br": "#996644", "In": "#bbbbbb", "Sn": "#aabbbb", "Sb": "#885566", "I": "#9922bb", "Tl": "#888888", "Pb": "#556666", "Bi": "#556655", "Cr": "#77cc77", "Mn": "#993399", "Fe": "#bb9988", "Co": "#9999dd", "Ni": "#77bb99", "Cu": "#eeaa99", "Zn": "#bbbbbb", "At": "#777777", "He": "eeeeee", "Ne": "#eeeeee", "Ar": "#eeeeee", "Kr": "#eeeeee", "Xe": "#eeeeee", "Rn": "#99bb99", "Sc": "#999999", "Y": "#999999", "La": "#999999", "Ce": "#999999", "Pr": "#999999", "Nd": "#999999", "Pm": "#999999", "Sm": "#999999", "Eu": "#999999", "Gd": "#999999", "Tb": "#999999", "Dy": "#999999", "Ho": "#999999", "Er": "#999999", "Tm": "#999999", "Yb": "#999999", "Lu": "#999999", "Ac": "#999999", "Th": "#999999", "Pa": "#999999", "U": "#99cc99", "Np": "#99aaaa", "Pu": "#bb9999", "Am": "#999999", "Cm": "#999999", "Bk": "#999999", "Cf": "#999999", "Es": "#999999", "Fm": "#999999", "Md": "#999999", "No": "#999999", "Lr": "#999999", "Rf": "#999999", "Db": "#999999", "Sg": "#999999", "Bh": "#999999", "Hs": "#999999", "Mt": "#999999", "Ds": "#999999", "Rg": "#999999", "Cn": "#999999", "Ti": "#999999", "Zr": "#999999", "Hf": "#999999", "V": "#999999", "Nb": "#999999", "Ta": "#999999", "Mo": "#999999", "W": "#999999", "Tc": "#999999", "Re": "#999999", "Ru": "#999999", "Os": "#666666", "Rh": "#dddddd", "Ir": "#dddddd", "Pd": "#dddddd", "Pt": "#dddddd", "Ag": "#dddddd", "Au": "#ffdd99", "Cd": "#999999", "Hg": "#dddddd"}
        self.parent = parent
        self.canvas = parent.canvas
        self.canvas.delete()
        self.dict_parameters = {}
        self.drag_start = [-1, -1, -1, -1]
        self.visibles = {"labels": self.parent.visual_preferences.get("labels"), "centers": self.parent.visual_preferences.get("centers")}
        self.rotation_z = 0
        self.pivot = Vec(0, 0, 1)
        self.movements = {}
        self.dimensions = (self.canvas.winfo_width(), self.canvas.winfo_height())
        self.edge_lines, self.atom_circles, self.shape_labels = [], [], []
        self.exists = True
        belt_dicts_fixed = prepare_belt_dicts(belt_dicts, direction_factor = 1)
        self.belt_dicts = deepcopy(belt_dicts_fixed)
        self.dict_bundles = dict_bundles
        self.dict_var_range = dict_var_range
        self.z_asymmetry = False
        for count_belt, item_belt in enumerate(belt_dicts[:len(belt_dicts) // 2]):
            if item_belt.get("z") != belt_dicts[-(count_belt + 1)].get("z"):
                self.z_asymmetry = True
        self.dict_bound_parameters = {}
        self.balance = True
        self.is_molecule = is_molecule
        self.v_total = sum([i.get("v") - len(i.get("erase", ())) for i in self.belt_dicts])
        self.parameter_count = parameter_count
        cross_axis_1 = axis.cross(Vec(2.77, 1.06, 0.32)).unit()
        cross_axis_2 = cross_axis_1.cross(axis).unit()
        self.coord_system_model = (axis, cross_axis_1, cross_axis_2)
        self.coord_system_model_copy = deepcopy(self.coord_system_model)
        if belt_dicts:
            for item_var in self.dict_bundles:
                dict_bundle = self.dict_bundles.get(item_var)
                for item_type in ALL_PARAMETER_TYPES:
                    for item_belt_dict in self.belt_dicts:
                        parameter = item_belt_dict.get(item_type)
                        if parameter in dict_bundle:
                            if not item_type in self.dict_bound_parameters:
                                self.dict_bound_parameters[item_type] = {}
                            self.dict_bound_parameters.get(item_type)[parameter] = self.dict_var_range.get(parameter)
            for item_var in self.dict_bundles:
                dict_bundle = self.dict_bundles.get(item_var)
                for item_type in ALL_PARAMETER_TYPES:
                    for item_belt_dict in self.belt_dicts:
                        if item_belt_dict.get(item_type) in dict_bundle:
                            parameter = item_belt_dict["$" + item_type] = item_belt_dict.pop(item_type) # limit refinement of bound parameters to the adjust_bundles() method.
                            self.dict_bound_parameters.get(item_type)[parameter] = 0
            self.dict_parameters = assemble_initial_parameters(belt_dicts_fixed, self.dict_bundles, self.dict_var_range, self.dict_bound_parameters)
            print("dict_parameters:", self.dict_parameters, "belt_dicts:", self.belt_dicts)
            self.number_of_belts = len(self.belt_dicts)
            self.shape_inits()
        self.belt_dicts_copy = deepcopy(self.belt_dicts)
        self.real = list(real_vecs)
        self.model = list(model_vecs) if model_vecs else [i for j in prototype_polyhedron(self.belt_dicts) for i in j]
        self.labels_real, self.labels_model = names, ["Z"] + list(model_labels) if model_labels else ["Z"] + list(range(1, len(self.model) + 1))
        self.colors = [DICT_ELEMENT_COLORS.get(i) if i in DICT_ELEMENT_COLORS else "#777777" for i in elements]
        for item_type in self.dict_parameters:
            for item_param in self.dict_parameters.get(item_type):
                self.movements[item_param] = [True, 0]
        if self.real:
            self.real_edges, self.real_faces = find_edges(real_vecs)
            self.real.insert(0, excentricity.scale(-1))
            self.visibles["real edges"] = True
        self.real_copy = deepcopy(self.real)
        if self.model:
            if self.real and self.is_molecule:
                self.model.insert(0, Vec(0, 0, 0))
            elif self.real:
                self.model.insert(0, get_centroid(self.model))
            else:
                self.adjust()
            self.model_edges, self.model_faces = ([], None) if self.is_molecule else find_edges(self.model[1:], vec_type = "real" if self.real else "model")
            self.visibles["model edges"] = True
            self.visibles["axis"] = self.parent.visual_preferences.get("axis")
            print(self.model_edges, self.model_faces, self.is_molecule)
        if self.model and self.real:
            self.visibles["real edges"] = self.parent.visual_preferences.get("real edges")
            self.visibles["model edges"] = self.parent.visual_preferences.get("model edges")
            self.visibles[u"\u0394 / pm"] = self.parent.visual_preferences.get(u"\u0394 / pm")
        self.model_copy = deepcopy(self.model)
        self.scale_factor = 0.4 / max([i.length() for i in self.real + self.model])
        if self.belt_dicts:
            self.scale_factor *= 0.72
            if self.dict_parameters.get("xy_mod"):
                self.scale_factor *= 0.8
        self.box_widgets, self.box_parameters = [], []
        self.canvas.bind("<B1-Motion>", self.manual_rotation_xy)
        self.canvas.bind("<Button-1>", self.prepare_rotation_xy)
        self.canvas.bind("<MouseWheel>", self.zoom)
        self.canvas.bind("<ButtonRelease-1>", self.canvas_clicked)
        self.build_boxes()
        self.display()
    
    def shape_inits(self):
        z_number, xy_number = len(self.dict_parameters.get("z")), len(self.dict_parameters.get("xy"))
        full_belt_count = self.number_of_belts - 1
        for count_belt, item_belt_dict in enumerate(self.belt_dicts):
            v, xy = item_belt_dict.get("v"), item_belt_dict.get("xy")
            if xy:
                if self.v_total > 4:
                    self.belt_dicts[count_belt]["xy_factor"] *= (1.25 - self.number_of_belts / self.v_total)
                if xy_number > 1:
                    if 0 < count_belt < full_belt_count:
                        if self.belt_dicts[count_belt + 1].get("xy") != xy != self.belt_dicts[count_belt - 1].get("xy"):
                            self.belt_dicts[count_belt]["xy_factor"] *= sqrt(1.3 - (2 * count_belt / full_belt_count - 1) ** 2)
                    else:
                        self.belt_dicts[count_belt]["xy_factor"] *= 0.8
            if item_belt_dict.get("z"):
                if z_number == 1:
                    self.belt_dicts[count_belt]["z_factor"] *= sqrt(self.number_of_belts / 4)
                elif z_number > 1:
                    if self.z_asymmetry:
                        if count_belt == (full_belt_count) / 2:
                            self.belt_dicts[count_belt]["z_factor"] = -0.2
                        elif item_belt_dict.get("z") == "z*" and count_belt == full_belt_count:
                            z_imbalance = 0
                            for item_nonstar_belt_dict in self.belt_dicts[:-1]:
                                z_imbalance += item_nonstar_belt_dict.get("z_factor") * item_nonstar_belt_dict.get("v")
                            if count_belt > 1:
                                corr_term = (z_imbalance - v) / (self.v_total - v)
                                for item_belt_dict_nested in self.belt_dicts[:-1]:
                                    item_belt_dict_nested["z_factor"] -= corr_term
                                self.belt_dicts[count_belt]["z_factor"] = -1
                            else:
                                self.belt_dicts[count_belt]["z_factor"] = -z_imbalance / v
                        else:
                            self.belt_dicts[count_belt]["z_factor"] = 1 - 2 * count_belt / (full_belt_count)
                    else:
                        self.belt_dicts[count_belt]["z_factor"] = cos(count_belt * pi / (full_belt_count))
                        if v == 1:
                            self.belt_dicts[count_belt]["z_factor"] *= 1.1
            for count_len, item_len in enumerate(self.dict_parameters.get("len")):
                if item_belt_dict.get("len") == item_len:
                    self.belt_dicts[count_belt]["z_factor"] *= 1 - 0.1 * count_len
                    self.belt_dicts[count_belt]["xy_factor"] *= 1 - 0.1 * count_len
    
    def canvas_clicked(self, event):
        for item_box in self.box_parameters:
            if item_box[0] <= event.x <= item_box[2] and item_box[3] <= event.y <= item_box[1]:
                pref_option = item_box[4]
                if pref_option in self.movements:
                    self.movements.get(pref_option)[0] = not self.movements.get(pref_option)[0]
                elif pref_option in self.visibles:
                    self.visibles[pref_option] = not self.visibles.get(pref_option)
                    if (self.model and self.real) or pref_option in ("labels", "centers", "axis"):
                        self.parent.visual_preferences[pref_option] = self.visibles.get(pref_option)
        self.build_boxes()
    
    def zoom(self, event):
        self.scale_factor *= 1.1 ** (event.delta / 120)
    
    def prepare_rotation_xy(self, event):
        self.drag_start[0], self.drag_start[1] = event.x, event.y
    
    def manual_rotation_xy(self, event):
        self.pivot = self.pivot.rotate_x(3 * (event.y - self.drag_start[1]) / self.dimensions[1]).rotate_y(3 * (event.x - self.drag_start[0]) / self.dimensions[0])
        self.rotation_z += Vec(self.pivot.x, self.pivot.y, 0).cross(Vec(3 * (event.x - self.drag_start[0]) / self.dimensions[0], 3 * (event.y - self.drag_start[1]) / self.dimensions[1], 0)).z
        self.drag_start[0], self.drag_start[1] = event.x, event.y
    
    def build_boxes(self):
        for item_box in self.box_widgets:
            self.canvas.delete(item_box)
        self.box_widgets, self.box_parameters = [], []
        height = self.canvas.winfo_height()
        font_size = height // 40 + 1
        char_width = int(font_size * 0.8)
        box_lengths = char_width
        dummy_font = font.Font(size = font_size)
        for item_param in self.movements:
            if not "*" in item_param:
                string_width = 1.05 * dummy_font.measure(item_param)
                self.box_parameters.append((box_lengths, height * 0.98, box_lengths + string_width + 12, height * 0.91, item_param))
                box_lengths += string_width + 18
        box_lengths = char_width
        for item_shape in self.visibles:
            string_width = 1.05 * dummy_font.measure(item_shape)
            self.box_parameters.append((box_lengths, height * 0.09, box_lengths + string_width + 12, height * 0.02, item_shape))
            box_lengths += string_width + 18
        for item_param in self.box_parameters:
            if item_param[4] in self.movements:
                box = self.canvas.create_rectangle(item_param[:4], width = 3, outline = "#ffffff" if self.movements.get(item_param[4])[0] else "#777777")
                text = self.canvas.create_text(item_param[0] + char_width, item_param[1] - char_width + 2, text = item_param[4], fill = "#ffffff" if self.movements.get(item_param[4])[0] else "#777777", font = ("Helvetica", font_size), anchor = "sw")
            elif item_param[4] in self.visibles:
                box = self.canvas.create_rectangle(item_param[:4], width = 3, outline = "#ffffff" if self.visibles.get(item_param[4]) else "#777777")
                text = self.canvas.create_text(item_param[0] + char_width, item_param[1] - char_width + 2, text = item_param[4], fill = "#ffffff" if self.visibles.get(item_param[4]) else "#777777", font = ("Helvetica", font_size), anchor = "sw")
            self.box_widgets.append(box), self.box_widgets.append(text)
    
    def display(self):
        for item_shape in self.edge_lines + self.atom_circles + self.shape_labels:
            self.canvas.delete(item_shape)
        if not self.exists:
            return
        width, height = self.canvas.winfo_width(), self.canvas.winfo_height()
        size_factor = min(width, height)
        if (width, height) != self.dimensions:
            self.dimensions = (width, height)
            self.build_boxes()
        vertex_numbers = tuple(range(len(self.real))) if self.real else tuple(range(len(self.model)))
        vertex_tuples = []
        if self.real:
            vertex_tuples += [("p", self.real[c].z, self.real[c], self.colors[c], self.labels_real[c]) for c in vertex_numbers]
            vertex_tuples += [("e", min(self.real[i[0] + 1].z, self.real[i[1] + 1].z) - 0.00002, (self.real[i[0] + 1], self.real[i[1] + 1]), (self.colors[i[0] + 1], self.colors[i[1] + 1])) for i in self.real_edges]
        if self.model:
            index_model = len(vertex_tuples)
            vertex_tuples += [("p", self.model[c].z - 0.00001, self.model[c], "#4455ff", self.labels_model[c]) for c in vertex_numbers]
            vertex_tuples.append(("a", self.coord_system_model[0].z * 0.01))
            if self.model_edges:
                vertex_tuples += [("m", min(self.model[i[0] + 1].z, self.model[i[1] + 1].z) - 0.00003, (self.model[i[0] + 1], self.model[i[1] + 1])) for i in self.model_edges]
            else:
                vertex_tuples += [("m", min(0, i.z) - 0.00003, (Vec(0, 0, 0), i)) for i in self.model[1:]]
        if self.real and self.model:
            vertex_tuples += [("c", min(self.real[c].z, self.model[c].z) - 0.00004, (self.real[c], self.model[c]), str(round(self.real[c].distance_points(self.model[c]) * 100, 1))) for c in vertex_numbers]
        if not self.visibles.get("centers"):
            if self.real and self.model:
                vertex_tuples.pop(index_model)
            vertex_tuples.pop(0)
        vertex_tuples.sort(key = lambda i: i[1])
        self.edge_lines, self.atom_circles, self.shape_labels = [], [], []
        scale = (self.scale_factor * size_factor)
        atom_upper, atom_lower, atom_right, atom_left = 0.5 * height + 0.025 * size_factor, 0.5 * height - 0.025 * size_factor, 0.5 * width + 0.025 * size_factor, 0.5 * width - 0.025 * size_factor
        label_size, label_dist = 0.025 * width, 0.06 * width
        for item_tuple in vertex_tuples:
            if item_tuple[0] == "p":
                atom_coordinates = (atom_left + item_tuple[2].x * scale, atom_lower + item_tuple[2].y * scale, atom_right + item_tuple[2].x * scale, atom_upper + item_tuple[2].y * scale)
                atom_circle = self.canvas.create_oval(atom_coordinates, fill = item_tuple[3], outline = "")
                self.atom_circles.append(atom_circle)
                if item_tuple[4] and self.visibles.get("labels"):
                    atom_label = self.canvas.create_text(atom_coordinates[0] + label_dist, atom_coordinates[1] + label_size, text = item_tuple[4], fill = item_tuple[3], font = "Helvetica 14 bold", anchor = tk.W)
                    self.shape_labels.append(atom_label)
            elif item_tuple[0] == "m" and self.visibles.get("model edges"):
                edge_coordinates = (0.5 * width + scale * item_tuple[2][0].x, 0.5 * height + scale * item_tuple[2][0].y, 0.5 * width + scale * item_tuple[2][1].x, 0.5 * height + scale * item_tuple[2][1].y)
                edge_line = self.canvas.create_line(edge_coordinates, fill = "#8899ff", width = 6)
                self.edge_lines.append(edge_line)
            elif item_tuple[0] == "e" and self.visibles.get("real edges"):
                x_midpoint = 0.5 * width + scale * (item_tuple[2][0].x + item_tuple[2][1].x) / 2
                y_midpoint = 0.5 * height + scale * (item_tuple[2][0].y + item_tuple[2][1].y) / 2
                edge_coords_1 = (0.5 * width + scale * item_tuple[2][0].x, 0.5 * height + scale * item_tuple[2][0].y, x_midpoint, y_midpoint)
                edge_coords_2 = (x_midpoint, y_midpoint, 0.5 * width + scale * item_tuple[2][1].x, 0.5 * height + scale * item_tuple[2][1].y)
                edge_line_1 = self.canvas.create_line(edge_coords_1, fill = item_tuple[3][0], width = 6)
                edge_line_2 = self.canvas.create_line(edge_coords_2, fill = item_tuple[3][1], width = 6)
                self.edge_lines.append(edge_line_1), self.edge_lines.append(edge_line_2)
            elif item_tuple[0] == "c" and self.visibles.get(u"\u0394 / pm"):
                connection_coordinates = (0.5 * width + scale * item_tuple[2][0].x, 0.5 * height + scale * item_tuple[2][0].y, 0.5 * width + scale * item_tuple[2][1].x, 0.5 * height + scale * item_tuple[2][1].y)
                connection_line = self.canvas.create_line(connection_coordinates, fill = "#8899ff", width = 6, dash = (2, 1))
                self.edge_lines.append(connection_line)
            elif item_tuple[0] == "a" and self.visibles.get("axis"):
                axis = self.coord_system_model[0]
                try:
                    axis_flat_unit = Vec(axis.x, axis.y, 0).unit()
                except ZeroDivisionError:
                    axis_flat_unit = Vec(0, 0, 0)
                shaft_coordinates = (0.5 * width + 0.02 * size_factor * axis_flat_unit.x, 0.5 * height + 0.02 * size_factor * axis_flat_unit.y, 0.5 * width + 0.16 * size_factor * axis.x, 0.5 * height + 0.16 * size_factor * axis.y)
                self.edge_lines.append(self.canvas.create_line(shaft_coordinates, fill = "#cc7777", width = 6))
                arrowhead_base_coordinates = ()
                for item_cross in self.coord_system_model[1:]:
                    tip = (0.5 * width + 0.2 * size_factor * axis.x, 0.5 * height + 0.2 * size_factor * axis.y)
                    barb_1 = (0.5 * width + 0.16 * size_factor * (0.9 * axis.x + 0.14 * item_cross.x), 0.5 * height + 0.16 * size_factor * (0.9 * axis.y + 0.14 * item_cross.y))
                    barb_2 = (0.5 * width + 0.16 * size_factor * (0.9 * axis.x - 0.14 * item_cross.x), 0.5 * height + 0.16 * size_factor * (0.9 * axis.y - 0.14 * item_cross.y))
                    arrowhead_edge = self.canvas.create_polygon(tip + barb_1 + barb_2, fill = "#cc7777")
                    arrowhead_base_coordinates += barb_1 + barb_2
                    self.atom_circles.append(arrowhead_edge)
                arrowhead_base = self.canvas.create_polygon(tuple([arrowhead_base_coordinates[i] for i in (0, 1, 4, 5, 2, 3, 6, 7)]), fill = "#cc7777")
                self.atom_circles.append(arrowhead_base)
        if self.model and self.real:
            for item_tuple in vertex_tuples:
                if item_tuple[0] == "c" and self.visibles.get(u"\u0394 / pm"):
                    x_midpoint = 0.5 * width + scale * (item_tuple[2][0].x + item_tuple[2][1].x) / 2
                    y_midpoint = 0.5 * height + scale * (item_tuple[2][0].y + item_tuple[2][1].y) / 2
                    label_distance = self.canvas.create_text(x_midpoint, y_midpoint, text = item_tuple[3], fill = "#ffffff", font = "Helvetica 12 bold")
                    self.shape_labels.append(label_distance)
        self.canvas.after(30, self.update)
    
    def update(self):
        if self.belt_dicts:
            self.adjust()
        else:
            self.real, self.model = deepcopy(self.real_copy), deepcopy(self.model_copy)
        self.rotate()
        self.display()
    
    def rotate(self):
        try:
            rock_axis = Vec(0, 0, 1).cross(self.pivot).unit()
            angle_rock = -Vec(0, 0, 1).dihedral_angle(rock_axis, self.pivot)
        except ZeroDivisionError:
            rock_axis = Vec(1, 0, 0)
            angle_rock = 0
        for count_vertex in range(len(self.real)):
            self.real[count_vertex] = self.real[count_vertex].rotate(rock_axis, angle_rock).rotate(self.pivot, self.rotation_z)
        for count_vertex in range(len(self.model)):
            self.model[count_vertex] = self.model[count_vertex].rotate(rock_axis, angle_rock).rotate(self.pivot, self.rotation_z)
        self.coord_system_model = tuple([i.rotate(rock_axis, angle_rock).rotate(self.pivot, self.rotation_z) for i in self.coord_system_model_copy])
        
    def adjust(self):
        dict_vars = {"pi": 3.141592653589793}
        if self.dict_bound_parameters:
            for count_var, item_var in enumerate(self.dict_bundles):
                if self.movements.get(item_var)[0] == True:
                    self.movements.get(item_var)[1] += 0.04
                lower, upper = self.dict_var_range.get(item_var)[:2]
                midpoint, wiggle_room = (lower + upper) / 2, (lower - upper) / 2
                if self.dict_parameters.get("len"):
                    dict_vars[item_var] = midpoint - wiggle_room * sin(self.movements.get(item_var)[1] * (count_var + 1))
                else:
                    frequency = 2 * self.movements.get(item_var)[1] * cos(count_var)
                    dict_vars[item_var] = midpoint + 0.3 * wiggle_room * sin(frequency)
        parameter_count = 0
        for item_type in self.dict_parameters:
            if item_type == "parameter_bundles":
                continue
            if not self.dict_parameters.get(item_type):
                continue
            for item_param in self.dict_parameters.get(item_type):
                if item_param in self.dict_bound_parameters.get(item_type, {}):
                    continue
                if self.movements.get(item_param)[0]:
                    self.movements.get(item_param)[1] += 0.8
                frequency = 0.1 * self.movements.get(item_param)[1] * cos(parameter_count)
                factor = 1 + 0.3 * sin(frequency) if self.z_asymmetry else 1 + 0.3 * abs(sin(frequency))
                if item_type == "len":
                    dict_vars[item_param] = factor
                for count_belt, item_belt_dict in enumerate(self.belt_dicts_copy):
                    if item_belt_dict.get(item_type) != item_param:
                        continue
                    if item_type == "xy":
                        self.belt_dicts[count_belt]["xy_factor"] = item_belt_dict.get("xy_factor") * factor
                    elif item_type == "z" and item_belt_dict.get("z") != "z*":
                        if count_belt == (self.number_of_belts - 1) / 2 and not self.z_asymmetry: # affects middle belt of heterobipyramids and the like.
                            self.belt_dicts[count_belt]["z_factor"] = 0.15 * sin(frequency)
                        else:
                            self.belt_dicts[count_belt]["z_factor"] = item_belt_dict.get("z_factor") * factor
                    elif item_type == "len":
                        self.belt_dicts[count_belt]["xy_factor"] = item_belt_dict.get("xy_factor") * factor
                        self.belt_dicts[count_belt]["z_factor"] = item_belt_dict.get("z_factor") * factor
                        if "xy_mod_init" in item_belt_dict:
                            self.belt_dicts[count_belt]["xy_mod_init"] = item_belt_dict.get("xy_mod_init") * factor
                        if "z_mod_init" in item_belt_dict:
                            self.belt_dicts[count_belt]["z_mod_init"] = item_belt_dict.get("z_mod_init") * factor
                    elif item_type.startswith("phi_mod") or item_type.startswith("xy_mod") or item_type.startswith("z_mod"):
                        self.belt_dicts[count_belt][f"{item_type}_init"] = 0.18 * sin(frequency)
                    elif item_type == "phi" and item_belt_dict.get("phi") != "phi*":
                        if self.movements.get(item_param)[1] > 2:
                            self.belt_dicts[count_belt]["phi_init"] = item_belt_dict.get("phi_init") + sin(frequency) * 0.5 * pi
                parameter_count += 1
        if self.z_asymmetry:
            z_imbalance = sum([i.get("z_factor") * (i.get("v") - len(i.get("erase", ()))) for i in self.belt_dicts])
            corr_term = z_imbalance / self.v_total
            for item_belt_dict in self.belt_dicts:
                item_belt_dict["z_factor"] -= corr_term
        for item_var in self.dict_parameters.get("parameter_bundles", {}):
            for item_type in self.dict_bound_parameters:
                for item_param in self.dict_bound_parameters.get(item_type):
                    if item_param not in self.dict_bundles.get(item_var):
                        continue
                    dependency = self.dict_bundles.get(item_var).get(item_param)
                    for count_belt, item_belt_dict in enumerate(self.belt_dicts_copy):
                        if "len" in item_belt_dict:
                            base_length = dict_vars.get(item_belt_dict.get("len"))
                            if item_belt_dict.get("$" + item_type) == item_param:
                                if item_type == "xy":
                                    self.belt_dicts[count_belt]["xy_factor"] = base_length * parse_math(dependency, dict_variables = dict_vars)
                                elif item_type == "z":
                                    z_sign = 1 if item_belt_dict.get("z_factor", 0) > 0 else -1
                                    self.belt_dicts[count_belt]["z_factor"] = z_sign * base_length * parse_math(dependency, dict_variables = dict_vars)
                                elif item_type == "len":
                                    self.belt_dicts[count_belt]["xy_factor"] = base_length * parse_math(dependency, dict_variables = dict_vars)
                                    self.belt_dicts[count_belt]["z_factor"] = base_length * parse_math(dependency, dict_variables = dict_vars)
                                elif item_type.startswith("phi"):
                                    self.belt_dicts[count_belt][item_type + "_init"] = (self.belt_dicts_copy[count_belt].get(item_type + "_init") + parse_math(dependency, dict_variables = dict_vars))
                                elif item_type.startswith("z_mod") or item_type.startswith("xy_mod"):
                                    self.belt_dicts[count_belt][item_type + "_init"] = base_length * (self.belt_dicts_copy[count_belt].get(item_type + "_init") + parse_math(dependency, dict_variables = dict_vars))
                        else:
                            if item_belt_dict.get("$" + item_type) == item_param:
                                if item_type == "xy":
                                    self.belt_dicts[count_belt]["xy_factor"] = parse_math(dependency, dict_variables = dict_vars)
                                elif item_type == "z":
                                    z_sign = 1 if item_belt_dict.get("z_factor", 0) > 0 else -1
                                    self.belt_dicts[count_belt]["z_factor"] = z_sign * parse_math(dependency, dict_variables = dict_vars)
                                elif item_type == "len":
                                    self.belt_dicts[count_belt]["xy_factor"] = parse_math(dependency, dict_variables = dict_vars)
                                    self.belt_dicts[count_belt]["z_factor"] = parse_math(dependency, dict_variables = dict_vars)
                                elif item_type.startswith("phi"):
                                    self.belt_dicts[count_belt][item_type + "_init"] = parse_math(dependency, dict_variables = dict_vars)
                                elif item_type.startswith("z_mod") or item_type.startswith("xy_mod"):
                                    self.belt_dicts[count_belt][item_type + "_init"] = parse_math(dependency, dict_variables = dict_vars)
                        
        prototype = prototype_polyhedron(self.belt_dicts)
        if self.balance:
            self.model = [Vec(0, 0, 0)] + [i for j in prototype for i in j]
        else:
            self.model = [get_centroid([i for j in prototype for i in j])] + [i for j in prototype for i in j]
        self.model_copy = deepcopy(self.model)
    
    def kill(self):
        self.exists = False
        for item_box in self.box_widgets:
            self.canvas.delete(item_box)

# Main GUI classes / windows:

class ConfigWindow:
    def __init__(self, parent):
        self.parent = parent
        self.dict_input = self.parent.dict_input
        self.visual_preferences = self.parent.visual_preferences
        self.menu = tk.Toplevel()
        self.menu.geometry("980x540+260+40")
        CHECKSTUFF_OUT = (("output_table", "data table", self.dict_input.get("output_table"), "If selected, a single .csv table containing data for all evaluated model polyhedra and all coordination environments in all input files will be part of the output. Data include csm, average deltas, excentricity and parameter values."), ("output_log", "log files", self.dict_input.get("output_log"), "If selected, a .log file containig potential problems with the calculations or the data will be generated."), ("output_details", "detailed output files", self.dict_input.get("output_details"), "If selected, a .out ascii file with vectors and precise deviations for each atom, average angular and radial deviations, free parameter values, fit qualities of special projections and more will be generated for each model polyhedron evaluated."), ("output_minicif_cart", "cartesian .cif files", self.dict_input.get("output_minicif_cart"), "If selected, a minimal .cif file for viewing in external programs will be generated for each model polyhedron evaluated. Representations are true to scale and proportions, but not true to the original unit cell in order to allow for simple visualization."), ("output_minicif_fract", "fractional .cif files", self.dict_input.get("output_minicif_fract"), "If selected, a minimal .cif file for viewing in external programs will be generated for each model polyhedron evaluated. Representations are true to the original unit cell, but may make for somewhat cumbersome viewing if equivalent polyhedra intersect."), ("overwrite_output", "overwrite existing output files", self.dict_input.get("overwrite_output"), "If selected, old output files in the same output directory with the same name (including .cif file name, central atom and rounded csm) will be overwritten when a new output is generated. If unselected, an enumerated copy is produced instead."))
        CHECKSTUFF_MISC = (("output_only_best_polyhedra", "output only best fit",  self.dict_input.get("output_only_best_polyhedra"), "Show and generate output files for only the best fitting polyhedron of each coordination environment."), ("automatic_output", "generate output files automatically", self.dict_input.get("automatic_output"), "Automatically generate output files for every run."), ("show_comments", "show comments", self.dict_input.get("show_comments"), "Show comments at the bottom of windows (such as this one)."))
        FRAME_PROPERTIES = (("output files:", 0, 0, 1, 1, {}, {0: 1}), ("miscellaneous:", 0, 1, 1, 1, {0: 1}, {0: 1}), ("active polyhedra:", 1, 0, 1, 2, {(1,): 1}, {(0, 1): 1}), ("viewer:", 0, 2, 2, 1, {0: 1}, {0: 1})) # label, row, column, rowspan, columnspan, rowconfigure weights, columnconfigure weights
        ENTRY_TUPLES_MISC = (("var_tax", "variable tax:", str(self.dict_input.get("var_tax")), "To counterbalance overfitting, an internal parameter is constructed by adding this 'tax' times the number of variables (degrees of freedom) a given polyhedron has to its csm value. The tax is not reflected in the csm given in the results box, nor in the output files. However, the taxed value compares to the csm cutoff value, determines the best polyhedron if 'output only best fitting polyhedra' has been selected and influences the greyscale applied in the results box.", float), ("csm_cap", "csm cutoff:", str(self.dict_input.get("csm_cap")), "Polyhedra with a taxed csm higher than this number will be discarded.", float), ("piggyback_cap", "piggy cap:", self.dict_input.get("piggyback_cap"), "Most derivative polyhedra can inherit their belt assignment from their aristohedra. This almost always enhances performance and improves results in some special cases. If the aristohedron is a bad match, it can also negatively affect the obtained orientation and the resulting csm. This sets a maximal value for the csm of the respective aristohedron. Set to 0 to disallow piggybacking entirely (not recommended).", float)) # entrystuff = tuple(variable_name, label, initial_text, hint, type)
        self.menu.title("Settings")
        self.menu.configure(background = "#ffffff")
        try:
            self.menu.iconbitmap(self.parent.icon)
        except tk.TclError:
            print("icon file unavailable.")
        self.list_blueprints = []
        self.all_blueprints = []
        self.selected_in_listbox = []
        self.deleted_custom_polyhedra = []
        self.excluded_polyhedra = set()
        self.visual_poly = None
        frames = []
        for item_frame in FRAME_PROPERTIES:
            frame = tk.LabelFrame(self.menu, text = item_frame[0], bg = "#ffffff")
            for item_dict_entry in item_frame[5]:
                frame.rowconfigure(item_dict_entry, weight = item_frame[5].get(item_dict_entry))
            for item_dict_entry in item_frame[6]:
                frame.columnconfigure(item_dict_entry, weight = item_frame[6].get(item_dict_entry))
            frames.append(frame)
        subframe_checklist_out = tk.Frame(frames[0], bg = "#ffffff", padx = 4, pady = 2)
        self.check_out = CheckList(self, subframe_checklist_out, CHECKSTUFF_OUT)
        subframe_checklist_out.grid(row = 0, column = 0, pady = (4, 0), sticky = "nw")
        subframe_checklist_misc = tk.Frame(frames[1], bg = "#ffffff", padx = 4, pady = 2)
        self.check_misc = CheckList(self, subframe_checklist_misc, CHECKSTUFF_MISC)
        subframe_checklist_misc.grid(row = 0, column = 0, pady = (4, 0), sticky = "nw")
        subframe_entries_misc = tk.Frame(frames[1], bg = "#ffffff", padx = 4, pady = 2)
        self.entries_misc = EntryList(self, subframe_entries_misc, ENTRY_TUPLES_MISC, fieldwidth = 10, n_columns = 3)
        subframe_entries_misc.grid(row = 1, column = 0, pady = 4, sticky = "nw")
        tk.Label(frames[2], text = "  inclusion criteria:", bg = "#ffffff", anchor = "nw").grid(row = 0, column = 1, sticky = "nw")
        subframe_crit_poly = tk.Frame(frames[2], bg = "#ffffff", padx = 4, pady = 2)
        self.crit_poly = CritSearch(self, subframe_crit_poly, POLY_KEYWORDS, {"inclusion_criteria": {"and": {}, "or": {}, "not": {}}, "var_max": 100}, height = 7, pad_top = 0, crit_type = "poly")
        give_hint(self.crit_poly.combox, self, "Enter criteria for model polyhedra useful to you. Press 'confirm' to review your selection in the list to the left. Leave empty and confirm to reset. Enter '*n', where n is any integer, to cap the number of degrees of freedom at n. Potentially useful entries: 1) '#free' for polyhedra which have all the degrees of freedom allowed by their point group. 2) '#essential' for a (subjective) selection of polyhedra commonly encountered in inorganic chemistry.")
        frame_bottom = tk.Frame(subframe_crit_poly, bg = "#ffffff", padx = 1, pady = 1)
        button_delete = Buttoon(frame_bottom, text = "remove", command = self.crit_poly.delete_search_term, width = 7)
        button_delete.frame.grid(row = 0, column = 0, ipady = 1, padx = 3, sticky = "w")
        button_apply = Buttoon(frame_bottom, text = "confirm", command = self.apply_poly_checks, width = 12)
        button_apply.frame.grid(row = 0, column = 1, ipady = 1, padx = 7, sticky = "w")
        frame_bottom.grid(row = 3, column = 0, sticky = "nsew")
        subframe_crit_poly.grid(row = 1, column = 1, pady = 2, sticky = "nsew")
        subframe_listbox_polyhedra = tk.Frame(frames[2], bg = "#a0a0a0", padx = 1, pady = 1)
        self.listbox_polyhedra = tk.Listbox(subframe_listbox_polyhedra, height = 15, width = 36, highlightthickness = 0, selectmode = "multiple", relief = tk.FLAT)
        self.listbox_polyhedra.grid(sticky = "nsew")
        give_hint(self.listbox_polyhedra, self, "Upon save & close, greyed out polyhedra will not appear in future calculations, even if Polynator is restarted. However, this change is fully reversible.")
        subframe_listbox_polyhedra.rowconfigure(0, weight = 1), subframe_listbox_polyhedra.columnconfigure(0, weight = 1)
        subframe_listbox_polyhedra.grid(row = 0, column = 0, rowspan = 3, padx = (4, 1), pady = 4, sticky = "nsew")
        self.fill_listbox_polyhedra()
        subframe_buttons = tk.Frame(frames[2], bg = "#ffffff")
        toggle_selection_button = Buttoon(subframe_buttons, width = 7, text = "flip selection", command = self.toggle_polyhedra, pady = 4)
        give_hint(toggle_selection_button.butt, self, "From among the polyhedra you manually selected, activate the inactive ones and vice versa.")
        add_custom_button = Buttoon(subframe_buttons, width = 15, text = "add custom polyhedron", command = self.parent.build_custom_polyhedron, pady = 4)
        give_hint(add_custom_button.butt, self, "Construct your own model polyhedra by stacking belts of vertices and defining a set of deformation parameters")
        delete_custom_button = Buttoon(subframe_buttons, width = 22, text = "delete selected custom polyhedra", command = self.delete_custom_polyhedra, pady = 4)
        give_hint(delete_custom_button.butt, self, "Delete selected custom polyhedra irreversibly.")
        poly_buttons = (toggle_selection_button, delete_custom_button, add_custom_button)
        for count_button, item_button in enumerate(poly_buttons):
            item_button.frame.grid(row = 0, column = count_button, padx = (4, 6), pady = 5, sticky = "nsew")
        subframe_buttons.columnconfigure((0, 1, 2), weight = 1), subframe_buttons.grid(row = 3, column = 0, columnspan = 2, sticky = "sew")
        self.canvas = tk.Canvas(frames[3], highlightthickness = 0, bg = "#000000", height = 480, width = 480)
        give_hint(self.canvas, self, "Click on one of the polyhedron names in the list of active polyhedra to visualize it here. Drag with your mouse to rotate the polyhedron and scroll to zoom in or out. Click on the boxes to hide parts of the image or to freeze the movements associated with the respective parameters.")
        self.canvas.grid(padx = 5, pady = 5, sticky = "nsew")
        self.listbox_polyhedra.bind("<ButtonRelease-1>", lambda event: self.visualize_polyhedron(event))
        frame_bottom = tk.Frame(self.menu, bg = "#ffffff")
        self.commentline = tk.Label(frame_bottom, justify = tk.LEFT, wraplength = 800, height = 3, width = 20, bg = "#ffffff", anchor = "nw")
        self.commentline.grid(row = 0, column = 0, padx = 8, sticky = "nsew")
        button_confirm = Buttoon(frame_bottom, text = "save & close", command = self.save_and_close, width = 20, pady = 6)
        button_confirm.frame.grid(row = 0, column = 1, padx = 5, pady = (2, 6), sticky = "nsew")
        frame_bottom.grid(row = 2, column = 0, columnspan = 3, sticky = "nsew")
        frame_bottom.columnconfigure(0, weight = 1)
        for count_frame, item_frame in enumerate(frames):
            properties = FRAME_PROPERTIES[count_frame]
            for item_dict_entry in properties[5]:
                item_frame.rowconfigure(item_dict_entry, weight = properties[5].get(item_dict_entry))
            for item_dict_entry in properties[6]:
                item_frame.columnconfigure(item_dict_entry, weight = properties[6].get(item_dict_entry))
            item_frame.grid(row = properties[1], column = properties[2], rowspan = properties[3], columnspan = properties[4], padx = 5, pady = 5, ipadx = 5, ipady = 5, sticky = "nsew")
        self.menu.columnconfigure((0, 1, 2), weight = 1), self.menu.rowconfigure(0, weight = 1), self.menu.rowconfigure(1, weight = 5)
    
    def fill_listbox_polyhedra(self):
        self.listbox_polyhedra.delete(0, tk.END)
        self.list_blueprints, self.all_polyhedron_names, self.selected_in_listbox = [], [], []
        for item_cn in self.parent.dict_blueprints_by_cn:
            self.listbox_polyhedra.insert(tk.END, f"CN = {item_cn}:")
            self.list_blueprints.append(())
            for item_poly in self.parent.dict_blueprints_by_cn.get(item_cn):
                if item_poly[0] in self.deleted_custom_polyhedra:
                    continue
                self.listbox_polyhedra.insert(tk.END, item_poly[0])
                self.list_blueprints.append(item_poly)
                self.all_blueprints.append(item_poly)
                if item_poly[0] in self.parent.blocked_polyhedra:
                    self.listbox_polyhedra.itemconfigure(tk.END, foreground = "#bbbbbb")
    
    def apply_poly_checks(self):
        self.excluded_polyhedra = set()
        dict_poly_criteria = self.crit_poly.read()
        included_polyhedra = set([i[0] for i in poly_search(self.all_blueprints, dict_poly_criteria.get("inclusion_criteria"), dict_poly_criteria.get("var_max"))])
        self.excluded_polyhedra = set([i[0] for i in self.all_blueprints]) - included_polyhedra
        for count_line, item_line in enumerate(self.listbox_polyhedra.get(0, tk.END)):
            if item_line in self.excluded_polyhedra:
                self.listbox_polyhedra.itemconfigure(count_line, foreground = "#bbbbbb")
            else:
                self.listbox_polyhedra.itemconfigure(count_line, foreground = "#000000")
    
    def toggle_polyhedra(self):
        for item_index in self.selected_in_listbox:
            poly_name = self.listbox_polyhedra.get(item_index)
            if poly_name[0] != ">":
                continue
            if poly_name in self.excluded_polyhedra:
                self.excluded_polyhedra.discard(poly_name)
                self.listbox_polyhedra.itemconfigure(item_index, foreground = "#000000")
            else:
                self.excluded_polyhedra.add(poly_name)
                self.listbox_polyhedra.itemconfigure(item_index, foreground = "#bbbbbb")
    
    def delete_custom_polyhedra(self):
        deletion_candidates = [self.list_blueprints[i][0] for i in self.selected_in_listbox]
        for item_name in deletion_candidates:
            if item_name in self.parent.dict_custom_blueprints:
                self.deleted_custom_polyhedra.append(item_name)
        self.fill_listbox_polyhedra()
        self.commentline.configure(text = "Selected polyhedra will be permanently removed upon exiting this window via save & close.")
    
    def save_and_close(self):
        for item_widget in (self.entries_misc, self.check_misc, self.check_out):
            self.parent.dict_input.update(item_widget.read())
        self.parent.blocked_polyhedra = tuple(self.excluded_polyhedra)
        for item_cn in self.parent.dict_blueprints_by_cn:
            for item_blueprint in self.parent.dict_blueprints_by_cn.get(item_cn):
                if item_blueprint[0] in self.deleted_custom_polyhedra:
                    self.parent.dict_blueprints_by_cn.get(item_cn).remove(item_blueprint)
        for item_name in self.deleted_custom_polyhedra:
            try:
                self.parent.dict_custom_blueprints.pop(item_name)
                self.parent.dict_blueprints_by_name.pop(item_name)
            except KeyError:
                print(self.parent.dict_custom_blueprints, self.parent.dict_blueprints_by_name, self.parent.dict_blueprints_by_cn)
                pass
        self.menu.destroy()
    
    def visualize_polyhedron(self, event):
        if self.visual_poly:
            self.visual_poly.kill()
        new_selection = self.listbox_polyhedra.curselection()
        index_new = next(iter(set(new_selection) ^ set(self.selected_in_listbox)))
        blueprint = self.list_blueprints[index_new]
        self.selected_in_listbox = new_selection
        if not blueprint:
            return
        self.commentline.configure(text = get_poly_comment(blueprint), foreground = "#000000")
        self.visual_poly = VisualPolyhedron(self, belt_dicts = blueprint[2], is_molecule = "#molecule" in blueprint[1].get("tags"), dict_bundles = blueprint[1].get("parameter_bundles", {}), dict_var_range = blueprint[1].get("var_range", {}))
        

class CustomPolyhedronWindow:
    def __init__(self, parent, blueprint = ()):
        self.parent = parent
        self.dict_input = self.parent.dict_input
        self.visual_preferences = parent.visual_preferences
        self.menu = tk.Toplevel()
        self.menu.geometry("1000x744+240+40")
        self.menu.title("Construct Custom Polyhedron")
        try:
            self.menu.iconbitmap(self.parent.icon)
        except tk.TclError:
            pass
        self.menu.configure(background = "#ffffff")
        self.BELT_TUPLES = (("v", "v (polygon order):", None, int), ("erase", "delete vertices:", None, None), ("phi", "phi parameter:", None, str), ("phi_init", "phi_init:", 0.0, None), ("xy", "xy parameter:", None, str), ("z", "z parameter:", None, str), ("xy_factor", "xy_factor:", 1.0, float), ("z_factor", "z_factor:", 1.0, float), ("phi_mod", "phi_mod parameter:", None, str), ("phi_mod_init", "phi_mod_init:", 0.0, float), ("phi_mod_frq", "phi_mod_frq:", 2, int), ("phi_mod_off", "phi_mod_off:", 0.5, float), ("xy_mod", "xy_mod parameter:", None, str), ("xy_mod_init", "xy_mod_init:", 0.0, float), ("xy_mod_frq", "xy_mod_frq:", 2, int), ("xy_mod_off", "xy_mod_off:", 0.0, float), ("z_mod", "z_mod parameter:", None, str), ("z_mod_init", "z_mod_init:", 0.0, float), ("z_mod_frq", "z_mod_frq:", 2, int), ("z_mod_off", "z_mod_off:", 0.0, float), ("len", "len parameter:", None, str))
        DICT_HINTS = {"v": "Construct a planar polygon with basic order of symmetry and number of vertices v. Enter integers. For example, entering 4 will produce a square, other modifications notwithstanding.", "phi_init": "Defines a fixed rotation angle around the main axis relative to the first belt with v > 1. Enter 's' for a staggered, 0 for an ecliptic conformation, or any other angle (read in degrees).", "xy_factor": "Initial width, will be multiplied with the free xy or len parameter to give the distance of these vertices from the main axis. Should usually be either 1 or 0 in xyz-type belts. If len-type, sqrt(xy_factor^2 + z_factor^2) should usually equal 1.", "z_factor": "Initial height, will be multiplied with the free z or len parameter to give the vertical component of the distance of these vertices from the origin. If len-type, sqrt(xy_factor^2 + z_factor^2) should usually equal 1.", "erase": "This is a convenience feature for belts lacking one or more vertices. Enter vertex indices smaller than v, separated by commas. For example, the middle belt of the seesaw (v = 3) has '0' ('1' or '2' would be equivalent in this case) here. Leave empty to include all vertices given by v. Non-#molecule polyhedra with overall horizontal imbalances will be inaccurately processed, so use this with care.", "phi": "Parameter allowing free rotation around the main axis. Enter any string of characters. Will be added to phi_init, not multiplied. Enter 'phi*' to have this belt counterbalance any rotations in other belts.", "xy": "Free parameter governing the belt width. Enter any string. Enter 'None' or leave empty if xy_factor is 0.", "z": "Free parameter governing the vertical cylindrical coordinate. Enter 'None' or leave empty if z_factor is 0. Enter 'z*' and set z_factor to 1 to have this belt balance out the others, centering the model polyhedron at the centroid of the real ligand atoms.", "len": "Free parameter for scaling the model polyhedron, basically constraining the xy and z parameters into a single variable. Enter any string of characters. The height to width ratio is determined by xy_factor and z_factor which should, on average, combine to sqrt(xy_factor^2 + z_factor^2) = 1 (unless xy_mod_init or z_mod_init have nonzero values).", "phi_mod": "Free parameter to longitudially modulate phi. Enter any string of characters to enable.", "phi_mod_init": "Defines a fixed modulation amplitude. Enter any angle (read in degrees). This is the angle each vertex is adjusted by if phi_mod_frq = v / 2 and phi_mod_off = 0.5.", "phi_mod_frq": "Modulation frequency. Enter integer denominators of v. For example, starting from a hexagon (v = 6, 6/mmm), at phi_mod_off = 0.5, phi_mod_frq = 3 will preserve -6m2 while phi_mod_frq = 2 will preserve mmm and choosing phi_mod_frq = 1 will preserve mm2.", "phi_mod_off": "Offset of a sine wave with amplitude phi_mod + phi_mod_init and frequency phi_mod_frq. Enter any angle (read as a fraction of v, for example phi_mod_off = 1 translates to 90 degrees if v = 4).", "xy_mod": "Free parameter to transversally modulate xy. Enter any string of characters.", "xy_mod_init": "Defines a fixed modulation amplitude. Enter any distance (read in Angstrom). This will be equal to the horizontal displacement of each vertex from the belt equator if xy_mod_frq = v / 2 and xy_mod_off = 0.", "xy_mod_frq": "Modulation frequency. Enter integer denominators of v. For example, at xy_mod_frq = 2 and xy_mod_off = 0, a square will distort into a rhombus", "xy_mod_off": "Offset of a cosine wave with amplitude xy_mod + xy_mod_init and frequency xy_mod_frq. Enter any angle (read as a fraction of v, for example xy_mod_off = 1 translates to 90 degrees if v = 4).", "z_mod": "Free parameter to transversally modulate z. Enter any string of characters.", "z_mod_init": "Defines a fixed modulation amplitude. Enter any distance (read in Angstrom). This will be equal to the displacement of each vertex from the belt equator if z_mod_frq = v / 2 and z_mod_off = 0.", "z_mod_frq": "Modulation frequency. Enter integer denominators of v. For example, at z_mod_off = 0, this will change a benzene-shaped belt into a cyclohexane-shaped belt (z_mod_frq = 3 -> chair conformation, z_mod_frq = 2 -> faux tub conformation). z_mod_frq = 1 will shear the belt in z direction.", "z_mod_off": "Offset of a cosine wave with amplitude z_mod + z_mod_init and frequency z_mod_frq. Enter any angle (read as a fraction of v, for example z_mod_off = 1 translates to 90 degrees if v = 4)."}
        self.POLY_FIELD_TUPLES = (("name", "name:", "Enter a unique name for your polyhedron (mandatory).", 36, 0, 0), ("point_group", "point group:", "You may enter the point group of your polyhedron (Schoenfließ or Hermann-Mauguin notation). This has no effect on the calculations.", 12, 0, 1), ("axis", "main axis symmetry:", "You may enter the polyhedron symmetry in the direction of the main axis (Hermann-Mauguin notation only, e. g. '-4' or '6/m'). If this axis has a rotoinversion or horizontal mirror plane, Polynator will only look at a half-spheres worth of possible orientations (the other half is redundant).", 12, 0, 2), ("aristohedra", "aristohedra:", "You may give a list of related polyhedra this is a distorted version of.", 32, 1, 0), ("tags", "tags:", "Enter categories your polyhedron fits into or other search terms, e. g. '#antiprism, #equilateral, #constrained'. Most tags have no side effects, but there are three exceptions: #molecule centers the polyhedron on its central atom (useful for e. g. pseudopolyhedra); #axis_fixed prevents Polynator from manipulating the initial main axis (useful for e. g. skew prisms) and #chiral adds 'clockwise' or 'anticlockwise' after fitting.", 50, 2, 0))
        self.BUNDLE_FIELD_LABELS = (("bundle", "variable:", "You may enter a variable name. Multiple parameters assigned to the same variable name will all depend on this single variable. The variable name must be distinct from mathematical operators such as cos, sqrt, etc.", 10), ("min", "min:", "Lower bound of the variable value.", 8), ("init", "initial:", "Initial variable value. This value must correspond to the shape of the polyhedron when it is first constructed, with free and constrained parameters unmodified.", 8), ("max", "max:", "Upper bound of the variable value.", 8), ("parameter", "dependent parameter:", "Enter parameters that serve a function in the 'belts' section. For example, if you have defined a parameter 'xy1' and want it to depend on variable 'var', you should enter 'xy1' here.", 14), ("dependency", "dependency expression:", "How is the bound parameter related to the variable? Example input: 'var / sqrt(2)' -> The parameter scales linearly with the variable 'var', with the inverse square root of 2 as the proportionality constant. Valid built-in symbols include +, -, *, /, ** or ^ (exponent symbols), sqrt(), sin(), cos(), tan(), asin(), acos(), atan(), pi, ( and ).", 32))
        OUTER_FRAME_PROPERTIES = ((0, 0, self.menu, {0: 0, 1: 1}, (0, 1), 5, 1), (0, 1, self.menu, {0: 0, 1: 1}, (0, 1), 5, 1), (1, 0, self.menu, {}, (0,), (0, 2), 2)) #row, column, frame, rowconfigure weights, columnconfigure indices, pady, columnspan
        INNER_FRAME_PROPERTIES = (("general properties:", 0, 0, 0, {}, {0: 1}), ("belts:", 0, 1, 0, {(0, 1): 0, 2: 1}, {0: 1}), ("viewer:", 1, 0, 1, {0: 1}, {0: 1}), ("parameter bundles:", 1, 1, 1, {(0, 1): 0, 2: 1}, {0: 1})) #label, outer frame, row, column, rowconfigure weights, columnconfigure weights
        BELT_BUTTON_TUPLES = (("  insert at index:", self.add_belt), ("edit entry", self.edit_belt), ("remove entry", lambda: self.remove_belt(self.listbox_belts.curselection())))
        BUNDLE_BUTTON_TUPLES = (("confirm entry", lambda: self.add_bundle_entry()), ("edit entry", lambda: self.edit_bundle_entry(self.listbox_parameter_bundles.curselection())), ("remove entry", lambda: self.remove_bundle_entry(self.listbox_parameter_bundles.curselection())))
        self.name = blueprint[0] if blueprint else ""
        self.dict_info = blueprint[1] if blueprint else {"parameter_bundles": {}, "var_range": {}}
        self.belt_dicts = blueprint[2] if blueprint else ()
        self.visual_poly = None
        outer_frames, inner_frames = [], []
        for item_frame in OUTER_FRAME_PROPERTIES:
            frame = tk.Frame(item_frame[2], bg = "#ffffff")
            outer_frames.append(frame)
        for item_frame in INNER_FRAME_PROPERTIES:
            frame = tk.LabelFrame(outer_frames[item_frame[1]], text = item_frame[0], bg = "#ffffff")
            inner_frames.append(frame)
        give_hint(inner_frames[3], self, "Parameter bundles allow you to tie multiple parameters to a single variable. They're useful for constructing polyhedra which deform in unusually constrained ways, as some of the built-in polyhedra with cubic point groups do (e. g. the tetrahedral cuboctahedron and the pyritohedron). The average vertex distance from the polyhedron center should be invariant under the deformation (use e. g. non-bundled len parameters for radial differences).")
        button_confirm = Buttoon(outer_frames[2], text = "save", command = self.save, width = 20, pady = 6)
        button_confirm.frame.grid(row = 0, column = 1, padx = 5, pady = (2, 4), sticky = "nsew")
        button_confirm.butt.bind("<Leave>", lambda event: self.commentline.configure(text = ""))
        frame_poly_fields = tk.Frame(inner_frames[0], borderwidth = 0, highlightthickness = 0, bg = "#ffffff")
        self.poly_fields = []
        poly_field_subframes = [(tk.Frame(frame_poly_fields, borderwidth = 0, highlightthickness = 0, bg = "#ffffff"), c) for c in (3, 1, 1)]
        for count_tuple, item_field_info in enumerate(self.POLY_FIELD_TUPLES):
            frame_tuple = poly_field_subframes[item_field_info[4]]
            label = tk.Label(frame_tuple[0], text = item_field_info[1], bg = "#ffffff")
            field = ttk.Entry(frame_tuple[0], width = item_field_info[3])
            give_hint(field, self, item_field_info[2])
            label.grid(row = 0, column = item_field_info[5], padx = 3, sticky = "w")
            field.grid(row = 1, column = item_field_info[5], padx = 5, pady = (0, 5), sticky = "ew")
            self.poly_fields.append(field)
            if item_field_info[0] == "name":
                field.insert(0, self.name)
            elif item_field_info[0] in ("axis", "point_group"):
                field.insert(0, self.dict_info.get(item_field_info[0], ""))
            else:
                for item_iterable_entry in ("aristohedra", "tags"):
                    if item_field_info[0] == item_iterable_entry:
                        entry_set = set(self.dict_info.get(item_iterable_entry, ()))
                        entry_set.discard("#custom")
                        entry_string = ", ".join(entry_set)
                        field.insert(0, entry_string)
        for count_row, item_frame_tuple in enumerate(poly_field_subframes):
            item_frame_tuple[0].columnconfigure(tuple(range(item_frame_tuple[1])), weight = 1)
            item_frame_tuple[0].grid(row = count_row, column = 0, sticky = "nsew")
        frame_belt_fields = tk.Frame(inner_frames[1], borderwidth = 0, highlightthickness = 0, bg = "#ffffff")
        self.belt_fields = []
        for count_entry, item_entry in enumerate(self.BELT_TUPLES):
            label = tk.Label(frame_belt_fields, text = item_entry[1], bg = "#ffffff")
            field = ttk.Entry(frame_belt_fields, width = 18)
            field.insert(0, (str(item_entry[2]) if item_entry[2] != None else ""))
            give_hint(field, self, DICT_HINTS.get(item_entry[0]))
            label.grid(row = 2 * (count_entry // 4), column = count_entry % 4, sticky = "w", padx = 3)
            field.grid(row = 2 * (count_entry // 4) + 1, column = count_entry % 4, sticky = "ew", padx = 5, pady = (0, 5))
            self.belt_fields.append(field)
        frame_belt_fields.columnconfigure((0, 1, 2, 3), weight = 1)
        frame_belt_fields.grid(row = 0, column = 0, sticky = "nsew")
        for count_button, item_button in enumerate(BELT_BUTTON_TUPLES):
            button = Buttoon(frame_belt_fields, text = item_button[0], command = item_button[1], width = 14, pady = 3)
            button.frame.grid(row = 10, column = count_button + 1, rowspan = 2, padx = 5, pady = (0, 5), sticky = "sew")
            if count_button == 0:
                button.butt.configure(anchor = "w")
                self.field_belt_index = ttk.Entry(button.frame, width = 2, background = "#666666")
                self.field_belt_index.grid(row = 0, column = 0, sticky = "e", padx = 3)
                give_hint(button.butt, self, "Click here to add a belt with properties specified in the fields above to your polyhedron. If belts are listed below already, you may enter an index to specify where your belt will be inserted (e. g. '0' to insert it at the top). Greyed out entries in the list below will be replaced by the new belt.")
        frame_listbox_belts = tk.Frame(inner_frames[1], bg = "#a0a0a0", padx = 1, pady = 1)
        self.listbox_belts = tk.Listbox(frame_listbox_belts, height = 7, width = 72, highlightthickness = 0, relief = tk.FLAT)
        self.listbox_belts.grid(sticky = "nsew")
        for item_dict in self.belt_dicts:
            self.listbox_belts.insert(tk.END, ", ".join([f"{i} = {str(item_dict.get(i))}" for i in item_dict]))
        frame_listbox_belts.rowconfigure(0, weight = 1), frame_listbox_belts.columnconfigure(0, weight = 1)
        frame_listbox_belts.grid(row = 2, column = 0, padx = 5, pady = 4, sticky = "nsew")
        self.canvas = tk.Canvas(inner_frames[2], highlightthickness = 0, bg = "#000000", height = 390, width = 390)
        self.canvas.grid(padx = 5, pady = 5, sticky = "nsew")
        self.bundle_fields = []
        frame_bundle_fields = tk.Frame(inner_frames[3], bg = "#ffffff")
        for count_entry, item_entry in enumerate(self.BUNDLE_FIELD_LABELS):
            label = tk.Label(frame_bundle_fields, text = item_entry[1], bg = "#ffffff")
            field = ttk.Entry(frame_bundle_fields, width = item_entry[3])
            give_hint(field, self, item_entry[2])
            label.grid(row = 2 * (count_entry // 5), column = count_entry % 5, sticky = "w", padx = 3)
            field.grid(row = 2 * (count_entry // 5) + 1, column = count_entry % 5, columnspan = 5 if count_entry == 5 else 1, sticky = "ew", padx = (5, 5), pady = (0, 5))
            self.bundle_fields.append(field)
        frame_bundle_fields.columnconfigure((0, 1, 2, 3, 4), weight = 1), frame_bundle_fields.grid(row = 0, column = 0, sticky = "nsew")
        frame_bundle_buttons = tk.Frame(inner_frames[3], borderwidth = 0, highlightthickness = 0, bg = "#ffffff")
        for count_button, item_button in enumerate(BUNDLE_BUTTON_TUPLES):
            button = Buttoon(frame_bundle_buttons, text = item_button[0], command = item_button[1], width = 20, pady = 3)
            button.frame.grid(row = 0, column = count_button, padx = 5, pady = 4, sticky = "nsew")
        frame_bundle_buttons.columnconfigure((0, 1, 2), weight = 1)
        frame_bundle_buttons.grid(row = 1, column = 0, sticky = "nsew")
        frame_listbox_parameter_bundles = tk.Frame(inner_frames[3], bg = "#a0a0a0", padx = 1, pady = 1)
        self.listbox_parameter_bundles = tk.Listbox(frame_listbox_parameter_bundles, height = 4, width = 72, highlightthickness = 0, relief = tk.FLAT)
        self.listbox_parameter_bundles.grid(sticky = "nsew")
        for item_var in self.dict_info.get("parameter_bundles", {}):
            dict_bundle = self.dict_info.get("parameter_bundles").get(item_var)
            range_tuple = self.dict_info.get('var_range').get(item_var, (0, 1, 0.5))
            self.listbox_parameter_bundles.insert(tk.END, f"{item_var} ({range_tuple[0]} <= {range_tuple[2]} <= {range_tuple[1]}):")
            for item_param in dict_bundle:
                self.listbox_parameter_bundles.insert(tk.END, f"    {item_param} = {dict_bundle.get(item_param)}")
        frame_listbox_parameter_bundles.rowconfigure(0, weight = 1), frame_listbox_parameter_bundles.columnconfigure(0, weight = 1)
        frame_listbox_parameter_bundles.grid(row = 2, column = 0, padx = 5, pady = 4, sticky = "nsew")
        frame_poly_fields.columnconfigure(0, weight = 1)
        frame_poly_fields.grid(row = 0, column = 0, sticky = "nsew")
        self.commentline = tk.Label(outer_frames[2], justify = tk.LEFT, wraplength = 825, height = 3, width = 36, bg = "#ffffff", anchor = "nw")
        self.commentline.grid(row = 0, column = 0, padx = 2, sticky = "nsew")
        for count_frame, item_frame in enumerate(inner_frames):
            properties = INNER_FRAME_PROPERTIES[count_frame]
            for item_dict_entry in properties[4]:
                item_frame.rowconfigure(item_dict_entry, weight = properties[4].get(item_dict_entry))
            for item_dict_entry in properties[5]:
                item_frame.columnconfigure(item_dict_entry, weight = properties[5].get(item_dict_entry))
            item_frame.grid(row = properties[2], column = properties[3], padx = 5, pady = 5, ipadx = 5, ipady = 5, sticky = "nsew")
        for count_frame, item_frame in enumerate(outer_frames):
            properties = OUTER_FRAME_PROPERTIES[count_frame]
            for item_dict_entry in properties[3]:
                item_frame.rowconfigure(item_dict_entry, weight = properties[3].get(item_dict_entry))
            item_frame.columnconfigure(properties[4], weight = 1)
            item_frame.grid(row = properties[0], column = properties[1], columnspan = properties[6], pady = properties[5], sticky = "nsew")
        self.menu.rowconfigure(0, weight = 1), self.menu.rowconfigure(1, weight = 0), self.menu.columnconfigure((0, 1), weight = 1) # I got something slightly wrong here, but am unable to find the mistake.
    
    def read_belt(self, index):
        DICT_BRACKETS = {"[": ("]", list), "(": (")", tuple)}
        belt_string = self.listbox_belts.get(index)
        belt_dict = {}
        for item_bra in DICT_BRACKETS:
            if f"erase = {item_bra}" in belt_string:
                ket_tuple = DICT_BRACKETS.get(item_bra)
                index_bra = belt_string.index(f", erase = {item_bra}")
                index_ket = belt_string.index(ket_tuple[0], index_bra + 10)
                erase_string = belt_string[index_bra + 11: index_ket]
                belt_dict["erase"] = ket_tuple[1]([int(i) for i in erase_string.replace(" ", "").split(";")])
                belt_string = belt_string[:index_bra] + belt_string[index_ket + 1:]
        entry_strings = belt_string.split(", ")
        for item_entry_string in entry_strings:
            key, value = item_entry_string.split(" = ")
            value_type = int
            for item_char in value:
                if item_char not in ".+-0123456789":
                    value_type = str
                    break
                elif item_char == ".":
                    value_type = float
            belt_dict[key] = value_type(value)
        return belt_dict
    
    def add_belt(self):
        def key_tuple_generator(fields):
            for count_field, item_field in enumerate(fields):
                entry = item_field.get()
                key_string = self.BELT_TUPLES[count_field][0]
                if 7 < count_field < 20:
                    entry_split = entry.replace(",", " ").split()
                    for count_plus, item_entry in enumerate(entry_split):
                        index_underline = key_string.index("mod") + 3
                        key_string = key_string[:index_underline] + count_plus * "+" + key_string[index_underline:]
                        yield count_field, key_string, item_entry
                else:
                    yield count_field, key_string, entry
        try:
            v = int(self.belt_fields[0].get())
            if v < 1:
                raise ValueError
        except ValueError:
            self.commentline.configure(text = "Please enter a positive integer for 'v'!", foreground = "#ff0000")
            return
        if self.belt_fields[4].get() == "" and self.belt_fields[5].get() == "" and self.belt_fields[20].get() == "":
            self.commentline.configure(text = "At least one parameter out of 'xy', 'z', and 'len' is required!", foreground = "#ff0000")
            return
        try:
            belt_index = int(self.field_belt_index.get())
            if belt_index > self.listbox_belts.size() or belt_index < 0:
                belt_index = self.listbox_belts.size() + 1
        except ValueError:
            belt_index = tk.END
        dict_belt = {}
        for count_field, key_string, entry in key_tuple_generator(self.belt_fields):
            default_value = str(self.BELT_TUPLES[count_field][2])
            required_type = self.BELT_TUPLES[count_field][3]
            if entry == default_value or entry == "":
                continue
            if required_type == str:
                if not entry[0].isalpha():
                    self.commentline.configure(text = f"Entry for {key_string} must start with a letter!", foreground = "#ff0000")
                    return
                for item_substring in (".", "(", ")", "+", "-", "/", "^", "$", "§", "sin", "cos", "tan", "asin", "acos", "atan", "sqrt", "pi"):
                    if item_substring in entry:
                        self.commentline.configure(text = f"Entry for {key_string} must not contain '{item_substring}!", foreground = "#ff0000")
                        return
                dict_belt[key_string] = entry
            elif required_type == float:
                try:
                    dict_belt[key_string] = parse_math(entry, dict_variables = {"pi": 3.141592653589793})
                except (NameError, SyntaxError, ValueError):
                    self.commentline.configure(text = f"Entry for {key_string} must be a valid number!", foreground = "#ff0000")
                    return
            elif required_type == int:
                try:
                    dict_belt[key_string] = int(entry)
                except ValueError:
                    self.commentline.configure(text = f"Entry for {key_string} must be an integer!", foreground = "#ff0000")
                    return
            elif key_string == "erase":
                list_entries = []
                entry_split = entry.replace(",", " ").strip("*([{}])").split()
                for item_index_string in entry_split:
                    try:
                        list_entries.append(int(item_index_string) % dict_belt.get("v"))
                    except ValueError:
                        self.commentline.configure(text = "Deleted vertices must be a set of integers!", foreground = "#ff0000")
                        return
                if entry[0] in "*[":
                    dict_belt[key_string] = str(list_entries).replace(",", ";")
                elif len(list_entries) == 1:
                    dict_belt[key_string] = f"({list_entries[0]})"
                else:
                    dict_belt[key_string] = str(tuple(list_entries)).replace(",", ";")
            elif key_string == "phi_init":
                if entry == "s":
                    dict_belt[key_string] = "s"
                    continue
                try:
                    dict_belt[key_string] = str(parse_math(entry, dict_variables = {"pi": 3.141592653589793}))
                except (NameError, SyntaxError, ValueError):
                    self.commentline.configure(text = "Entry for phi_init must be a valid number!", foreground = "#ff0000")
                    return
        for item_index in range(self.listbox_belts.size() - 1, -1, -1):
            if "" != self.listbox_belts.itemcget(item_index, "foreground") != "#000000":
                self.listbox_belts.delete(item_index)
        paint_it_black(self.listbox_belts)
        belt_string = str(dict_belt)[1:-1].replace(":", " =").replace("'", "")
        self.listbox_belts.insert(belt_index, belt_string)
        self.visualize_polyhedron()
    
    def edit_belt(self):
        if not self.listbox_belts.size():
            return
        paint_it_black(self.listbox_belts)
        index_tuple = self.listbox_belts.curselection()
        if index_tuple:
            index_entry = index_tuple[0]
        else:
            index_entry = 0
        self.field_belt_index.delete(0, last = tk.END)
        self.field_belt_index.insert(0, str(index_entry))
        self.listbox_belts.itemconfigure(index_entry, foreground = "#bbbbbb")
        belt_dict = self.read_belt(index_entry)
        for count_tuple, item_tuple in enumerate(self.BELT_TUPLES):
            field = self.belt_fields[count_tuple]
            field.delete(0, tk.END)
            if 7 < count_tuple < 20:
                entries = []
                for count_plus in range(100):
                    index_underline = item_tuple[0].index("mod") + 3
                    key_string = item_tuple[0][:index_underline] + count_plus * "+" + item_tuple[0][index_underline:]
                    if key_string in belt_dict:
                        entries.append(str(belt_dict.get(key_string)))
                    elif not count_plus and item_tuple[2] != None:
                        entries.append(str(item_tuple[2]))
                    else:
                        break
                entry_string = ", ".join(entries)
            elif item_tuple[0] in belt_dict:
                entry_string = str(belt_dict.get(item_tuple[0]))
            elif item_tuple[2] != None:
                entry_string = str(item_tuple[2])
            else:
                entry_string = ""
            field.insert(0, entry_string)
    
    def remove_belt(self, index_tuple):
        if not self.listbox_belts.size():
            return
        paint_it_black(self.listbox_belts)
        if index_tuple:
            index_entry = index_tuple[0]
        else:
            index_entry = 0
        self.listbox_belts.delete(index_entry)
        self.visualize_polyhedron()
    
    def read_all_belts(self):
        list_belt_dicts = []
        z_init = inf
        for count_belt in range(self.listbox_belts.size()):
            belt_dict = self.read_belt(count_belt)
            z_init_new = belt_dict.get("z_factor", 1.0)
            if z_init_new > z_init:
                self.commentline.configure(text = "Belts must be sorted by z_factor values in descending order!", foreground = "#ff0000")
                return
            list_belt_dicts.append(belt_dict)
        self.belt_dicts = tuple(list_belt_dicts)
    
    def add_bundle_entry(self):
        bundle = self.bundle_fields[0].get()
        min_string = self.bundle_fields[1].get()
        init_string = self.bundle_fields[2].get()
        max_string = self.bundle_fields[3].get()
        parameter = self.bundle_fields[4].get()
        dependency = self.bundle_fields[5].get().replace(" ", "")
        for count, item_range in enumerate((min_string, max_string, init_string)):
            if item_range:
                try:
                    parse_math(item_range)
                except (NameError, SyntaxError, ValueError):
                    self.commentline.configure(text = f"Variable range {self.BUNDLE_FIELD_LABELS[count + 1][0]} must be a valid number!", foreground = "#ff0000")
                    return
        if not bundle or not parameter or not dependency:
            if bundle and min_string and max_string and init_string:
                if not self.listbox_parameter_bundles.size():
                    self.listbox_parameter_bundles.insert(0, f"{bundle} ({min_string} <= {init_string} <= {max_string}):")
                    return
                for count_string, item_string in enumerate(self.listbox_parameter_bundles.get(0, tk.END)):
                    if bundle in item_string and item_string[-1] == ":":
                        self.listbox_parameter_bundles.delete(count_string)
                        self.listbox_parameter_bundles.insert(count_string, f"{bundle} ({min_string} <= {init_string} <= {max_string}):")
                        return
            self.commentline.configure(text = "Bundle entries incomplete!", foreground = "#ff0000")
            return
        domain = []
        if not self.listbox_parameter_bundles.size():
            domain = [max((min_string, "0"), key = lambda i: len(i)), max((init_string, "0"), key = lambda i: len(i)), max((max_string, "1"), key = lambda i: len(i))]
            if not parse_math(domain[0]) <= parse_math(domain[1]) <= parse_math(domain[2]):
                self.commentline.configure(text = "Invalid variable range!", foreground = "#ff0000")
                return
        index_bundle, index_last = -1, -1
        for count_string, item_string in enumerate(self.listbox_parameter_bundles.get(0, tk.END)):
            if bundle in item_string and item_string[-1] == ":":
                index_bundle = count_string
                if domain == []:
                    domain = [parse_math(i) for i in item_string[item_string.index("(") + 1:-2].split(" <= ")]
            elif index_bundle > -1:
                index_last = count_string + 1
                if item_string[-1] == ":":
                    index_last -= 1
                    break
        if index_last > 0:
            self.listbox_parameter_bundles.delete(index_bundle)
            self.listbox_parameter_bundles.insert(index_bundle, f"{bundle} ({domain[0]} <= {domain[1]} <= {domain[2]}):")
            self.listbox_parameter_bundles.insert(index_last, f"    {parameter} = {dependency}")
        else:
            self.listbox_parameter_bundles.insert(tk.END, f"{bundle} ({domain[0]} <= {domain[1]} <= {domain[2]}):")
            self.listbox_parameter_bundles.insert(tk.END, f"    {parameter} = {dependency}")
        self.visualize_polyhedron()
    
    def edit_bundle_entry(self, index_tuple):
        if not self.listbox_parameter_bundles.size():
            return
        paint_it_black(self.listbox_parameter_bundles)
        if index_tuple:
            index_entry = index_tuple[0]
        else:
            index_entry = 0
        string_entry = self.listbox_parameter_bundles.get(index_entry)
        dict_entries = {}
        if string_entry[-1] == ":":
            dict_entries["bundle"], var_range = string_entry[:-2].split(" (")
            dict_entries["min"], dict_entries["init"], dict_entries["max"] = var_range.split(" <= ")
        else:
            dict_entries["parameter"], dict_entries["dependency"] = string_entry[4:].split(" = ")
            all_entries = self.listbox_parameter_bundles.get(0, tk.END)
            for item_entry in all_entries[index_entry::-1]:
                if item_entry[-1] == ":":
                    dict_entries["bundle"], var_range = item_entry[:-2].split(" (")
                    dict_entries["min"], dict_entries["init"], dict_entries["max"] = var_range.split(" <= ")
                    break
        for count_tuple, item_field_tuple in enumerate(self.BUNDLE_FIELD_LABELS):
            self.bundle_fields[count_tuple].delete(0, tk.END)
            self.bundle_fields[count_tuple].insert(0, dict_entries.get(item_field_tuple[0]))
    
    def remove_bundle_entry(self, index_tuple):
        if not self.listbox_parameter_bundles.size():
            return
        if index_tuple:
            index_entry = index_tuple[0]
        else:
            index_entry = tk.END
        string_entry = self.listbox_parameter_bundles.get(index_entry)
        if string_entry[-1] == ":":
            self.listbox_parameter_bundles.delete(index_entry)
            further_entries = self.listbox_parameter_bundles.get(index_entry, tk.END)
            for count_entry, item_entry in enumerate(further_entries):
                if item_entry[-1] == ":":
                    break
                self.listbox_parameter_bundles.delete(index_entry)
        else:
            self.listbox_parameter_bundles.delete(index_entry)
        all_entries = self.listbox_parameter_bundles.get(0, tk.END)
        init = False
        for count_entry, item_entry in enumerate(all_entries):
            if item_entry[-1] == ":":
                if init == True:
                    self.listbox_parameter_bundles.delete(count_entry - 1)
                if count_entry + 1 == self.listbox_parameter_bundles.size():
                    self.listbox_parameter_bundles.delete(count_entry)
                init = True
            else:
                init = False
        self.visualize_polyhedron()
    
    def read_bundles(self):
        all_entries = self.listbox_parameter_bundles.get(0, tk.END)
        for item_entry in all_entries:
            if item_entry[-1] == ":":
                bundle_n_range = item_entry[:-2].split(" (")
                bundle = bundle_n_range[0]
                var_range = bundle_n_range[1].split(" <= ")
                self.dict_info.get("var_range")[bundle] = tuple([float(var_range[i]) for i in (0, 2, 1)])
                dict_bundle = self.dict_info.get("parameter_bundles")[bundle] = {}
            else:
                parameter, dependency = item_entry[4:].split(" = ")
                dict_bundle[parameter] = dependency
    
    def read_info(self):
        tag_set = {"#custom"}
        for count_tuple, item_tuple in enumerate(self.POLY_FIELD_TUPLES):
            string_entry = self.poly_fields[count_tuple].get()
            if item_tuple[0] == "name":
                self.name = string_entry
            elif item_tuple[0] == "tags":
                tags_split = string_entry.replace(",", " ").split()
                for item_tag in tags_split:
                    tag_set.add("#" + item_tag.removeprefix("#"))
            elif item_tuple[0] in ("point_group", "axis"):
                entry_string = string_entry.replace(" ", "")
                self.dict_info[item_tuple[0]] = entry_string if entry_string else "1"
            elif item_tuple[0] == "aristohedra":
                dict_aristo = {}
                split_entry = string_entry.replace(",", " ").split()
                for item_aristo_name in split_entry:
                    dict_aristo[item_aristo_name] = ()
                self.dict_info["aristohedra"] = dict_aristo
        self.dict_info["tags"] = tuple(tag_set)
    
    def read_full(self):
        self.read_info()
        self.read_all_belts()
        self.read_bundles()
        set_free_parameters, set_constrained_parameters = set(), set()
        for item_belt_dict in self.belt_dicts:
            for item_type in ALL_PARAMETER_TYPES:
                parameter = item_belt_dict.get(item_type)
                if not parameter:
                    continue
                if parameter in ("phi*", "z*"):
                    set_constrained_parameters.add(parameter)
                else:
                    set_free_parameters.add(parameter)
                if item_type.endswith("_mod"):
                    for count_plus in range(100):
                        bonus_type = item_type + count_plus * "+"
                        bonus_parameter = item_belt_dict.get(bonus_type)
                        if not bonus_parameter:
                            break
                        set_free_parameters.add(bonus_parameter)
        for item_variable in self.dict_info.get("parameter_bundles", {}):
            set_free_parameters.add(item_variable)
            for item_constrained_parameter in self.dict_info.get("parameter_bundles").get(item_variable):
                set_constrained_parameters.add(item_constrained_parameter)
        self.dict_info["parameter_count"] = len(set_free_parameters - set_constrained_parameters)
        return (self.name, self.dict_info, self.belt_dicts)
    
    def save(self):
        finished_blueprint = self.read_full()
        if self.name == "":
            self.commentline.configure(text = "Please enter a name for this polyhedron!", foreground = "#ff0000")
            return None
        elif self.name in POLY_KEYWORDS:
            self.commentline.configure(text = "This polyhedron name is already taken!", foreground = "#ff0000")
            return None
        keyword_candidates = [self.name, self.dict_info.get("point_group")] + list(self.dict_info.get("tags"))
        for item_keyword in keyword_candidates:
            if item_keyword not in self.parent.main_contents[2].keywords:
                self.parent.main_contents[2].keywords.append(item_keyword)
        coordination_number = sum([i.get("v") - len(i.get("erase", ())) for i in finished_blueprint[2]])
        if coordination_number not in self.parent.dict_blueprints_by_cn:
            self.parent.dict_blueprints_by_cn[coordination_number] = []
        self.parent.dict_blueprints_by_cn.get(coordination_number).append(finished_blueprint)
        self.parent.dict_custom_blueprints[self.name] = finished_blueprint
        self.parent.dict_blueprints_by_name[self.name] = finished_blueprint
        self.commentline.configure(text = f"{self.name} saved successfully!")
    
    def visualize_polyhedron(self):
        if self.visual_poly != None:
            self.visual_poly.kill()
        blueprint = self.read_full()
        if not blueprint[2]:
            return
        self.visual_poly = VisualPolyhedron(self, belt_dicts = blueprint[2], is_molecule = "#molecule" in blueprint[1].get("tags"), dict_bundles = blueprint[1].get("parameter_bundles", {}), dict_var_range = self.dict_info.get("var_range", {}))

class MainWindow: # Manages main GUI window and oversees the main process at the highest level. Should probably have been split in two, but it's too late now.
    def __init__(self, root, icon = "icon.ico"):
        self.dict_input = deepcopy(DICT_INPUT_DEFAULT)
        self.icon = icon
        self.preview_listboxes = []
        self.cifs = ()
        self.displayed_polyhedron = None
        self.main_contents = ()
        self.dict_ignore_envs = {}
        self.ignored_polyhedra = []
        self.warnings = set()
        self.visual_preferences = {"labels": False, "centers": True, u"\u0394 / pm": True, "real edges": False, "model edges": True, "axis": False}
        self.cifs = []
        self.list_output = []
        self.blocked_polyhedra = []
        self.set_partable_head = set()
        self.partable_line_data = []
        self.var_axis_fixed = tk.BooleanVar()
        self.n_evaluated_polyhedra = 0
        self.dict_blueprints_by_name = {}
        self.dict_blueprints_by_cn = deepcopy(DICT_BLUEPRINTS)
        self.dict_custom_blueprints = {}
        self.configpath = str(os.getcwd())
        self.dict_config_files = {}
        self.process_config()
        for item_name in self.dict_custom_blueprints:
            custom_blueprint = self.dict_custom_blueprints.get(item_name)
            coordination_number = sum([i.get("v") - len(i.get("erase", ())) for i in custom_blueprint[2]])
            if not coordination_number in self.dict_blueprints_by_cn:
                self.dict_blueprints_by_cn[coordination_number] = []
            self.dict_blueprints_by_cn.get(coordination_number).append(custom_blueprint)
        self.dict_blueprint_symmetrees, self.dict_upright_nodes = construct_symmetrees(self.dict_blueprints_by_cn)
        self.dict_upside_down_symmetrees, self.dict_upside_down_nodes = flip_symmetrees(self.dict_blueprint_symmetrees)
        # The rest of __init__ is GUI only:
        self.root = root
        FRAMESTUFF_1 = (("central atom criteria:", 0, 0, 1), ("ligand atom criteria:", 0, 1, 1), ("model criteria:", 0, 2, 1), ("fit settings:", 1, 0, 2), ("coordination environment filters:", 1, 1, 1), ("set main axis:", 2, 1, 1), ("input files:", 3, 0, 1), ("coordination environments:", 3, 1, 1), ("model polyhedra:", 3, 2, 1), ("viewer:", 0, 3, 3), ("results:", 3, 3, 2)) # title, row, column, rowspan
        frames_labelled = [tk.LabelFrame(self.root, text = i[0], bg = "#ffffff") for i in FRAMESTUFF_1]
        FRAMESTUFF_2 = ((frames_labelled[3], 0, 0, 1, "nsew"), (frames_labelled[3], 1, 0, 1, "nsew"), (self.root, 4, 0, 3, "nsew")) # parent frame, row, column, columnspan, sticky
        frames_structural = [tk.Frame(i[0], borderwidth = 0, highlightthickness = 0, bg = "#ffffff") for i in FRAMESTUFF_2]
        DICT_SEARCH_POLY_INIT = {"poly_criteria": {"and": {}, "or": {}, "not": {}}, "var_max": inf}
        DICT_SEARCH_CENTRAL_INIT = {"central_atom_criteria": {"and": {}, "or": {}, "not": {}}, "central_coordinates": []}
        DICT_SEARCH_LIGAND_INIT = {"ligand_criteria": {"and": {}, "or": {}, "not": {}}, "ligand_coordinates": []}
        CHECKSTUFF_FIT = (("preserve_radial_avg", "preserve radial average", self.dict_input.get("preserve_radial_avg"), "Find the best model polyhedron given the restriction that model vectors need to have the same average distance from the center as real vectors (optimal model vectors are generally slightly shorter)."), ("include_central_atom_in_fit", "include central atom in fit", self.dict_input.get("include_central_atom_in_fit"), "The central atom factors in when fitting model polyhedra. The excentricity is treated as an additional distance contributing to the sum of least squares when calculating the csm."))
        RADIOSTUFF_CENTERMETHODS = {("center_method", "polyhedron centering:"): (("default", "best fit", "Model polyhedra are generally centered on the centroid of all atoms. This includes the central atom if it's included in the fit (see options above). Exception: Model polyhedra with the #molecule tag (this includes all pseudopolyhedra) are centered on the central atom."), ("force_central_atom", "central atom", "The model polyhedron is centered at the central atom. This treatment is given to model polyhedra with the #molecule tag by default."), ("force_centroid", "ligand centroid", "The model polyhedron is centered on the centroid of its ligands. The central atom doesn't factor in, even if it is relevant to the fit and even if it has the #molecule tag."))}
        ENTRYSTUFF_ENV_FILTERS = (("d_min", "d_min:", "0.1", "Minimal distance in Angstrom from each ligand atom to the central atom. Useful for accessing coordination spheres beyond the first.", float), ("d_max", "d_max:", "3.5", "Maximal distance in Angstrom of each ligand atom to the central atom.", float), ("cn_max", "CN_max:", "60", "Maximal number of ligand atoms. Excess atoms with the largest distance to the central atom will be dismissed until the specified number of atoms is left.", int), ("find_coordination_spheres", "coord. sphere:", "", "You may enter the order of one or more coordination spheres (e. g. '1, 2' for the first and second sphere), to automatically find the atoms constituting those spheres. This will override d_min, d_max and CN_max. Warning: This feature is not fully reliable, especially for disorderly coordination environments.", str))
        COMMENTS_BUTTONS = ("Configure Polynator to suit your needs.", "See what you're working with to prevent unnecessary runs.", "Depending on the total number and types of model polyhedra to be evaluated, the calculations may take some time. You can speed up the process by filtering out model polyhedra you aren't interested in, either in the 'polyhedron criteria' section or more permanently in the settings menu. Ready? Let's go!", "Generate output files for the last run. You will find them in a subdirectory of the folder you selected the input file(s) from. You may customize the output via the 'Settings' menu.")
        COMMENTS_CRITSEARCH = ("Enter any element symbol or atom name specific to your .cif file. Press Enter to submit with the currently selected logic operator ('and' / 'or' / 'not'). You may enter coordinates (e. g. '0, 1/2, 0.83') to generate a dummy center (useful if you want to evaluate an empty polyhedron).", "Enter any element symbol or atom name specific to your .cif file. Press 'Enter' to submit with the currently selected logic operator ('and' / 'or' / 'not'). Every atom in the input .cif file within range will qualify as a ligand if this is left empty. You may enter coordinates (e. g. '0, 1/2, 0.83') to generate a dummy ligand which will not be replicated by symmetry.", "Enter any polyhedron name or fragment thereof. Press 'Enter' to submit with the currently selected logic operator ('and' / 'or' / 'not'). All polyhedra of the appropriate coordination number are evaluated if left empty. Enter '*n', where n is any positive integer, to set a maximum for the degrees of freedom per polyhedron (polyhedra with more than n are filtered out).")
        frame_output = tk.Frame(frames_labelled[10], bd = 1, bg = "#808080")
        self.listbox_results = tk.Listbox(frame_output, height = 14, width = 80, relief = tk.FLAT)
        self.canvas = tk.Canvas(frames_labelled[9], highlightthickness = 0, bg = "#000000", height = 390, width = 440)
        give_hint(self.canvas, self, "Click on one of the items in the 'coordination environments', 'polyhedra' or 'results' lists to visualize the respective content here. Drag with your mouse to rotate the polyhedron and scroll to zoom in or out. Click on the boxes to hide parts of the image or to freeze the movements associated with the respective parameters.")
        self.field_main_axis = ttk.Entry(frames_labelled[5], width = 12)
        self.commentline = tk.Label(frames_structural[2], justify = tk.LEFT, wraplength = 660, height = 3, bg = "#ffffff", anchor = "nw")
        self.commentline.grid(row = 0, column = 0, sticky = "nsew")
        self.root.grid_rowconfigure((0, 3), minsize = 150, weight = 1)
        self.root.grid_columnconfigure((0, 1, 2), weight = 1)
        self.root.grid_columnconfigure(3, weight = 2)
        for count in range(3):
            frame = tk.Frame(frames_labelled[count + 6], bd = 1, bg = "#808080")
            listbox = tk.Listbox(frame, height = 9, width = 40 if count == 2 else 32, relief = tk.FLAT, highlightthickness = 0)
            listbox.grid(sticky = "nsew"), frame.grid(row = 0, column = 0, padx = 4, pady = 5, columnspan = 2, sticky = "nsew")
            frame.rowconfigure(0, weight = 1), frame.columnconfigure(0, weight = 1)
            self.preview_listboxes.append(listbox)
            frame_buttons = ttk.Frame(frames_labelled[count + 6])
            if count == 0:
                for item_filepath in self.dict_input.get("cif_filepaths"):
                    listbox.insert(tk.END, item_filepath)
                button_get_cifs = Buttoon(frame_buttons, width = 15, text = "load input files", command = self.get_cifs)
                give_hint(button_get_cifs.butt, self, "Select one or more .cif files to be analyzed. Output files will go into a subfolder of the folder you selected from.")
                button_get_cifs.frame.grid(row = 1, column = 0, padx = 5, pady = (0, 5))
            elif count == 1:
                button_remove = Buttoon(frame_buttons, width = 8, text = "ignore", command = lambda i = listbox: self.ignore_envs(i))
                button_remove.frame.grid(row = 1, column = 0, padx = 5, pady = (0, 5))
                give_hint(button_remove.butt, self, "Temporarily ignore the selected coordination environment. It will not be evaluated.")
            elif count == 2:
                button_remove = Buttoon(frame_buttons, width = 8, text = "ignore", command = lambda i = listbox: self.ignore_poly(i))
                button_edit = Buttoon(frame_buttons, width = 14, text = "use as blueprint", command = lambda i = listbox: self.custom_from_blueprint(i))
                give_hint(button_edit.butt, self, "Use the selected polyhedron as a jumping-off point to define a new model polyhedron.")
                button_remove.frame.grid(row = 1, column = 0, padx = 5, pady = (0, 5)), button_edit.frame.grid(row = 1, column = 1, padx = 5, pady = (0, 5), sticky = tk.E)
                give_hint(button_remove.butt, self, "Temporarily ignore the selected model polyhedron. It will not be evaluated.")
            frame_buttons.grid(row = 1, column = 0, sticky = tk.W)
        for item_box in (self.preview_listboxes[1], self.preview_listboxes[2], self.listbox_results):
            item_box.bind("<<ListboxSelect>>", lambda event, i = item_box: self.visualize_polyhedron(i))
        self.preview_listboxes[2].bind("<ButtonRelease-1>", lambda event: self.polyhedron_facts())
        self.preview_listboxes[2].bind("<Leave>", lambda event: self.commentline.configure(text = ""))
        centralcrit = CritSearch(self, frames_labelled[0], ATOM_KEYWORDS, DICT_SEARCH_CENTRAL_INIT, crit_type = "central_coordinates")
        ligandcrit = CritSearch(self, frames_labelled[1], ATOM_KEYWORDS, DICT_SEARCH_LIGAND_INIT, crit_type = "ligand_coordinates")
        polycrit = CritSearch(self, frames_labelled[2], POLY_KEYWORDS, DICT_SEARCH_POLY_INIT, crit_type = "poly")
        fit_checks = CheckList(self, frames_structural[0], CHECKSTUFF_FIT)
        fit_radio = RadioList(self, frames_structural[1], RADIOSTUFF_CENTERMETHODS)
        env_filters = EntryList(self, frames_labelled[4], ENTRYSTUFF_ENV_FILTERS)
        check_axis_fixed = ttk.Checkbutton(frames_labelled[5], text = "don't adjust", variable = self.var_axis_fixed)
        give_hint(check_axis_fixed, self, "If selected, Polynator will not try to optimize the axis direction you gave. In this case, make sure to enter precise coordinates.")
        give_hint(self.field_main_axis, self, "You may define a main axis to bypass Polynator's axis finding process. This may be useful if you already know the (approximate) best main axis and want to safe computation time or if you suspect that the correct axis wasn't found automatically. Enter coordinates, e. g. '1/3, 2/3, 0.5' (fractional, no unit vector required).")
        env_filters.variables[3].trace("w", lambda name, index, mode, var = env_filters.variables[3]: self.toggle_entries(var))
        self.main_contents = (centralcrit, ligandcrit, polycrit, fit_checks, fit_radio, env_filters)
        for count_crit, item_crit in enumerate(self.main_contents[:3]):
            give_hint(item_crit.combox, self, COMMENTS_CRITSEARCH[count_crit])
            frame_bottom = tk.Frame(item_crit.frame, bg = "#ffffff", padx = 1, pady = 1)
            button_delete = Buttoon(frame_bottom, text = "remove", command = item_crit.delete_search_term, width = 7)
            button_delete.frame.grid(row = 0, column = 0, ipady = 1, padx = 5, pady = (0, 5), sticky = "w")
            if count_crit < 2:
                button_radius = Buttoon(frame_bottom, text = "  set radius:", width = 16)
                field_radius = ttk.Entry(button_radius.frame, width = 6)
                button_radius.butt.bind("<ButtonRelease-1>", lambda event, i = item_crit.listbox, j = field_radius: self.add_radius(event, i, j))
                field_radius.bind("<Return>", lambda event, i = item_crit.listbox, j = field_radius: self.add_radius(event, i, j))
                button_radius.butt.configure(anchor = "w")
                button_radius.frame.grid(row = 0, column = 1, columnspan = 2, ipady = 1, pady = (0, 5), padx = 5, sticky = "w")
                field_radius.grid(row = 0, column = 0, padx = 4, sticky = "e")
                for item_widget in (field_radius, button_radius.butt):
                    give_hint(item_widget, self, "You may enter a maximal atom radius here. Press this button to assign it to the selected atoms in the list above. If you assigned radii to both central and ligand atoms, d_max is replaced by the sum of the respective radii. Tip: You can differentiate atoms using radii, but if you don't need to, it's more convenient to use d_max or cn_max instead.")
            frame_bottom.grid(row = 3, column = 0, sticky = "nsew")
        frame_buttons = tk.Frame(self.root, bg = "#ffffff")
        button_config = Buttoon(frame_buttons, width = 20, text = "Settings", command = self.configure, pady = 6)
        button_preview = Buttoon(frame_buttons, width = 20, text = "Preview", command = self.preview, pady = 6)
        button_start = Buttoon(frame_buttons, width = 20, text = "Run", command = self.run_full, pady = 6)
        button_output = Buttoon(frame_buttons, width = 20, text = "Generate output files", command = self.output, pady = 6)
        buttons_bottom = (button_config, button_preview, button_start, button_output)
        for count, item in enumerate(buttons_bottom):
            give_hint(item.butt, self, COMMENTS_BUTTONS[count])
            item.frame.grid(row = count, column = 0, pady = (7, 0), sticky = "nsew")
        self.field_main_axis.grid(row = 0, column = 0, sticky = "ew", padx = 6, pady = 6)
        check_axis_fixed.grid(row = 0, column = 3, sticky = tk.W, padx = 3, pady = 6)
        self.canvas.grid(row = 0, column = 0, padx = 5, pady = 5, sticky = "nsew")
        self.listbox_results.grid(sticky = "nsew")
        frame_output.grid(row = 0, column = 0, padx = 5, pady = 5, sticky = "nsew")
        frame_output.rowconfigure(0, weight = 1), frame_output.columnconfigure(0, weight = 1)
        for count_frame, item_frame in enumerate(frames_labelled):
            item_frame.grid(row = FRAMESTUFF_1[count_frame][1], column = FRAMESTUFF_1[count_frame][2], rowspan = FRAMESTUFF_1[count_frame][3], sticky = "nsew", padx = 4, pady= 4)
        for count_frame, item_frame in enumerate(frames_structural):
            item_frame.grid(row = FRAMESTUFF_2[count_frame][1], column = FRAMESTUFF_2[count_frame][2], columnspan = FRAMESTUFF_2[count_frame][3], sticky = FRAMESTUFF_2[count_frame][4])
        frame_buttons.columnconfigure(0, weight = 1), frame_buttons.rowconfigure((0, 1, 2, 3, 4), weight = 1)
        frame_buttons.grid(row = 1, column = 2, rowspan = 2, sticky = "nsew", padx = (4, 5), pady = 4)
        for item_frame in frames_labelled[:3]:
            item_frame.rowconfigure(2, weight = 1)
            item_frame.columnconfigure((0, 1, 2), weight = 1)
        for item_frame in frames_labelled[6:]:
            item_frame.rowconfigure(0, weight = 1)
            item_frame.columnconfigure(0, weight = 1)
        self.root.wm_protocol("WM_DELETE_WINDOW", self.write_config_files_and_close)
    
    def __str__(self):
        return f"{self.name}: {self.csm}"
    
    def process_config(self):
        existing_files = os.listdir(self.configpath)
        for item_filename in CONFIG_FILENAMES:
            if not item_filename + ".cfg" in existing_files:
                config_file = open(f"{self.configpath}/{item_filename}.cfg", "w")
                config_file.close()
            with open(f"{self.configpath}/{item_filename}.cfg") as config_file:
                self.dict_config_files[item_filename] = config_file.readlines()
            for item_line in self.dict_config_files.get(item_filename):
                if item_filename == "custom_polyhedra":
                    try:
                        blueprint = literal_eval(item_line.removesuffix("\n"))
                    except ValueError:
                        warning = "Couldn't read all custom polyhedra from configuration file. Bad string: " + item_line.removesuffix("\n")
                        self.commentline.configure(text = warning, foreground = "#ff0000")
                        self.warnings.add(warning)
                        continue
                    self.dict_custom_blueprints[blueprint[0]] = blueprint
                elif item_filename == "blocked_polyhedra": 
                    self.blocked_polyhedra.append(item_line.removesuffix("\n"))
                elif item_filename == "settings":
                    line_split = item_line.removesuffix("\n").split(";")
                    try:
                        self.dict_input[line_split[0].strip("'")] = literal_eval(line_split[1])
                    except NameError:
                        self.dict_input[line_split[0].strip("'")] = line_split[1].strip("'")
                    except SyntaxError:
                        self.dict_input[line_split[0].strip("'")] = line_split[1].strip("'")
                    except ValueError:
                        self.dict_input[line_split[0].strip("'")] = line_split[1].strip("'")
    
    def toggle_entries(self, var):
        coord_filters = self.main_contents[5]
        if var.get() != "":
            for item_field in coord_filters.fields[:3]:
                item_field.config(state = "disabled", foreground = "#cccccc")
        else:
            for item_field in coord_filters.fields[:3]:
                item_field.config(state = "normal", foreground = "#000000")
    
    def visualize_polyhedron(self, listbox):
        indices = listbox.curselection()
        if not indices:
            return
        entry = listbox.get(indices[0])
        if entry.endswith(".cif:"):
            return
        belt_dicts, ligvecs_real, atom_labels, list_elements, vecs_model, model_numbers, excentricity_vec, is_molecule, axis = (), [], [], [], [], [], Vec(0, 0, 0), False, Vec(0, 0, 1)
        parameter_bundles, var_range = {}, {}
        if not " " in entry: # model preview
            blueprint = self.dict_blueprints_by_name.get(entry)
            is_molecule = "#molecule" in blueprint[1].get("tags")
            belt_dicts = blueprint[2]
            parameter_bundles, var_range = blueprint[1].get("parameter_bundles", {}), blueprint[1].get("var_range", {})
        elif entry[:4] == "    ": # real + model
            info_tuple = self.list_output[indices[0]]
            if len(info_tuple) == 8:
                vecs_mapped = list(zip([i for j in info_tuple[4] for i in j], [i for j in info_tuple[0] for i in j], [i for j in info_tuple[3] for i in j], list(range(1, len(info_tuple[1])))))
                vecs_mapped.sort(key = lambda i: i[0])
                maps, ligvecs_real, vecs_model, model_numbers = zip(*vecs_mapped)
                atom_labels, list_elements, excentricity_vec, is_molecule, axis = info_tuple[1], info_tuple[2], info_tuple[5], info_tuple[6], info_tuple[7]
            else:
                ligvecs_real, atom_labels, list_elements = info_tuple[0], info_tuple[1], info_tuple[2]
        elif entry[-1] in ") ": # real env preview
            entry_split = entry.split(" ")
            central_name = entry_split[0][:-1]
            cif_string = "".join(entry_split[2:])[1:-1]
            for item_cif in self.cifs:
                if item_cif.filepath.replace(" ", "") == cif_string:
                    for item_env in item_cif.envs:
                        if item_env.env_fract[0][0] == central_name and item_env.env_fract[0][1] + item_env.sum_ligands == entry_split[1]:
                            ligvecs_real = item_env.ligands_around_central_atom
                            atom_labels = [i[0] for i in item_env.env_fract]
                            list_elements = [i[1] for i in item_env.env_fract]
                            break
                    break
        if self.displayed_polyhedron:
            self.displayed_polyhedron.kill()
        self.displayed_polyhedron = VisualPolyhedron(self, names = atom_labels, elements = list_elements, belt_dicts = belt_dicts, real_vecs = ligvecs_real, model_vecs = vecs_model, excentricity = excentricity_vec, is_molecule = is_molecule, model_labels = model_numbers, dict_bundles = parameter_bundles, dict_var_range = var_range, axis = axis)
    
    def polyhedron_facts(self):
        indices = self.preview_listboxes[2].curselection()
        if not indices:
            return
        poly_name = self.preview_listboxes[2].get(indices[0])
        blueprint = self.dict_blueprints_by_name.get(poly_name)
        self.commentline.configure(text = get_poly_comment(blueprint), foreground = "#000000")
    
    def custom_from_blueprint(self, listbox):
        index = listbox.curselection()[0] if listbox.curselection() else tk.END
        poly_name = listbox.get(index)
        blueprint = deepcopy(self.dict_blueprints_by_name.get(poly_name))
        self.build_custom_polyhedron(blueprint = blueprint)
    
    def add_radius(self, event, listbox, field_radius):
        tuple_indices = listbox.curselection()
        string_radius = field_radius.get()
        if not listbox.size():
            self.commentline.configure(text = "Select an atom before assigning a radius!", foreground = "#ff0000")
            return
        if string_radius == "":
            return
        for item_char in string_radius:
            if item_char not in "0123456789 .,-":
                self.commentline.configure(text = "Couldn't read the number you entered!", foreground = "#ff0000")
                return
        if not tuple_indices:
            tuple_indices = (tk.END,)
        for item_index in tuple_indices:
            split_string_entry = listbox.get(item_index).split(" ")
            if split_string_entry[0] == "add":
                self.commentline.configure(text = "Cannot add a radius to a dummy atom. Use d_max instead.", foreground = "#ff0000")
                continue
            listbox.delete(item_index)
            listbox.insert(item_index, f"{split_string_entry[0]} {split_string_entry[1]} (r = {string_radius.replace(',', '.')})")
    
    def ignore_poly(self, listbox, permanent = False):
        indices = listbox.curselection() if listbox.curselection() else (tk.END,)
        for item_index in indices:
            poly_name = listbox.get(item_index)
            if not poly_name in self.ignored_polyhedra:
                self.ignored_polyhedra.append(poly_name)
                listbox.itemconfig(item_index, foreground = "#bbbbbb")
            else:
                self.ignored_polyhedra.remove(poly_name)
                listbox.itemconfig(item_index, foreground = "#000000")
            if permanent:
                pass
    
    def ignore_envs(self, listbox):
        indices = listbox.curselection()
        for item_index in indices:
            entry = listbox.get(item_index)
            if entry[-1] == " ": # Trailing space marks coordination environments which have an unpopulated coordination number (0, 1, 19, etc.). Kind of a hack, but I like it.
                self.commentline.configure(text = "No model polyhedra for this coordination number! Construct a new one (-> settings -> add custom polyhedron) or adjust your coordination environment filters.", foreground = "#ff0000")
                continue
            entry_split = entry.split(" ")
            central_name = entry_split[0][:-1]
            cif_string = "".join(entry_split[2:])[1:-1]
            if not self.dict_ignore_envs.get(cif_string):
                self.dict_ignore_envs[cif_string] = []
            if not central_name in self.dict_ignore_envs.get(cif_string):
                self.dict_ignore_envs.get(cif_string).append(central_name)
                listbox.itemconfig(item_index, foreground = "#bbbbbb")
            else:
                self.dict_ignore_envs.get(cif_string).remove(central_name)
                listbox.itemconfig(item_index, foreground = "#000000")
    
    def gather_inputs(self):
        dict_entries = {}
        raw_axis_strings = self.field_main_axis.get().replace(",", " ").split()
        main_axis_list = []
        for item_string in raw_axis_strings:
            try:
                main_axis_list.append(parse_math(item_string))
            except ValueError:
                pass
        if len(main_axis_list) == 3:
            dict_entries["custom_axis"] = Vec(main_axis_list[0], main_axis_list[1], main_axis_list[2])
        elif main_axis_list:
            self.commentline.configure(text = "Entry for main axis invalid!", foreground = "#ff0000")
            return
        for item in self.main_contents:
            dict_entries.update(item.read())
        if isinstance(dict_entries.get("find_coordination_spheres"), str):
            sphere_integers = set()
            spheres_string_list = dict_entries.get("find_coordination_spheres").replace(",", " ").split()
            for item_sphere_string in spheres_string_list:
                try:
                    integer = int(item_sphere_string)
                    if integer < 10:
                        sphere_integers.add(integer)
                except ValueError:
                    pass
            dict_entries["find_coordination_spheres"] = tuple(sorted(sphere_integers))
        try:
            os.mkdir(self.dict_input.get("input_dir") + "/polynator_output/")
        except FileExistsError:
            pass
        except FileNotFoundError:
            self.get_cifs()
            self.gather_inputs()
        dict_entries["axis_fixed"] = self.var_axis_fixed.get()
        for item_key in ("d_min", "d_max", "cn_max", "find_coordination_spheres"):
            if not dict_entries.get(item_key):
                dict_entries[item_key] = DICT_INPUT_DEFAULT.get(item_key)
        for count_key, item_key in enumerate((("d_min", 10, "0.1"), ("d_max", 10, "10"), ("cn_max", max(self.dict_blueprints_by_cn), str(max(self.dict_blueprints_by_cn))))):
            if dict_entries.get(item_key[0]) > item_key[1]:
                dict_entries[item_key[0]] = 10
                self.main_contents[5].fields[count_key].delete(0, tk.END)
                self.main_contents[5].fields[count_key].insert(0, item_key[2])
        for item_entry in dict_entries:
            if item_entry:
                self.dict_input[item_entry] = dict_entries.get(item_entry)
    
    def build_custom_polyhedron(self, blueprint = ()):
        CustomPolyhedronWindow(self, blueprint = blueprint)
    
    def configure(self):
        ConfigWindow(self)
    
    def get_cifs(self):
        filepaths = tkfd.askopenfilenames(title = "Select input files", filetypes = (("cif files", "*.cif"), ("all files", "*.*")))
        self.dict_input["input_dir"] = filepaths[0][:filepaths[0].rindex("/")]
        self.dict_input["cif_filepaths"] = [i.removeprefix(self.dict_input.get("input_dir") + "/") for i in filepaths]
        for item_listbox in self.preview_listboxes + [self.listbox_results]:
            item_listbox.delete(0, last = tk.END)
        for item_cif_path in self.dict_input.get("cif_filepaths"):
            self.preview_listboxes[0].insert(tk.END, item_cif_path.removeprefix(self.dict_input.get("input_dir") + "/"))
    
    def preview(self):
        self.preview_listboxes[1].delete(0, last = tk.END)
        self.preview_listboxes[2].delete(0, last = tk.END)
        self.gather_inputs()
        self.cifs = []
        self.dict_blueprints_by_name = {}
        for item_cif_path_run in self.dict_input.get("cif_filepaths"):
            current_cif = CifFile(self, item_cif_path_run)
            current_cif.get_coordination_environments()
            ignored_envs = self.dict_ignore_envs.get(item_cif_path_run.replace(" ", ""))
            for item_env_run in current_cif.list_all_coordination_envs:
                current_env = CoordinationEnvironment(current_cif, item_env_run)
                current_env.get_matching_blueprints()
                current_cif.envs.append(current_env)
                for item_blueprint in self.dict_blueprints_by_cn.get(current_env.coordination_number, []):
                    if item_blueprint[0] not in self.blocked_polyhedra:
                        self.preview_listboxes[1].insert(tk.END, f"{current_env.env_fract[0][0]}: {current_env.env_fract[0][1]}{current_env.sum_ligands} ({item_cif_path_run})")
                        break
                else: 
                    self.preview_listboxes[1].insert(tk.END, f"{current_env.env_fract[0][0]}: {current_env.env_fract[0][1]}{current_env.sum_ligands} ({item_cif_path_run}) ") # the trailing space prevents the color from being manually altered.
                    self.preview_listboxes[1].itemconfigure(tk.END, foreground = "#bbbbbb")
                if ignored_envs:
                    if current_env.env_fract[0][0] in ignored_envs:
                        self.preview_listboxes[1].itemconfigure(tk.END, foreground = "#bbbbbb")
                for item_blueprint in current_env.matching_blueprints:
                    self.dict_blueprints_by_name[item_blueprint[0]] = item_blueprint
            self.cifs.append(current_cif)
        for item_poly in self.dict_blueprints_by_name:
            self.preview_listboxes[2].insert(tk.END, item_poly)
            if item_poly in self.ignored_polyhedra:
                self.preview_listboxes[2].itemconfigure(tk.END, foreground = "#bbbbbb")
    
    def run_full(self):
        dict_all_poly_parameters = {}
        self.partable_line_data, self.set_partable_head, self.warnings = [], set(), set()
        self.listbox_results.delete(0, last = tk.END), self.list_output.clear()
        self.preview()
        for current_cif in self.cifs:
            ignored_envs = self.dict_ignore_envs.get(current_cif.filepath.replace(" ", ""))
            for current_env in current_cif.envs:
                if ignored_envs:
                    if current_env.env_fract[0][0] in ignored_envs:
                        continue
                best_poly_in_env = Polyhedron("placeholder", [">", {}, [], {}])
                partable_rows_env = []
                for item_blueprint in current_env.matching_blueprints:
                    if item_blueprint[0] in self.ignored_polyhedra:
                        continue
                    if item_blueprint[1].get("aristohedra") and self.dict_input.get("skip_redundant_polyhedra"):
                        if evaluate_redundancy(item_blueprint[0], current_env.polyhedra):
                            continue
                    current_poly = run_blueprint(item_blueprint, current_env, self.dict_input)
                    if current_poly.csm_taxed < best_poly_in_env.csm_taxed:
                        best_poly_in_env = current_poly.copy()
                        print("best matching polyhedron so far! csm:", current_poly.csm, "| main_axis:", str(current_poly.main_axis))
                    for item_type in current_poly.dict_parameters:
                        if current_poly.dict_parameters.get(item_type):
                            if not item_type in dict_all_poly_parameters:
                                dict_all_poly_parameters[item_type] = set()
                            for item_param in current_poly.dict_parameters.get(item_type):
                                dict_all_poly_parameters.get(item_type).add(item_param)
                    current_env.polyhedra.append(current_poly.copy())
                    if not self.dict_input.get("output_only_best_polyhedra"):
                        partable_rows_env.append((current_cif.filepath.removesuffix(".cif"), current_env.env_fract[0][0], current_env.sum_ligands, current_poly.name, current_poly.csm, current_poly.avg_delta, current_poly.avg_delta_radial, current_poly.avg_delta_angular, current_poly.excentricity_vec_cart.length(), current_poly.parameter_count, current_poly.dict_parameters))
                if current_env.matching_blueprints and self.dict_input.get("output_only_best_polyhedra"):
                    partable_rows_env.append((current_cif.filepath.removesuffix(".cif"), current_env.env_fract[0][0], current_env.sum_ligands, best_poly_in_env.name, best_poly_in_env.csm, best_poly_in_env.avg_delta, best_poly_in_env.avg_delta_radial, best_poly_in_env.avg_delta_angular, best_poly_in_env.excentricity_vec_cart.length(), best_poly_in_env.parameter_count, best_poly_in_env.dict_parameters))
                partable_rows_env.sort(key = lambda i: i[4])
                self.partable_line_data += partable_rows_env
        for item_type in dict_all_poly_parameters:
            for item_param in dict_all_poly_parameters.get(item_type):
                self.set_partable_head.add(item_param)
        self.n_evaluated_polyhedra = 0
        for item_cif in self.cifs:
            self.listbox_results.insert(tk.END, f"{item_cif.filepath}:"), self.list_output.append(())
            for item_env in item_cif.envs:
                item_env.polyhedra.sort(key = lambda i: i.csm_taxed)
                if self.dict_input.get("output_only_best_polyhedra") and item_env.polyhedra:
                    self.n_evaluated_polyhedra += 1
                    self.fill_results(item_env.polyhedra[0], prefix = f"    {item_env.env_fract[0][0]}: ")
                elif item_env.polyhedra:
                    self.listbox_results.insert(tk.END, f"    {item_env.env_fract[0][0]}:")
                    self.list_output.append((item_env.ligands_around_central_atom, [i[0] for i in item_env.env_fract], [i[1] for i in item_env.env_fract]))
                    branches = [[(i[0].content, i[1]) for i in j.get_branches_sorted() if i[0].content] for j in item_env.symmetree_root.children]
                    branches.sort(key = lambda j: min([i[0].csm for i in j]))
                    for item_branch in branches:
                        for item_poly_depth_tuple in item_branch:
                            if item_poly_depth_tuple[0].csm_taxed < self.dict_input.get("csm_cap"):
                                self.n_evaluated_polyhedra += 1
                                self.fill_results(item_poly_depth_tuple[0], prefix = "    " * (item_poly_depth_tuple[1] + 1))
        give_hint(self.listbox_results, self, "The continuous symmetry measure (csm, see Pinsky & Avnir, Inorg. Chem., 1998) is a scale-indepent metric for the similarity between the shapes formed by real atoms and model vertices. As a rough guideline - csm < 3: (almost) undistorted; 3 < csm < 9: slightly distorted; 9 < csm < 15: moderately distorted; 15 < csm < 27: heavily distorted; csm > 27: probably not a good fit.")
        if self.dict_input.get("automatic_output"):
            self.output()
    
    def output(self):
        if self.dict_input.get("output_table"):
            self.write_partable()
        for item_cif in self.cifs:
            for item_env in item_cif.envs:
                edges, faces = find_edges(item_env.ligands_around_centroid)
                if self.dict_input.get("output_only_best_polyhedra") and item_env.polyhedra:
                    item_env.polyhedra[0].write_output()
                else:
                    for item_poly in item_env.polyhedra:
                        if item_poly.csm_taxed < self.dict_input.get("csm_cap"):
                            item_poly.write_output()
        if self.dict_input.get("output_log"):
            self.write_logfile()
        output_factor = sum([1 for i in ("output_minicif_cart", "output_minicif_fract", "output_details") if self.dict_input.get(i)])
        output_bonus = sum([1 for i in ("output_table", "output_log") if self.dict_input.get(i)])
        n_output = self.n_evaluated_polyhedra * output_factor + output_bonus
        self.commentline.configure(text = f"{n_output} output file{'s' if n_output != 1 else ''} written to '{self.dict_input.get('input_dir')}/polynator_output/'.")
                                
    def fill_results(self, poly, prefix = "        "):
        name = poly.name
        if len(name) > 40:
            for item_abbreviation in DICT_ABBREVIATIONS:
                name = name.replace(item_abbreviation, DICT_ABBREVIATIONS.get(item_abbreviation))
        csm_adjusted = poly.csm_taxed if poly.csm_taxed < 40 else 40
        hex_color = "#" + 6 * hex(int(csm_adjusted / 4))[-1]
        self.listbox_results.insert(tk.END, f"{prefix}{name} ({poly.dict_info.get('point_group')}, {poly.parameter_count} var{'' if poly.parameter_count == 1 else 's'}): cs{'' if '#free' in poly.dict_info.get('tags') else 'h'}m = {stround(poly.csm, 3)}")
        self.listbox_results.itemconfig(tk.END, foreground = hex_color)
        self.list_output.append((poly.belts_real, [i[0] for i in poly.env.env_fract], [i[1] for i in poly.env.env_fract], poly.belts_model, poly.map_belts, poly.excentricity_vec_cart, "#molecule" in poly.dict_info.get("tags"), poly.main_axis))
    
    def write_partable(self):
        input_dir = self.dict_input.get("input_dir")
        round_to = 5
        list_partable_head = list(self.set_partable_head)
        list_partable_head.sort()
        partable = get_file(f"{input_dir}/polynator_output/", "data_table", ".csv", overwrite = self.dict_input.get("overwrite_output"))
        try:
            partable.write("filename,central_atom,ligands,polyhedron,csm/percent,avg_delta/pm,avg_rad_delta/pm,avg_ang_delta/deg,excentricity_length/pm,variable_count")
        except PermissionError:
            warning = "Cannot write on data_table file, it's probably opened in another window."
            self.commentline.configure(text = warning, foreground = "#ff0000")
            self.warnings.add(warning)
        for item_free_parameter in list_partable_head:
            if "phi" in item_free_parameter:
                partable.write("," + item_free_parameter + "/rad")
            else:
                partable.write("," + item_free_parameter + "/Angstrom")
        partable.write("\n")
        for item_line in self.partable_line_data:
            partable.write(f"{item_line[0]},{item_line[1].replace(',', '/')},{item_line[2].replace(',', '/')},{item_line[3]},{stround(item_line[4], round_to - 2)},{stround(100 * item_line[5], round_to - 2)},{stround(100 * item_line[6], round_to - 2)},{stround(item_line[7], round_to)},{stround(100 * item_line[8], round_to - 2)},{item_line[9]}")
            for item_free_parameter in list_partable_head:
                parameter_found = False
                for item_dict_type in item_line[10].values():
                    if item_dict_type.get(item_free_parameter):
                        partable.write(f",{stround(item_dict_type.get(item_free_parameter), round_to)}")
                        parameter_found = True
                        break
                if not parameter_found:
                    partable.write(",0")
            partable.write("\n")
        partable.flush(), partable.close()
    
    def write_logfile(self):
        logfile = get_file(f"{self.dict_input.get('input_dir')}/polynator_output/", "report", ".log", overwrite = self.dict_input.get("overwrite_output"))
        warnings = ""
        for item_warning in self.warnings:
            warnings += f"Warning: {item_warning}\n"
        for item_cif in self.cifs:
            for item_warning in item_cif.warnings:
                warnings += f"Warning: {item_warning}\n"
            for item_env in item_cif.envs:
                for item_warning in item_env.warnings:
                    warnings += f"Warning: {item_warning}\n"
                for item_poly in item_env.polyhedra:
                    for item_warning in item_poly.warnings:
                        warnings += f"Warning: {item_warning}\n"
        logfile.write(warnings + "Run successful!")
        logfile.flush(), logfile.close()
    
    def write_config_files_and_close(self):
        for item_filename in self.dict_config_files:
            try:
                with open(f"{item_filename}.cfg", "r+") as config_file:
                    config_file.truncate(0)
                    config_file.seek(0)
                    if item_filename == "custom_polyhedra":
                        for item_name in self.dict_custom_blueprints:
                            config_file.write(f"{self.dict_custom_blueprints.get(item_name)}\n")
                    if item_filename == "blocked_polyhedra":
                        for item_erased_poly in self.blocked_polyhedra:
                            config_file.write(f"{item_erased_poly}\n")
                    if item_filename == "settings":
                        for item_entry in self.dict_input:
                            if item_entry in CONFIGURABLES:
                                if isinstance(self.dict_input.get(item_entry), str):
                                    config_file.write(f"'{item_entry}';'{self.dict_input.get(item_entry)}'\n")
                                else:
                                    config_file.write(f"'{item_entry}';{self.dict_input.get(item_entry)}\n")
            except BaseException:
                pass
        self.root.destroy()

# three classes in hierarchical order: Each CifFile (= crystal structure) may have several CoordinationEnvironments, to each of which several Polyhedra may be fitted.

class CifFile:
    def __init__(self, main_object, cif_filepath):
        self.main = main_object
        self.dict_input = main_object.dict_input
        self.filepath = cif_filepath
        self.warnings = set()
        self.envs = []
        self.list_all_coordination_envs = []
        self.symop_strings = []
        self.get_file_data()
        
    def get_file_data(self):
        CHARS_XYZ = "xyz12345678+-/,"
        dict_cell_parameter_entries_cif = {"_cell_length_a": "a", "_cell_length_b": "b", "_cell_length_c": "c", "_cell_angle_alpha": "alpha", "_cell_angle_beta": "beta", "_cell_angle_gamma": "gamma"}
        dict_coordtable = {"_atom_site_label": "name", "_atom_site_type_symbol": "element", "_atom_site_fract_x": "x", "_atom_site_fract_y": "y", "_atom_site_fract_z": "z"}
        try:
            with open(f"{self.dict_input.get('input_dir')}/{self.filepath}") as cif:
                cif_lines = cif.readlines()
        except FileNotFoundError:
            self.main.commentline.configure(text = f"Couldn't find input file '{self.filepath}'. Is it still there?", foreground = "#ff0000")
            self.main.get_cifs()
            self.main.preview_listboxes[0].delete(0, tk.END)
        symop, cif_valid = False, True
        self.symop_strings = []
        dict_cell_parameters = {}
        dict_table_indices = {}
        readlist_atoms = []
        count_column_coordtable = 0
        for count_line, item_line in enumerate(cif_lines):
            item_line = item_line.replace("  ", " ").strip(" ")
            if "_space_group_IT_number" in item_line or "_symmetry_Int_Tables_number" in item_line:
                self.n_spgr = int(item_line.split()[1])
                continue
            elif "_space_group_symop_operation_xyz" in item_line or "symmetry_equiv_pos_as_xyz" in item_line:
                symop = True
                continue
            elif self.symop_strings and ("loop_" in item_line or item_line.replace(" ", "") == "\n" or item_line.replace(" ", "") == ""):
                symop = False
                if count_column_coordtable or not cif_valid:
                    break
                continue
            elif symop and "," in item_line: #find symmetry operations
                indicator = "'" if "'" in item_line else " "
                symop_builder = ""
                symchar = False
                for item_char_symop in item_line:
                    if indicator in item_char_symop:
                        symchar = not symchar
                    elif symchar and item_char_symop in CHARS_XYZ:
                        symop_builder += item_char_symop
                if symop_builder not in self.symop_strings:
                    self.symop_strings.append(symop_builder)
                continue
            for item_key in dict_cell_parameter_entries_cif: #find cell parameters
                if item_key in item_line:
                    precursor = ""
                    for item_char_axes in item_line:
                        if item_char_axes == "(":
                            break
                        elif item_char_axes in CHARS_FLOAT:
                            precursor += item_char_axes
                    try:
                        cell_parameter = float(precursor)
                        dict_cell_parameters[dict_cell_parameter_entries_cif.get(item_key)] = cell_parameter
                    except ValueError:
                        self.warnings.add(f"Could not read cell parameter data in {self.filepath}. Skipped this file")
                        cif_valid = False
            for item_key in dict_coordtable:
                if item_key in item_line:
                    dict_table_indices[dict_coordtable.get(item_key)] = count_column_coordtable
            if 0 < len(dict_table_indices) < 5:
                count_column_coordtable += 1
            elif count_column_coordtable and not "_" in item_line:
                line_coordtab = item_line.split()
                name_atom = line_coordtab[dict_table_indices.get("name")]
                element_pre = line_coordtab[dict_table_indices.get("element")]
                element_atom = ""
                for item_char in element_pre:
                    if item_char.isalpha():
                        element_atom += item_char
                for item_char in line_coordtab[dict_table_indices.get("x")] + line_coordtab[dict_table_indices.get("y")] + line_coordtab[dict_table_indices.get("z")]:
                    if not item_char in CHARS_FLOAT + "()":
                        cif_valid = False
                        self.warnings.add(f"Atom coordinates in {self.filepath} contain a character '{item_char}' that shouldn't be there. Skipped this file.")
                        break
                if cif_valid:
                    x_init, y_init, z_init = translate_to_unit_cell(parse_math(build_coordinate(line_coordtab[dict_table_indices.get("x")]))), translate_to_unit_cell(parse_math(build_coordinate(line_coordtab[dict_table_indices.get("y")]))), translate_to_unit_cell(parse_math(build_coordinate(line_coordtab[dict_table_indices.get("z")])))
                    data_set_atom = (name_atom, element_atom, x_init, y_init, z_init)
                    readlist_atoms.append(data_set_atom)
        self.cell_parameters = (dict_cell_parameters.get("a"), dict_cell_parameters.get("b"), dict_cell_parameters.get("c"), rad(dict_cell_parameters.get("alpha")), rad(dict_cell_parameters.get("beta")), rad(dict_cell_parameters.get("gamma")))
        self.MATRIX_TO_CART = get_transformation_matrix_to_cart(self.cell_parameters)
        self.MATRIX_TO_FRACT = get_transformation_matrix_to_fract(self.cell_parameters)
        print(self.filepath, "| transformation_matrix_cart:", self.MATRIX_TO_CART)
        self.dict_central = atom_search(readlist_atoms, self.dict_input.get("central_atom_criteria"), self.dict_input.get("central_coordinates"))
        self.dict_ligand = atom_search(readlist_atoms, self.dict_input.get("ligand_criteria"), self.dict_input.get("ligand_coordinates"))
        for count_dict in range(2):
            dict_current = self.dict_central if count_dict == 0 else self.dict_ligand
            dict_copy = deepcopy(dict_current)
            if len(dict_copy) < 2:
                continue
            for count_atom_1, item_atom_1 in enumerate(tuple(dict_copy.keys())[:-1]):
                for count_atom_2, item_atom_2 in enumerate(tuple(dict_copy.keys())[count_atom_1 + 1:]):
                    if (item_atom_1[2], item_atom_1[3], item_atom_1[4]) == (item_atom_2[2], item_atom_2[3], item_atom_2[4]):
                        try:
                            radius = dict_current.pop(item_atom_1)
                            dict_current.pop(item_atom_2)
                            dict_current[(f"({item_atom_1[0]},{item_atom_2[0]})", f"({item_atom_1[1]},{item_atom_2[1]})", item_atom_1[2], item_atom_1[3], item_atom_1[4])] = radius
                        except KeyError:
                            pass

    def get_coordination_environments(self):
        EXTENSION_VECS_FRACT = (Vec(1, 0, 0), Vec(0, 1, 0), Vec(0, 0, 1), Vec(1, 1, 0), Vec(1, 0, 1), Vec(0, 1, 1), Vec(1, -1, 0), Vec(1, 0, -1), Vec(0, 1, -1), Vec(1, 1, 1), Vec(1, 1, -1), Vec(1, -1, 1), Vec(-1, 1, 1))
        smallest_cell_size = min([i.coordinate_transformation(self.MATRIX_TO_CART).length() for i in EXTENSION_VECS_FRACT])
        required_size = 4 * sqrt(max(self.dict_input.get("find_coordination_spheres"))) if self.dict_input.get("find_coordination_spheres") else self.dict_input.get("d_max")
        range_cell_extension = range(int(-1 - required_size // smallest_cell_size), int(2 + required_size // smallest_cell_size))
        d_min = 0.01 if self.dict_input.get("find_coordination_spheres") else self.dict_input.get("d_min")
        for count_cen, item_cen in enumerate(self.dict_input.get("central_coordinates")):
            data_set_atom = ("Aux" + str(count_cen), "Cen", item_cen[0], item_cen[1], item_cen[2])
            self.dict_central[data_set_atom] = None
        for count_lig, item_lig in enumerate(self.dict_input.get("ligand_coordinates")):
            data_set_atom = ("Aux" + str(count_lig), "Lig", item_lig[0], item_lig[1], item_lig[2])
            self.dict_ligand[data_set_atom] = None
        for item_central in self.dict_central:
            radius_central = self.dict_central.get(item_central)
            list_atoms = []
            central_vec_fract = Vec(item_central[2], item_central[3], item_central[4])
            central_vec_cart = central_vec_fract.coordinate_transformation(self.MATRIX_TO_CART)
            central_atom = (item_central[0], item_central[1], central_vec_fract, 0)
            list_atoms.append(central_atom)
            for item_ligand in self.dict_ligand:
                radius_ligand = self.dict_ligand.get(item_ligand)
                d_max = inf if self.dict_input.get("find_coordination_spheres") else self.dict_input.get("d_max")
                try:
                    if radius_central or radius_ligand:
                        if radius_central and not radius_ligand:
                            radius_ligand = 0
                        elif radius_ligand and not radius_central:
                            radius_central = 0
                        d_max = min(d_max, parse_math(radius_central) + parse_math(radius_ligand))
                except TypeError:
                    pass
                if item_ligand[1] == "Lig":
                    auxlig_vec = Vec(item_ligand[2], item_ligand[3], item_ligand[4])
                    list_atoms.append((item_ligand[0], "Lig", auxlig_vec, (central_vec_fract - auxlig_vec).coordinate_transformation(self.MATRIX_TO_CART).length()))
                else:
                    coord_log_ligands_1, coord_log_ligands_2 = set(), set()
                    for item_sym in parse_symops(self.symop_strings, item_ligand, warnings = self.warnings):
                        try:
                            for count in range(3):
                                item_sym[count] = translate_to_unit_cell(item_sym[count])
                        except IndexError:
                            self.warnings.add(f"Couldn't read symmetry terms given in {self.filepath}. Skipped this .cif file.")
                            return
                        ligand_signature = (round(item_sym[0], 2), round(item_sym[1], 2), round(item_sym[2], 2))
                        if ligand_signature in coord_log_ligands_1: # filter out duplicates
                            continue
                        coord_log_ligands_1.add(ligand_signature)
                        for item_ligand_vecs in generate_translated_atoms(d_max, central_vec_fract, item_sym, range_cell_extension, self.MATRIX_TO_CART):
                            d = central_vec_cart.distance_points(item_ligand_vecs[0])
                            if not d_min <= d <= d_max:
                                continue
                            ligand_signature_translated = (round(item_ligand_vecs[1].x, 2), round(item_ligand_vecs[1].y, 2), round(item_ligand_vecs[1].z, 2))
                            if ligand_signature_translated in coord_log_ligands_2:
                                continue
                            coord_log_ligands_2.add(ligand_signature_translated)
                            list_atoms.append((item_ligand[0], item_ligand[1], item_ligand_vecs[1], d))
            list_atoms.sort(key = lambda i: i[-1])
            if self.dict_input.get("find_coordination_spheres"):
                coordination_spheres = find_coordination_spheres(list_atoms, self.dict_input.get("find_coordination_spheres"), max_size = max(self.main.dict_blueprints_by_cn))
                if coordination_spheres:
                    for item_sphere in coordination_spheres:
                        self.list_all_coordination_envs.append(deepcopy(item_sphere))
                else:
                    self.warnings.add(f"Couldn't find the requested coordination spheres for atom {list_atoms[0][0]} ({self.filepath}). Try defining them manually using d_min, d_max, cn_max and/or atomic radii.")
            else:
                if len(list_atoms) - 1 > self.dict_input.get("cn_max"):
                    list_atoms = list_atoms[:self.dict_input.get("cn_max") + 1]
                self.list_all_coordination_envs.append(deepcopy(list_atoms))


class CoordinationEnvironment:
    def __init__(self, cif_object, list_coordination_environment_fract):
        self.cif = cif_object
        self.dict_input = cif_object.dict_input
        self.env_fract = list_coordination_environment_fract
        self.central_atom_vec_cart = list_coordination_environment_fract[0][2].coordinate_transformation(self.cif.MATRIX_TO_CART)
        self.ligand_vecs_cart = [i[2].coordinate_transformation(self.cif.MATRIX_TO_CART) for i in list_coordination_environment_fract[1:]]
        self.coordination_number = len(list_coordination_environment_fract) - 1
        self.warnings = set()
        self.matching_blueprints = []
        self.polyhedra = []
        self.symmetree_root = TreeNode(None, "root", [])
        self.dict_symmetree_nodes =  {"root": self.symmetree_root}
        self.atom_name_log, ligand_count_dict = [], {}
        if self.env_fract[0][0] not in self.atom_name_log:
            self.atom_name_log.append(self.env_fract[0][0])
        for item_ligand in self.env_fract[1:]:
            if item_ligand[1] not in ligand_count_dict:
                ligand_count_dict[item_ligand[1]] = 1
            else:
                ligand_count_dict[item_ligand[1]] += 1
            if item_ligand[0] not in self.atom_name_log:
                self.atom_name_log.append(item_ligand[0])
        self.sum_ligands = ""
        for item_ligdict in ligand_count_dict:
            coefficient = ligand_count_dict.get(item_ligdict)
            self.sum_ligands += item_ligdict if not "/" in item_ligdict else f"({item_ligdict.replace('/', ',')})"
            self.sum_ligands += str(coefficient) if coefficient != 1 else ""
        self.centroid_cart = get_centroid(self.ligand_vecs_cart)
        self.centroid_fract = self.centroid_cart.coordinate_transformation(self.cif.MATRIX_TO_FRACT)
        self.vec_centroid_to_central_atom = self.centroid_cart - self.central_atom_vec_cart
        self.ligands_around_central_atom = [i - self.central_atom_vec_cart for i in self.ligand_vecs_cart]
        self.ligands_around_centroid = [i - self.centroid_cart for i in self.ligand_vecs_cart]
        try:
            self.ligand_sphere = [i.unit() for i in self.ligands_around_centroid]
        except ZeroDivisionError:
            self.ligand_sphere = self.ligands_around_centroid

    def get_matching_blueprints(self):
        if self.coordination_number == 0:
            self.warnings.add(f"No atoms found in the coordination environment of {self.env_fract[0][0]}! Skipped it. Check your coordination environment filters and ligand criteria.")
            return
        if not self.coordination_number in self.cif.main.dict_blueprints_by_cn:
            self.warnings.add(f"No polyhedron found for the coordination environment of {self.env_fract[0][0]} (coordination number = {self.coordination_number}, file = {self.cif.filepath})! Skipped it. Check your coordination environment filters and ligand criteria.")
            return
        candidate_blueprints = [i for i in self.cif.main.dict_blueprints_by_cn.get(self.coordination_number, []) if not i[0] in self.cif.main.blocked_polyhedra]
        self.matching_blueprints = poly_search(candidate_blueprints, self.dict_input.get("poly_criteria"), self.dict_input.get("var_max"))

class Polyhedron:
    def __init__(self, env_object, polyhedron_blueprint, warnings = set(), dict_parameters = {}, main_axis = Vec(0, 0, 1), belts_model = [], belts_real = [], map_belts = [], csm = inf, csm_taxed = inf, excentricity_vec = None, avg_delta = 0, avg_delta_angular = 0, avg_delta_radial = 0):
        self.csm, self.csm_taxed = csm, csm_taxed
        if env_object == "placeholder":
            return
        self.blueprint = polyhedron_blueprint
        self.name = polyhedron_blueprint[0]
        self.dict_info = polyhedron_blueprint[1]
        self.belt_dicts = deepcopy(polyhedron_blueprint[2])
        self.dict_bundles = self.dict_info.get("parameter_bundles", {})
        self.belts_real = belts_real
        self.map_belts = map_belts
        self.env = env_object
        self.dict_input = env_object.dict_input
        self.warnings = warnings
        self.belt_intensities = []
        self.assemble_belt_intensities()
        self.dict_bound_parameters = {}
        if self.dict_bundles:
            self.assemble_bound_parameters()
        self.dict_parameters = dict_parameters if dict_parameters else assemble_initial_parameters(self.belt_dicts, self.dict_bundles, self.dict_info.get("var_range"), self.dict_bound_parameters)
        self.parameter_count = self.dict_info.get("parameter_count")
        self.is_molecule = ("#molecule" in self.dict_info.get("tags") or self.dict_input.get("center_method") == "force_central_atom") and self.dict_input.get("center_method") != "force_centroid"
        self.axis_is_fixed = self.dict_input.get("axis_fixed") or "#axis_fixed" in self.dict_info.get("tags")
        self.point_group_hm = self.dict_info.get("point_group")
        self.point_group_sch = DICT_POINT_GROUP_CONVERSION.get(self.dict_info.get("point_group"), "?")
        self.assemble_factor_sets()
        self.main_axis = main_axis
        self.optimization_progress = 0
        self.belts_model = belts_model
        self.prototype = deepcopy(belts_model)
        if excentricity_vec:
            self.excentricity_vec_cart = excentricity_vec
        elif self.is_molecule:
            self.excentricity_vec_cart = Vec(0, 0, 0)
        else:
            self.excentricity_vec_cart = self.env.vec_centroid_to_central_atom
        self.avg_delta, self.avg_delta_angular, self.avg_delta_radial = avg_delta, avg_delta_angular, avg_delta_radial
        if "phi*" in self.dict_parameters.get("phi", {}):
            self.assemble_phi_star_ratios()
        else:
            self.dict_phi_star_ratios = {}
        print(self.dict_parameters)
    
    def assemble_phi_star_ratios(self):
        dict_aux = {}
        for item_belt_dict in self.belt_dicts:
            phi_param = item_belt_dict.get("phi")
            if phi_param and phi_param not in dict_aux:
                dict_aux[phi_param] = 0
            dict_aux[phi_param] += item_belt_dict.get("v") - len(item_belt_dict.get("erase", ()))
        self.dict_phi_star_ratios = {}
        for item_param in dict_aux:
            self.dict_phi_star_ratios[item_param] = dict_aux.get(item_param) / dict_aux.get("phi*")
    
    def assemble_bound_parameters(self):
        for item_var in self.dict_bundles:
            dict_bundle = self.dict_bundles.get(item_var)
            for item_type in ALL_PARAMETER_TYPES:
                for item_belt_dict in self.belt_dicts:
                    for count_plus in range(100):
                        full_type = item_type + count_plus * "+"
                        if not item_belt_dict.get(full_type):
                            break
                        if item_belt_dict.get(full_type) in dict_bundle:
                            if not full_type in self.dict_bound_parameters:
                                self.dict_bound_parameters[full_type] = {}
                            parameter = item_belt_dict["$" + full_type] = item_belt_dict.pop(full_type) # limit refinement of bound parameters to the adjust_bundles() method.
                            self.dict_bound_parameters.get(full_type)[parameter] = dict_bundle.get(parameter)
    
    def assemble_belt_intensities(self):
        self.belt_intensities = []
        for item_belt_dict in self.belt_dicts:
            dict_types = {}
            for item_base_type in ("phi_mod", "xy_mod", "z_mod"):
                for count_plus in range(100):
                    full_type = item_base_type + count_plus * "+"
                    if full_type not in item_belt_dict:
                        break
                    vec_ints = []
                    offset, frequency, v = item_belt_dict.get(full_type + "_off"), item_belt_dict.get(full_type + "_frq"), item_belt_dict.get("v")
                    w = [i for i in range(v) if not i in item_belt_dict.get("erase", ())]
                    for item_w in w:
                        if full_type == "phi_mod":
                            vec_ints.append(sin(2 * pi * (item_w - offset) * frequency / v))
                        else:
                            vec_ints.append(cos(2 * pi * (item_w - offset) * frequency / v))
                    dict_types[full_type] = vec_ints
            self.belt_intensities.append(dict_types)
    
    def assemble_factor_sets(self): # 'Simple' parameters with the same _factors for every belt they appear in allow for simple calculations without iteration. This basically checks for that condition.
        self.dict_factor_sets = {}
        for item_type in self.dict_parameters:
            if item_type in ("phi", "parameter_bundles") or item_type.startswith("phi_mod"):
                continue
            for item_param in self.dict_parameters.get(item_type):
                self.dict_factor_sets[item_param] = set()
            if item_type in ("xy", "z", "len"):
                for item_belt_dict in self.belt_dicts:
                    parameter = item_belt_dict.get(item_type)
                    if not parameter:
                        continue
                    if item_type == "len":
                        factor = round(sqrt(item_belt_dict.get("xy_factor") ** 2 + item_belt_dict.get("z_factor") ** 2), 12)
                    else:
                        factor = abs(item_belt_dict.get(item_type + "_factor"))
                    self.dict_factor_sets.get(parameter).add(factor)
            elif item_type.startswith("xy_mod") or item_type.startswith("z_mod"):
                for count_belt, item_belt_intensity_dict in enumerate(self.belt_intensities):
                    belt_dict = self.belt_dicts[count_belt]
                    parameter = belt_dict.get(item_type)
                    if not parameter: # This occurs if the $parameter is part of a bundle, in which case it need not be included in the factor set, as full vector iteration is required anyway.
                        continue
                    for item_vec_factor in item_belt_intensity_dict.get(item_type):
                        self.dict_factor_sets.get(parameter).add(abs(item_vec_factor))
    
    def evaluate(self):
        if self.dict_input.get("axis_fixed") and self.dict_input.get("custom_axis"):
            self.main_axis = self.dict_input.get("custom_axis").coordinate_transformation(self.env.cif.MATRIX_TO_CART).unit()
        else:
            self.find_initial_main_axis()
        self.match_real_vecs_to_model()
        for item_iterstep in range(3):
            if not self.axis_is_fixed:
                self.wrap_kabsch()
            elif not self.dict_parameters.get("phi"):
                print("main_axis preserved")
                self.rotate_belts_fixed_axis()
            if len(self.dict_parameters.get("phi")) > 1:
                self.adjust_phi()
            if self.dict_parameters.get("phi_mod"):
                self.adjust_phi_mod()
            if self.dict_parameters.get("len"):
                self.adjust_len()
            if self.dict_parameters.get("xy") or self.dict_parameters.get("z"):
                self.adjust_xyz()
            self.adjust_xyz_mod()
            if self.dict_parameters.get("parameter_bundles"):
                self.adjust_bundles()
            self.optimization_progress += 2
        self.polish()
        self.get_csm()
    
    def find_initial_main_axis(self):
        if len(self.belts_real) == 1 and not self.is_molecule:
            if len(self.belts_real[0]) == 2:
                try:
                    self.main_axis = get_centroid(self.belts_real[0]).unit()
                except ZeroDivisionError:
                    self.main_axis = Vec(5, 7, 11).cross(self.belts_real[0][0]).unit() # Any vector normal to any of the ligands is fine in this case (equidistant linear coordination)
            else:
                self.main_axis = covariance_eigen(self.belts_real[0])[0][1].unit()
        else:
            list_centroids = []
            for count_belt, item_belt in enumerate(self.belts_real):
                belt_centroid = get_centroid(item_belt)
                list_centroids.append(belt_centroid)
            if self.is_molecule and len(self.belts_real) == 1:
                list_centroids.append(Vec(0, 0, 0))
            if len(self.belts_real) == 2:
                if list_centroids[0].distance_points(list_centroids[1]) < 0.01:
                    self.main_axis = covariance_eigen(self.belts_real[0] + self.belts_real[1])[0][1].unit()
                    return
            self.main_axis = covariance_eigen(list_centroids)[-1][1].unit()
            if self.main_axis.dot(list_centroids[0]) < 0:
                self.main_axis = self.main_axis.scale(-1)
    
    def match_real_vecs_to_model(self): # This method first rotates ("rocks") the model polyhedron from the z axis to the main axis, then pairs up real and model vertices within each belt.
        try:
            rock_axis = self.main_axis.cross(Vec(0, 0, 1)).unit()
        except ZeroDivisionError:
            rock_axis = Vec(1, 0, 0)
        angle_rock = self.main_axis.dihedral_angle(rock_axis, Vec(0, 0, 1))
        index_xy = 0
        for count_belt, item_belt_dict in enumerate(self.belt_dicts):
            if item_belt_dict.get("xy_factor"):
                index_xy = count_belt
                break 
        dihedral_zero_vec = self.belts_model[index_xy][0].rotate(rock_axis, -angle_rock).scale_along_axis(0, self.main_axis).unit()
        dihedral_stuff_model = []
        for count_belt, item_belt_model in enumerate(self.belts_model):
            dihedral_angles_n_vecs_model = []
            for count_vec, item_vec in enumerate(item_belt_model):
                vertex_model_rocked = item_vec.rotate(rock_axis, -angle_rock)
                dihedral_angle_model = vertex_model_rocked.dihedral_angle(self.main_axis, dihedral_zero_vec)
                dihedral_angles_n_vecs_model.append((dihedral_angle_model, vertex_model_rocked))
            dihedral_stuff_model.append(dihedral_angles_n_vecs_model)
        model_tara = dihedral_stuff_model[index_xy][0][0]
        dihedral_angles_model = [[(i[0] - model_tara) % (2 * pi) for i in j] for j in dihedral_stuff_model]
        self.belts_model = [[i[1] for i in j] for j in dihedral_stuff_model]
        all_angles_real, dihedral_angles_belts_real = set(), []
        for count_belt, item_belt in enumerate(self.belts_real):
            dihedral_angles_vecs_real = []
            for item_vec in item_belt:
                angle_vec_real = item_vec.dihedral_angle(self.main_axis, rock_axis)
                if self.belt_dicts[count_belt].get("xy_factor"):
                    all_angles_real.add(angle_vec_real)
                dihedral_angles_vecs_real.append(angle_vec_real)
            dihedral_angles_belts_real.append(dihedral_angles_vecs_real)
        total_delta_min = inf
        print(all_angles_real)
        for item_turn in iter(all_angles_real):
            belts_stuff_real, total_delta = [], 0
            for count_belt, item_belt in enumerate(dihedral_angles_belts_real):
                vecs_stuff_real = []
                for count_vec, item_vec_angle in enumerate(item_belt):
                    dihedral_angle_real_eff = (item_vec_angle - item_turn) % (2 * pi)
                    vecs_stuff_real.append((dihedral_angle_real_eff, self.belts_real[count_belt][count_vec], self.map_belts[count_belt][count_vec]))
                vecs_stuff_real.sort(key = lambda i: i[0])
                belts_stuff_real.append(vecs_stuff_real)
            for count_belt, item_belt_model in enumerate(dihedral_angles_model):
                weight = self.belt_dicts[count_belt].get("xy_factor")
                if weight:
                    for count_angle, item_angle_model in enumerate(item_belt_model):
                        total_delta += weight * abs(belts_stuff_real[count_belt][count_angle][0] - item_angle_model) # weight * sin(0.5 * diff) ** 2 would probably be better, but this works just fine...
            if total_delta < total_delta_min:
                total_delta_min = total_delta
                belts_stuff_real_best = belts_stuff_real
        self.belts_real = [[i[1] for i in j] for j in belts_stuff_real_best]
        self.map_belts = [[i[2] for i in j] for j in belts_stuff_real_best]
    
    def adjust_phi_mod(self):
        for item_param in self.dict_parameters.get("phi_mod"):
            belts_real = [i if self.belt_dicts[c].get("phi_mod") == item_param else [] for c, i in enumerate(self.belts_real)]
            belts_model = [i if self.belt_dicts[c].get("phi_mod") == item_param else [] for c, i in enumerate(self.belts_model)]
            count_iter, increment = 0, 0.5 * 10 ** -self.optimization_progress
            variable = variable_opt = 0
            delta_squared_min = inf
            while abs(increment) > 10 ** -(6 + self.optimization_progress) and count_iter < 60:
                delta_squared = 0
                for count_belt, item_belt_real in enumerate(belts_real):
                    for count_vec, item_vec_real in enumerate(item_belt_real):
                        model_vec_rotated = belts_model[count_belt][count_vec].rotate(self.main_axis, variable * self.belt_intensities[count_belt].get("phi_mod")[count_vec])
                        delta_squared += item_vec_real.distance_points_squared(model_vec_rotated)
                if delta_squared < delta_squared_min:
                    delta_squared_min = delta_squared
                    variable_opt = variable
                else:
                    increment *= -0.5
                    variable = variable_opt
                variable += increment
                count_iter += 1
            self.dict_parameters.get("phi_mod")[item_param] += variable_opt
            for count_belt, item_belt_model in enumerate(belts_model):
                for count_vec, item_vec_model in enumerate(item_belt_model):
                    self.belts_model[count_belt][count_vec] = item_vec_model.rotate(self.main_axis, variable_opt * self.belt_intensities[count_belt].get("phi_mod")[count_vec])
    
    def rotate_belts_fixed_axis(self):
        count_iter, increment, angle_iter, angle_iter_best, delta_squared_min = 0, 3 * 10 ** -self.optimization_progress, 0, 0, inf
        while abs(increment) > 10 ** -(6 + self.optimization_progress) and count_iter < 50:
            delta_squared = 0
            for count_belt, item_belt_real in enumerate(self.belts_real):
                for count_vec, item_vec_real in enumerate(item_belt_real):
                    vec_model = self.belts_model[count_belt][count_vec]
                    if count_iter:
                        model_vec_rotated = vec_model.rotate(self.main_axis, angle_iter)
                    else:
                        model_vec_rotated = vec_model
                    delta_squared += item_vec_real.distance_points_squared(model_vec_rotated)
            if delta_squared < delta_squared_min:
                delta_squared_min = delta_squared
                angle_iter_best = angle_iter
            else:
                increment *= -0.5
                angle_iter = angle_iter_best
            angle_iter += increment
            count_iter += 1
        self.belts_model = [[i.rotate(self.main_axis, angle_iter_best) for i in j] for j in self.belts_model]
    
    def adjust_phi(self):
        for item_param in self.dict_parameters.get("phi"):
            if item_param == "phi*":
                continue
            belts_real = [i if self.belt_dicts[c].get("phi") in (item_param, "phi*") else [] for c, i in enumerate(self.belts_real)]
            belts_model = [i if self.belt_dicts[c].get("phi") == item_param else [] for c, i in enumerate(self.belts_model)]
            belts_model_star = [i if self.belt_dicts[c].get("phi") == "phi*" else [] for c, i in enumerate(self.belts_model)]
            star_ratio = self.dict_phi_star_ratios.get(item_param, 1)
            count_iter, increment = 0, 0.5 * 10 ** -self.optimization_progress
            variable = variable_opt = 0
            delta_squared_min = inf
            while abs(increment) > 10 ** -(6 + self.optimization_progress) and count_iter < 60:
                delta_squared = 0
                for count_belt, item_belt_real in enumerate(belts_real):
                    for count_vec, item_vec_real in enumerate(item_belt_real):
                        if belts_model[count_belt]:
                            model_vec_rotated = belts_model[count_belt][count_vec].rotate(self.main_axis, variable)
                        elif belts_model_star[count_belt]:
                            model_vec_rotated = belts_model_star[count_belt][count_vec].rotate(self.main_axis, -variable * star_ratio)
                        delta_squared += item_vec_real.distance_points_squared(model_vec_rotated)
                if delta_squared < delta_squared_min:
                    delta_squared_min = delta_squared
                    variable_opt = variable
                else:
                    increment *= -0.5
                    variable = variable_opt
                variable += increment
                count_iter += 1
            self.dict_parameters.get("phi")[item_param] += variable_opt
            for count_belt, item_belt_model in enumerate(belts_model):
                for count_vec, item_vec_model in enumerate(item_belt_model):
                    self.belts_model[count_belt][count_vec] = item_vec_model.rotate(self.main_axis, variable_opt)
            if "phi*" in self.dict_parameters.get("phi"):
                self.dict_parameters.get("phi")["phi*"] -= variable_opt * star_ratio
                for count_belt, item_belt_model in enumerate(belts_model_star):
                    for count_vec, item_vec_model in enumerate(item_belt_model):
                        self.belts_model[count_belt][count_vec] = item_vec_model.rotate(self.main_axis, -variable_opt * star_ratio)
    
    def wrap_kabsch(self): # convert to numpy array for kabsch algorithm (which spits out a rotation matrix for the optimal orientation of the model polyhedron), apply said matrix, then deconvert.
        npvecs_real, npvecs_model = [], []
        for count_belt, item_belt_real in enumerate(self.belts_real):
            for count_vec, item_vec_real in enumerate(item_belt_real):
                vec_model = self.belts_model[count_belt][count_vec]
                npvecs_model.append([vec_model.x, vec_model.y, vec_model.z])
                npvecs_real.append([item_vec_real.x, item_vec_real.y, item_vec_real.z])
        rotation_matrix = kabsch_algorithm(npvecs_real, npvecs_model)
        self.belts_model, vecs_model_kabsch = [], []
        count_belts, count_vecs_belt = 0, 0
        for item_model_vec in npvecs_model:
            model_npvec_kabsch = rotation_matrix @ item_model_vec
            vec_model_deconverted = Vec(model_npvec_kabsch[0], model_npvec_kabsch[1], model_npvec_kabsch[2])
            vecs_model_kabsch.append(vec_model_deconverted)
            count_vecs_belt += 1
            if count_vecs_belt == len(self.belts_real[count_belts]):
                self.belts_model.append(vecs_model_kabsch)
                vecs_model_kabsch = []
                count_vecs_belt = 0
                count_belts += 1
        np_main_axis = array([self.main_axis.x, self.main_axis.y, self.main_axis.z])
        np_main_axis_kabsch = rotation_matrix @ np_main_axis
        self.main_axis = Vec(np_main_axis_kabsch[0], np_main_axis_kabsch[1], np_main_axis_kabsch[2])
        if self.main_axis.dot(self.belts_real[0][0]) < -0.0000001 and len(self.belts_real) > 1:
            self.env.warnings.add(f"Bad main axis encountered! High likelyhood of unreliable data ({self.name}, {self.env.env_fract[0][0]}, {self.env.cif.filepath}, axis =  ({stround(self.main_axis.x, 5)}, {stround(self.main_axis.y, 5)}, {stround(self.main_axis.z, 5)}).")
    
    def adjust_len(self):
        dict_len_aux = {}
        for count_belt, item_belt_dict in enumerate(self.belt_dicts):
            if item_belt_dict.get("len"):
                dict_len_aux[item_belt_dict.get("len")] = []
        for count_belt, item_belt_real in enumerate(self.belts_real): #measure free parameter data
            belt_dict = self.belt_dicts[count_belt]
            parameter = self.belt_dicts[count_belt].get("len")
            for count_vec, item_vec_real in enumerate(item_belt_real):
                vec_model = self.belts_model[count_belt][count_vec]
                if parameter:
                    dict_len_aux[parameter].append((count_belt, item_vec_real.dot(vec_model.unit()), vec_model.length()))
        dict_old = {}
        for item_param in dict_len_aux:
            old_lengths = [i[2] for i in dict_len_aux.get(item_param)]
            dict_old[item_param] = sum(old_lengths) / len(old_lengths)
        if self.dict_input.get("preserve_radial_avg") == True:
            dict_real_lengths = {}
            for count_belt, item_belt_dict in enumerate(self.belt_dicts):
                if item_belt_dict.get("len"):
                    if not dict_real_lengths.get(item_belt_dict.get("len")):
                        dict_real_lengths[item_belt_dict.get("len")] = []
                    dict_real_lengths[item_belt_dict.get("len")] += [i.length() for i in self.belts_real[count_belt]]
            for item_type in dict_real_lengths:
                avg_length = sum(dict_real_lengths.get(item_type)) / len(dict_real_lengths.get(item_type))
                self.dict_parameters.get("len")[item_type] = avg_length
        else: 
            for item_param in dict_len_aux:
                real_values = [i[1] for i in dict_len_aux.get(item_param)]
                model_values = [i[2] for i in dict_len_aux.get(item_param)]
                factor_set = self.dict_factor_sets.get(item_param)
                if len(factor_set) == 1:
                    fit_variable_opt = sum([abs(i) for i in real_values]) / (sum([abs(i) for i in model_values]) * next(iter(factor_set)))
                else:
                    fit_variable_opt, increment, loop_count, delta_squared_min = 1, 10 ** -self.optimization_progress, 0, inf
                    while abs(increment) > 10 ** -(6 + self.optimization_progress) and loop_count < 120:
                        delta_squared = 0
                        fit_variable = fit_variable_opt
                        if loop_count:
                            fit_variable += increment
                        for count_vertex, item_real in enumerate(real_values):
                            delta_squared += (item_real - fit_variable * model_values[count_vertex]) ** 2
                        if delta_squared < delta_squared_min:
                            delta_squared_min = delta_squared
                            fit_variable_opt = fit_variable
                        else:
                            fit_variable = fit_variable_opt
                            increment *= -0.5
                        loop_count += 1
                self.dict_parameters.get("len")[item_param] *= fit_variable_opt
        for count_belt, item_belt_model in enumerate(self.belts_model):
            belt_dict = self.belt_dicts[count_belt]
            parameter = belt_dict.get("len")
            for count_vec, item_vec_model in enumerate(item_belt_model):
                self.belts_model[count_belt][count_vec] = item_vec_model.scale(self.dict_parameters.get("len").get(parameter) / dict_old.get(parameter))
    
    def adjust_xyz(self):
        dict_parameters_old = deepcopy(self.dict_parameters)
        if self.dict_input.get("preserve_radial_avg"):
            dict_real_lengths, dict_best_lengths = {}, {}
            for count_belt, item_dict in enumerate(self.belt_dicts):
                xyz_tuple = (item_dict.get("xy"), item_dict.get("z"))
                if not dict_real_lengths.get(xyz_tuple):
                    dict_real_lengths[xyz_tuple] = []
                    dict_best_lengths[xyz_tuple] = []
                best_xy = self.dict_parameters.get("xy").get(item_dict.get("xy")) * item_dict.get("xy_factor") if item_dict.get("xy") else 0
                best_z = self.dict_parameters.get("z").get(item_dict.get("z")) * item_dict.get("z_factor") if item_dict.get("z") else 0
                dict_real_lengths[xyz_tuple] += [i.length() for i in self.belts_real[count_belt]]
                dict_best_lengths[xyz_tuple] += [sqrt(best_xy ** 2 + best_z ** 2) for i in self.belts_real[count_belt]]
            for item_type in dict_real_lengths:
                avg_length = sum(dict_real_lengths.get(item_type)) / len(dict_real_lengths.get(item_type))
                best_length = sum(dict_best_lengths.get(item_type)) / len(dict_best_lengths.get(item_type))
                if item_type[0]:
                    self.dict_parameters.get("xy")[item_type[0]] *= avg_length / best_length
                if item_type[1] and item_type[1] != "z*":
                    self.dict_parameters.get("z")[item_type[1]] *= avg_length / best_length
        else:
            for item_type in ("xy", "z"):
                for item_param in self.dict_parameters.get(item_type):
                    real_values, model_factors = [], []
                    for count_belt, item_belt in enumerate(self.belts_real):
                        belt_dict = self.belt_dicts[count_belt]
                        if belt_dict.get(item_type) != item_param:
                            continue
                        for count_vec, item_vec_real in enumerate(item_belt):
                            if item_type == "xy":
                                plane_normal = self.belts_model[count_belt][count_vec].cross(self.main_axis).unit()
                                xy_measured = item_vec_real.scale_along_axis(0, plane_normal).scale_along_axis(0, self.main_axis).length()
                                real_values.append(xy_measured)
                                model_factors.append(belt_dict.get("xy_factor"))
                            elif item_type == "z":
                                real_values.append(item_vec_real.dot(self.main_axis))
                                model_factors.append(belt_dict.get("z_factor"))
                    factor_set = self.dict_factor_sets.get(item_param)
                    if len(factor_set) == 1:
                        fit_variable_opt = sum([i / model_factors[c] for c, i in enumerate(real_values)]) / len(model_factors)
                    else:
                        fit_variable_opt, increment, loop_count, delta_squared_min = dict_parameters_old.get(item_type).get(item_param), 10 ** -self.optimization_progress, 0, inf
                        while abs(increment) > 10 ** -(6 + self.optimization_progress) and loop_count < 120:
                            delta_squared = 0
                            fit_variable = fit_variable_opt
                            if loop_count:
                                fit_variable += increment
                            for count_vertex, item_real in enumerate(real_values):
                                delta_squared += (item_real - fit_variable * model_factors[count_vertex]) ** 2
                            if delta_squared < delta_squared_min:
                                delta_squared_min = delta_squared
                                fit_variable_opt = fit_variable
                            else:
                                increment *= -0.5
                            loop_count += 1
                    self.dict_parameters.get(item_type)[item_param] = fit_variable_opt
        for count_belt, item_belt_model in enumerate(self.belts_model):
            belt_dict = self.belt_dicts[count_belt]
            z_parameter, xy_parameter = belt_dict.get("z"), belt_dict.get("xy")
            if z_parameter == "z*": # z* serves as a special dependent parameter to guarantee the centering of the polyhedron by compensating for the other z values.
                z_numer = 0
                for count_belt, item_belt_dict in enumerate(self.belt_dicts):
                    if item_belt_dict.get("z") != "z*":
                        z_numer += self.dict_parameters.get("z").get(item_belt_dict.get("z")) * item_belt_dict.get("z_factor") * len(self.belts_model[count_belt])
                self.dict_parameters.get("z")["z*"] = -z_numer / len(item_belt_model)
            for count_vec, item_vec_model in enumerate(item_belt_model):
                vertex_model = item_vec_model
                if z_parameter:
                    vertex_model = vertex_model + self.main_axis.scale(belt_dict.get("z_factor") * (self.dict_parameters.get("z").get(z_parameter) - dict_parameters_old.get("z").get(z_parameter)))
                if xy_parameter:
                    xy_unit = vertex_model.scale_along_axis(0, self.main_axis).unit()
                    vertex_model = vertex_model + xy_unit.scale(belt_dict.get("xy_factor") * (self.dict_parameters.get("xy").get(xy_parameter) - dict_parameters_old.get("xy").get(xy_parameter)))
                self.belts_model[count_belt][count_vec] = vertex_model
    
    def adjust_bundles(self):
        if self.dict_parameters.get("len"):
            dict_lengths_aux = {}
            for count_belt, item_belt in enumerate(self.belt_dicts):
                param = item_belt.get("len")
                if not dict_lengths_aux.get(param):
                    dict_lengths_aux[param] = [0, 0]
                for item_vec in self.belts_model[count_belt]:
                    dict_lengths_aux.get(param)[0] += item_vec.length()
                    dict_lengths_aux.get(param)[1] += 1
            dict_lengths = {}
            for item_param in dict_lengths_aux:
                dict_lengths[item_param] = dict_lengths_aux.get(item_param)[0] / dict_lengths_aux.get(item_param)[1]
        dict_vars = {}
        for item_var in self.dict_bundles:
            dict_vars[item_var] = self.dict_parameters.get("parameter_bundles").get(item_var)
        for item_var in self.dict_bundles:
            dict_init = {}
            for item_param in self.dict_bundles.get(item_var):
                dependency = self.dict_bundles.get(item_var).get(item_param)
                dict_init[item_param] = parse_math(dependency, dict_variables = dict_vars)
            bundle_lower, bundle_upper = self.dict_info.get("var_range").get(item_var)[:2]
            increment = (bundle_upper - bundle_lower) / 3
            loop_count, delta_squared_min = 0, inf
            variable = variable_best = var_init = self.dict_parameters.get("parameter_bundles").get(item_var)
            while abs(increment) > 10 ** -(6 + self.optimization_progress) and loop_count < 120:
                dict_recycle = {}
                dict_vars[item_var] = variable
                belts_model_copy = deepcopy(self.belts_model)
                for item_type in self.dict_bound_parameters:
                    for count_belt, item_belt_real in enumerate(self.belts_real):
                        belt_dict = self.belt_dicts[count_belt]
                        parameter = belt_dict.get("$" + item_type)
                        if not parameter in self.dict_bundles.get(item_var):
                            continue
                        if item_type == "z":
                            z_sign = 1 if belt_dict.get("z_factor") > 0 else -1
                        dependency = self.dict_bundles.get(item_var).get(parameter)
                        if parameter in dict_recycle:
                            value = dict_recycle.get(parameter)
                        else:
                            value = dict_recycle[parameter] = parse_math(dependency, dict_variables = dict_vars)
                        value_init = dict_init.get(parameter)
                        value_diff = value - value_init
                        avg_len = dict_lengths.get(belt_dict.get("len")) if belt_dict.get("len") else 1
                        belt_intensities = self.belt_intensities[count_belt].get(item_type)
                        try:
                            for count_vec in range(len(item_belt_real)):
                                vec_model = belts_model_copy[count_belt][count_vec]
                                if item_type == "phi":
                                    belts_model_copy[count_belt][count_vec] = vec_model.rotate(self.main_axis, value_diff)
                                elif item_type == "xy":
                                    xy_unit = vec_model.scale_along_axis(0, self.main_axis).unit()
                                    belts_model_copy[count_belt][count_vec] = vec_model + xy_unit.scale(avg_len * value_diff)
                                elif item_type == "z":
                                    belts_model_copy[count_belt][count_vec] = vec_model + self.main_axis.scale(z_sign * avg_len * value_diff)
                                elif item_type.startswith("phi_mod"):
                                    belts_model_copy[count_belt][count_vec] = vec_model.rotate(self.main_axis, -value_diff * belt_intensities[count_vec])
                                elif item_type.startswith("xy_mod"):
                                    xy_unit = vec_model.scale_along_axis(0, self.main_axis).unit()
                                    belts_model_copy[count_belt][count_vec] = vec_model + xy_unit.scale(avg_len * value_diff * belt_intensities[count_vec])
                                elif item_type.startswith("z_mod"):
                                    belts_model_copy[count_belt][count_vec] = vec_model + self.main_axis.scale(avg_len * value_diff * belt_intensities[count_vec])
                        except ValueError:
                            self.commentline.configure(text = f"Problem when optimizing model '{self.name}' for central atom '{self.env.env_fract[0][0]}' ({self.env.cif.cif_filepath}): the dependency expression {parameter} = {self.dict_bundles.get(item_var).get(parameter)} is undefined for {item_var} = {value_diff}. Try adjusting the expression or its range.", foreground = "#ff0000")
                            raise ValueError
                delta_squared = 0
                for count_belt, item_belt_real in enumerate(self.belts_real):
                    for count_vec, item_vec_real in enumerate(item_belt_real):
                        delta_squared += belts_model_copy[count_belt][count_vec].distance_points_squared(item_vec_real)
                if delta_squared < delta_squared_min:
                    delta_squared_min = delta_squared
                    variable_best = variable
                else:
                    increment *= -0.5
                    variable = variable_best
                variable += increment
                if variable < bundle_lower or variable > bundle_upper:
                    variable -= 2 * increment
                    increment *= -1
                loop_count += 1
            self.dict_parameters.get("parameter_bundles")[item_var] += variable_best - var_init
            self.belts_model = deepcopy(belts_model_copy)
    
    def adjust_xyz_mod(self):
        dict_parameters_old = deepcopy(self.dict_parameters)
        for item_type in self.dict_parameters:
            if not item_type.startswith("xy_mod") and not item_type.startswith("z_mod"):
                continue
            for item_param in self.dict_parameters.get(item_type):
                aux_belts = []
                for count_belt, item_belt in enumerate(self.belts_real):
                    belt_dict = self.belt_dicts[count_belt]
                    base_type = item_type[:item_type.index("_")]
                    factor = belt_dict.get(base_type + "_factor")
                    if belt_dict.get("len"):
                        base = self.dict_parameters.get("len").get(belt_dict.get("len")) * factor
                    else:
                        base = self.dict_parameters.get(base_type).get(belt_dict.get(base_type), 0)
                    aux_vecs = []
                    if item_type.startswith("z_mod"):
                        for count_vec, item_vec_real in enumerate(item_belt):
                            aux_vecs.append(item_vec_real.dot(self.main_axis) - base * factor)
                    else:
                        for count_vec, item_vec_real in enumerate(item_belt):
                            try:
                                plane_normal = self.belts_model[count_belt][count_vec].cross(self.main_axis).unit()
                            except ZeroDivisionError:
                                plane_normal = self.main_axis
                            xy_real = item_vec_real.scale_along_axis(0, plane_normal).scale_along_axis(0, self.main_axis).length()
                            aux_vecs.append(xy_real - base * factor)
                    aux_belts.append(aux_vecs)
                if len(self.dict_factor_sets.get(item_param)) == 1:
                    vec_count, fit_variable_opt = 0, 0
                    for count_belt, item_aux_belt in enumerate(aux_belts):
                        if not self.belt_intensities[count_belt].get(item_type):
                            continue
                        for count_vec, item_vec in enumerate(item_aux_belt):
                            fit_variable_opt += item_vec / self.belt_intensities[count_belt].get(item_type)[count_vec]
                            vec_count += 1
                    fit_variable_opt /= vec_count
                else:
                    fit_variable_opt, increment, loop_count, delta_squared_min = dict_parameters_old.get(item_type).get(item_param), 10 ** -self.optimization_progress, 0, inf
                    while abs(increment) > 10 ** -(6 + self.optimization_progress) and loop_count < 120:
                        delta_squared = 0
                        fit_variable = fit_variable_opt
                        if loop_count:
                            fit_variable += increment
                        for count_belt, item_belt in enumerate(aux_belts):
                            belt_intensities = self.belt_intensities[count_belt].get(item_type)
                            if belt_intensities:
                                for count_vec, item_diff_base in enumerate(item_belt):
                                    delta_squared += (item_diff_base - fit_variable * belt_intensities[count_vec]) ** 2
                        if delta_squared < delta_squared_min:
                            delta_squared_min = delta_squared
                            fit_variable_opt = fit_variable
                        else:
                            increment *= -0.5
                        loop_count += 1
                self.dict_parameters.get(item_type)[item_param] = fit_variable_opt
                for count_belt, item_belt_model in enumerate(deepcopy(self.belts_model)):
                    if item_param == self.belt_dicts[count_belt].get(item_type):
                        for count_vec, item_vec_model in enumerate(item_belt_model):
                            unitvec_direction = item_vec_model.scale_along_axis(0, self.main_axis).unit() if item_type.startswith("xy_mod") else self.main_axis
                            intensity = self.belt_intensities[count_belt].get(item_type)[count_vec]
                            modval = intensity * (self.dict_parameters.get(item_type).get(item_param) - dict_parameters_old.get(item_type).get(item_param))
                            self.belts_model[count_belt][count_vec] = item_vec_model + unitvec_direction.scale(modval)
    
    def polish(self):
        if self.dict_input.get("include_central_atom_in_fit") and self.dict_input.get("center_method") == "default":
            corr_vec = self.excentricity_vec_cart.scale(-1 / (len(self.env.env_fract)))
            self.excentricity_vec_cart = self.excentricity_vec_cart + corr_vec
            for count_belt, item_belt_real in enumerate(self.belts_real):
                for count_vec, item_vec_real in enumerate(item_belt_real):
                    self.belts_real[count_belt][count_vec] = item_vec_real - corr_vec
        if self.dict_bundles:
            self.dict_parameters = register_bound_parameters(self.dict_parameters, self.dict_bundles, self.dict_bound_parameters)
        
    def get_csm(self):
        total_squared_delta, total_squared_distances_from_center = 0, 0
        total_linear_delta, total_radial_delta, total_angular_delta = 0, 0, 0
        for count_belt, item_belt_real in enumerate(self.belts_real):
            for count_vec, item_vec_real in enumerate(item_belt_real):
                vec_model = self.belts_model[count_belt][count_vec]
                delta_squared = vec_model.distance_points_squared(item_vec_real)
                total_squared_distances_from_center += item_vec_real.squared_length()
                total_squared_delta += delta_squared
                total_linear_delta += sqrt(delta_squared)
                total_angular_delta += abs(item_vec_real.angle_vecs(vec_model))
                total_radial_delta += abs(vec_model.length() - item_vec_real.length())
        self.avg_delta = total_linear_delta / self.env.coordination_number
        self.avg_delta_angular = deg(total_angular_delta / self.env.coordination_number)
        self.avg_delta_radial = total_radial_delta / self.env.coordination_number
        if self.dict_input.get("include_central_atom_in_fit"):
            excentricity_squared = self.excentricity_vec_cart.squared_length()
            total_squared_delta += excentricity_squared
            total_squared_distances_from_center += excentricity_squared
        self.csm = 100 * total_squared_delta / total_squared_distances_from_center
        self.csm_taxed = self.csm + self.dict_input.get("var_tax") * self.parameter_count
    
    def add_to_symmetree(self):
        parent_candidates = []
        if self.dict_info.get("aristohedra"):
            parent_candidates = [i for i in self.env.polyhedra if i.name in self.dict_info.get("aristohedra", {})]
            if parent_candidates:
                best_parent = min(parent_candidates, key = lambda i: i.csm)
            else:
                try:
                    aristohedral_branch = self.env.cif.main.dict_upside_down_nodes.get(self.name).get_branches()
                    qualifying_polyhedra = [self.env.dict_symmetree_nodes.get(i.name).content for i in aristohedral_branch if i.name in self.env.dict_symmetree_nodes]
                    if qualifying_polyhedra:
                        best_parent = min(qualifying_polyhedra, key = lambda i: i.csm)
                    else:
                        best_parent = self.env.symmetree_root
                except AttributeError:
                    best_parent = self.env.symmetree_root
        else:
            best_parent = self.env.symmetree_root
        this_poly_node = TreeNode(self, self.name, [])
        self.env.dict_symmetree_nodes[self.name] = this_poly_node
        self.env.dict_symmetree_nodes.get(best_parent.name).children.append(this_poly_node)
            
    def write_output(self):
        round_to = 5
        input_dir = self.dict_input.get("input_dir")
        atomlist = self.env.env_fract
        cell_parameters = self.env.cif.cell_parameters
        excentricity_vec_fract = self.excentricity_vec_cart.coordinate_transformation(self.env.cif.MATRIX_TO_FRACT)
        main_axis_fract = self.main_axis.coordinate_transformation(self.env.cif.MATRIX_TO_FRACT).unit()
        central_atom = atomlist[0][0]
        try:
            axis_transformed_to_z = self.main_axis.cross(Vec(0, 0, 1)).unit()
        except ZeroDivisionError:
            axis_transformed_to_z = self.main_axis
        transformation_angle = -self.main_axis.angle_vecs(Vec(0, 0, 1))
        try:
            if self.dict_input.get("output_details"):
                file_type, file_end = "", ".out"
                output = get_file(f"{input_dir}/polynator_output/", f"{self.env.cif.filepath.removesuffix('.cif')}_{central_atom}_{self.name.removeprefix('>')}_{stround(self.csm, 0)}", ".out", overwrite = self.dict_input.get("overwrite_output"))
                output.write(f"central atom: {central_atom} (element: {atomlist[0][1]})\ncoordination number: {self.env.coordination_number}\n")
                output.write(f"model polyhedron: {self.name} | point group: {self.dict_info.get('point_group')} ({self.point_group_sch})\n")
                output.write(f"csm = {stround(self.csm, round_to)}\n")
                output.write(f"main axis: ({stround(main_axis_fract.x, round_to)}, {stround(main_axis_fract.y, round_to)}, {stround(main_axis_fract.z, round_to)}) | model symmetry in this direction: {self.dict_info.get('axis', '?')}\n")
                output.write(f"ligand centroid: ({stround(self.env.centroid_fract.x, round_to)}, {stround(self.env.centroid_fract.y, round_to)}, {stround(self.env.centroid_fract.z, round_to)}) | central_atom vector: ({stround(atomlist[0][2].x, round_to)}, {stround(atomlist[0][2].y, round_to)}, {stround(atomlist[0][2].z, round_to)}) | excentricity vector: ({stround(excentricity_vec_fract.x, round_to)}, {stround(excentricity_vec_fract.y, round_to)}, {stround(excentricity_vec_fract.z, round_to)}) | excentricity length: {stround(100 * self.excentricity_vec_cart.length(), 3)} pm\n")
                output.write("\n")
            if self.dict_input.get("output_minicif_cart"):
                file_type, file_end = "_minicart", ".cif"
                cifput_cart = get_file(f"{input_dir}/polynator_output/", f"{self.env.cif.filepath.removesuffix('.cif')}_minicart_{central_atom}_{self.name.removeprefix('>')}_{stround(self.csm, 0)}", ".cif", overwrite = self.dict_input.get("overwrite_output"))
            if self.dict_input.get("output_minicif_fract"):
                file_type, file_end = "_minifract", ".cif"
                cifput_fract = get_file(f"{input_dir}/polynator_output/", f"{self.env.cif.filepath.removesuffix('.cif')}_minifract_{central_atom}_{self.name.removeprefix('>')}_{stround(self.csm, 0)}", ".cif", overwrite = self.dict_input.get("overwrite_output"))
        except PermissionError:
            warning = f"Cannot open {input_dir}/polynator_output/{self.env.cif.filepath.removesuffix('.cif')}{file_type}_{central_atom}_{self.name.removeprefix('>')}_{stround(self.csm, 0)}{file_end}, it's probably opened in another window."
            self.commentline.configure(text = warning, foreground = "#ff0000")
            self.warnings.add(warning)
        if self.dict_input.get("output_minicif_cart"):
            cifput_cart.write("\ndata_minimal\nloop_\n_cell_length_a 20\n_cell_length_b 20\n_cell_length_c 20\n_cell_angle_alpha 90\n_cell_angle_beta 90\n_cell_angle_gamma 90\n_symmetry_Int_Tables_number 1\nloop_\n_symmetry_equiv_pos_site_id\n_symmetry_equiv_pos_as_xyz\n1 'x, y, z'\nloop_\n_atom_type_symbol\n")
            cifput_cart.write(central_atom + "\n")
            for item_ligand in atomlist[1:]:
                cifput_cart.write(item_ligand[0] + "\n")
            cifput_cart.write("loop_\n_atom_site_label\n_atom_site_type_symbol\n_atom_site_fract_x\n_atom_site_fract_y\n_atom_site_fract_z\n")
            cifput_cart.write(f"{central_atom}_Z {atomlist[0][1]} {stround(0.5 + self.excentricity_vec_cart.scale(0.05).x, 5)} {stround(0.5 + self.excentricity_vec_cart.scale(0.05).y, 5)} {stround(0.5 + self.excentricity_vec_cart.scale(0.05).z, 5)}\nZ Z 0.5 0.5 0.5\n")
        if self.dict_input.get("output_minicif_fract"):
            cifput_fract.write(f"\ndata_minimal\nloop_\n_cell_length_a {cell_parameters[0]}\n_cell_length_b {cell_parameters[1]}\n_cell_length_c {cell_parameters[2]}\n_cell_angle_alpha {stround(deg(cell_parameters[3]), 3)}\n_cell_angle_beta {stround(deg(cell_parameters[4]), 3)}\n_cell_angle_gamma {stround(deg(cell_parameters[5]), 3)}\n_symmetry_Int_Tables_number 1\nloop_\n_symmetry_equiv_pos_site_id\n_symmetry_equiv_pos_as_xyz\n1 'x, y, z'\nloop_\n_atom_type_symbol\n")
            cifput_fract.write(central_atom + "\n")
            for item_ligand in atomlist[1:]:
                cifput_fract.write(item_ligand[0] + "\n")
            cifput_fract.write("loop_\n_atom_site_label\n_atom_site_type_symbol\n_atom_site_fract_x\n_atom_site_fract_y\n_atom_site_fract_z\n")
            cifput_fract.write(f"{central_atom}_Z {atomlist[0][1]} {stround(atomlist[0][2].x, 5)} {stround(atomlist[0][2].y, 5)} {stround(atomlist[0][2].z, 5)}\nZ Z {stround(atomlist[0][2].x + excentricity_vec_fract.x, 5)} {stround(atomlist[0][2].y + excentricity_vec_fract.y, 5)} {stround(atomlist[0][2].z + excentricity_vec_fract.z, 5)}\n")
        count_atom = 0
        total_model_length, total_real_length, unit_proj_numerator, squared_dotprod_deltas = 0, 0, 0, 0
        delta_vecs, real_vecs = [], []
        for count_belt, item_belt_real in enumerate(self.belts_real):
            for count_vec, item_vec_real in enumerate(item_belt_real):
                vec_model = self.belts_model[count_belt][count_vec]
                diff_real_model_vec_cart = vec_model - item_vec_real
                total_model_length += vec_model.length()
                total_real_length += item_vec_real.length()
                angular_difference = item_vec_real.angle_vecs(vec_model)
                radial_difference = vec_model.length() - item_vec_real.length()
                vec_real_cif_cart = item_vec_real.rotate(axis_transformed_to_z, transformation_angle).scale(0.05)
                vec_model_cif_cart = self.belts_model[count_belt][count_vec].rotate(axis_transformed_to_z, transformation_angle).scale(0.05)
                vertex_real_fract = item_vec_real.coordinate_transformation(self.env.cif.MATRIX_TO_FRACT) + atomlist[0][2]
                model_fract = vec_model.coordinate_transformation(self.env.cif.MATRIX_TO_FRACT) + atomlist[0][2]
                diff_real_model_vec_fract = model_fract - vertex_real_fract
                delta_vecs.append(diff_real_model_vec_cart)
                real_vecs.append(item_vec_real)
                if self.dict_input.get("output_minicif_cart"):
                    cifput_cart.write(f"{atomlist[1 + self.map_belts[count_belt][count_vec]][0]}_L{count_atom} {atomlist[1 + self.map_belts[count_belt][count_vec]][1]} {stround(0.5 + vec_real_cif_cart.x, round_to)} {stround(0.5 + vec_real_cif_cart.y, round_to)} {stround(0.5 + vec_real_cif_cart.z, round_to)}\nL_{count_atom} L {stround(0.5 + vec_model_cif_cart.x, round_to)} {stround(0.5 + vec_model_cif_cart.y, round_to)} {stround(0.5 + vec_model_cif_cart.z, round_to)}\n")
                if self.dict_input.get("output_minicif_fract"):
                   cifput_fract.write(f"{atomlist[1 + self.map_belts[count_belt][count_vec]][0]}_L{count_atom} {atomlist[1 + self.map_belts[count_belt][count_vec]][1]} {stround(vertex_real_fract.x, round_to)} {stround(vertex_real_fract.y, round_to)} {stround(vertex_real_fract.z, round_to)}\nL_{count_atom} L {stround(model_fract.x, round_to)} {stround(model_fract.y, round_to)} {stround(model_fract.z, round_to)}\n")
                if self.dict_input.get("output_details"):
                    real_unit = item_vec_real.unit()
                    unit_proj_numerator += vec_model.unit().distance_points_squared(real_unit)
                    squared_dotprod_deltas += (vec_model.dot(self.main_axis) - item_vec_real.dot(self.main_axis)) ** 2
                    angular_difference_theta = vec_model.angle_vecs(self.main_axis) - item_vec_real.angle_vecs(self.main_axis) 
                    angular_difference_phi = vec_model.dihedral_angle(self.main_axis, real_unit)
                    if angular_difference_phi > pi:
                        angular_difference_phi = 2 * pi - angular_difference_phi
                    output.write(f"ligand: {atomlist[1 + self.map_belts[count_belt][count_vec]][0]} | coordinates: ({stround(vertex_real_fract.x, round_to)}, {stround(vertex_real_fract.y, round_to)}, {stround(vertex_real_fract.z, round_to)}) | matching model vertex: ({stround(model_fract.x, round_to)}, {stround(model_fract.y, round_to)}, {stround(model_fract.z, round_to)}) | delta vector: ({stround(diff_real_model_vec_fract.x, round_to)}, {stround(diff_real_model_vec_fract.y, round_to)}, {stround(diff_real_model_vec_fract.z, round_to)})\n")
                    output.write(f"delta length: {stround(100 * diff_real_model_vec_cart.length(), 3)} pm | radial delta: {stround(100 * radial_difference, 3)} pm | angular delta: {stround(deg(angular_difference), 3)} deg | delta phi: {stround(deg(angular_difference_phi), 3)} deg | delta theta: {stround(deg(angular_difference_theta), 3)} deg\n")
                count_atom += 1
        if self.dict_input.get("output_details"):
            csm_unit_sphere_proj = 100 * unit_proj_numerator / self.env.coordination_number
            csm_main_axis_proj = 100 * squared_dotprod_deltas / self.env.coordination_number
            total_deviation_squared = 0
            for item_delta_vec in delta_vecs:
                total_deviation_squared += (item_delta_vec.length() - self.avg_delta) ** 2
            std_deviation = sqrt(total_deviation_squared / self.env.coordination_number)
            output.write(f"\navg distance model vertices to polyhedron center: {stround(total_model_length / self.env.coordination_number, 5)} Angstrom | avg distance real vertices to polyhedron center: {stround(total_real_length / self.env.coordination_number, 5)} Angstrom")
            output.write(f"\navg distance real-model vertex pairs: {stround(100 * self.avg_delta, 3)} pm | standard deviation thereof: {stround(100 * std_deviation, 3)} pm | avg radial difference: {stround(100 * self.avg_delta_radial, 3)} pm | avg angular difference: {stround(self.avg_delta_angular, 3)} deg\n")
            output.write(f"csm of unit sphere projection: {stround(csm_unit_sphere_proj, round_to)} | csm of main axis projection: {stround(csm_main_axis_proj, round_to)}\n")
            output.write("\nmodel constructor:\n")
            for item_belt in self.belt_dicts:
                output.write(f"    {item_belt}\n")
            output.write("\nfree parameters: \n")
            for item_type in self.dict_parameters:
                if self.dict_parameters.get(item_type) == {} or item_type == "parameter_bundles":
                    continue
                count_parameter = 0
                for item_param in self.dict_parameters.get(item_type):
                    if item_param in self.dict_bound_parameters.get(item_type, {}):
                        continue
                    if count_parameter == 0:
                        output.write(f"    {item_type}-type: ")
                    if "phi" in item_type:
                        parameter_value = deg(self.dict_parameters.get(item_type).get(item_param))
                    else:
                        parameter_value = self.dict_parameters.get(item_type).get(item_param)
                    output.write(f"{item_param} = {stround(parameter_value, round_to)} {'deg' if 'phi' in item_type else 'Angstrom'}")
                    count_parameter += 1
                    if count_parameter == len(self.dict_parameters.get(item_type)):
                        output.write("\n")
                    elif count_parameter:
                        output.write(", ")
            if self.dict_info.get("parameter_bundles"):
                output.write("\nbound parameters:\n")
                for item_var in self.dict_info.get("parameter_bundles"):
                    output.write(f"    variable: {item_var} | range: {self.dict_info.get('var_range').get(item_var)[0]} <= {item_var} <= {self.dict_info.get('var_range').get(item_var)[1]} | initial value: {self.dict_info.get('var_range').get(item_var)[2]} | optimized value: {stround(self.dict_parameters.get('parameter_bundles').get(item_var), 5)}\n")
                parameters = set()
                for item_type in self.dict_bound_parameters:
                    for item_param in self.dict_bound_parameters.get(item_type):
                        if item_param in parameters:
                            continue
                        parameters.add(item_param)
                        dependency_expression = self.dict_bound_parameters.get(item_type).get(item_param)
                        if item_type.startswith("phi"):
                            unit, parameter_value, decimal_places = "deg", deg(self.dict_parameters.get("parameter_bundles").get(item_param)), 3
                        else:
                            unit, parameter_value, decimal_places = "Angstrom", self.dict_parameters.get("parameter_bundles").get(item_param), 5
                        output.write(f"        {item_param} = {dependency_expression} = {stround(parameter_value, decimal_places)} {unit}\n")
            output.flush()
            output.close()
        if self.dict_input.get("output_minicif_cart"):
            cifput_cart.flush()
            cifput_cart.close()
        if self.dict_input.get("output_minicif_fract"):
            cifput_fract.flush()
            cifput_fract.close()
    
    def copy(self):
        return Polyhedron(self.env, self.blueprint, warnings = self.warnings.copy(), dict_parameters = deepcopy(self.dict_parameters), main_axis = self.main_axis.copy(), belts_model = deepcopy(self.belts_model), belts_real = deepcopy(self.belts_real), map_belts = deepcopy(self.map_belts), csm = self.csm, csm_taxed = self.csm_taxed, excentricity_vec = self.excentricity_vec_cart.copy(), avg_delta = self.avg_delta, avg_delta_angular = self.avg_delta_angular, avg_delta_radial = self.avg_delta_radial)
    
def give_hint(widget, window, hint):
    def show(event):
        if window.dict_input.get("show_comments"):
            window.commentline.configure(text = hint, foreground = "#000000")
    def clear(event):
        window.commentline.configure(text = "")
    widget.bind("<Enter>", show)
    widget.bind("<Leave>", clear)

def paint_it_black(listbox):
    for item_index in range(listbox.size()):
        listbox.itemconfigure(item_index, foreground = "#000000")

def get_poly_comment(blueprint):
    infostrings = [", ".join(blueprint[1].get(i)) if blueprint[1].get(i) else "" for i in ("synonyms", "tags", "aristohedra")]
    point_group_string = f"{blueprint[1].get('point_group')} ({DICT_POINT_GROUP_CONVERSION.get(blueprint[1].get('point_group'), '?')})"
    return f"{'alternative names: ' + infostrings[0] + ' | ' if infostrings[0] else ''}point group: {point_group_string}{' | tags: ' + infostrings[1] if infostrings[1] else ''}{' | derivative of: ' + infostrings[2] if infostrings[2] else ''}"

def validate_input(input_candidate, default_input):
    try:
        if isinstance(input_candidate, dict):
            for item in input_candidate:
                input_valid = validate_input(input_candidate.get(item), default_input.get(item))
                if not input_valid:
                    return False
        elif type(input_candidate) in (tuple, list, set):
            for count, item in enumerate(input_candidate):
                input_valid = validate_input(item, default_input[count])
                if not input_valid:
                    return False
        elif type(input_candidate) not in [type(i) for i in default_input]:
            return False
    except IndexError:
        return True
    return True

def get_face_borders(vecs_face, indices_face, face_normal, minor_axis):
    set_confirmed_edges = set()
    centroid = get_centroid(vecs_face)
    vecs_face_tare = [i - centroid for i in vecs_face]
    indexed_angles = []
    for count_vec, item_vec in enumerate(vecs_face_tare):
        indexed_angles.append((item_vec.dihedral_angle(face_normal, minor_axis), indices_face[count_vec]))
    indexed_angles.sort(key = lambda i: i[0])
    for count_vec, item_vec in enumerate(indexed_angles):
        set_confirmed_edges.add(tuple(sorted((item_vec[1], indexed_angles[count_vec - 1][1]))))
    return set_confirmed_edges

def find_edges(list_poly_vecs, convexicity_tolerance = 0.0001, vec_type = "real"):
    if len(list_poly_vecs) < 2:
        return set(), []
    if len(list_poly_vecs) == 2:
        return {(0, 1)}, []
    if len(list_poly_vecs) == 3:
        return {(0, 1), (0, 2), (1, 2)}, [{0, 1, 2}]
    if len(list_poly_vecs) > 60:
        return set(), []
    probe_normal = list_poly_vecs[0].cross(list_poly_vecs[1])
    if probe_normal.length() < 0.00001:
        probe_normal = list_poly_vecs[0].cross(list_poly_vecs[2])
    if not [True for i in list_poly_vecs if abs(probe_normal.dot(i)) > 0.0001]:
        return get_face_borders(list_poly_vecs, tuple(range(len(list_poly_vecs))), probe_normal.unit(), list_poly_vecs[1].unit()), [set(range(len(list_poly_vecs)))]
    distances = []
    set_initial_edges, set_confirmed_edges = set(), set()
    for count_vec_1, item_vec_1 in enumerate(list_poly_vecs[:len(list_poly_vecs) - 1]):
        for count_vec_2, item_vec_2 in enumerate(list_poly_vecs[count_vec_1 + 1:]):
            distances.append(((count_vec_1, count_vec_1 + count_vec_2 + 1), round(item_vec_1.distance_points(item_vec_2), 5)))
    distances.sort(key = lambda i: i[1])
    if len(list_poly_vecs) < 25:
        dist_limit = inf
    elif vec_type == "model":
        dist_limit = 2.8 * distances[0][1]
    else:
        dist_limit = 0.5 * distances[-1][1]
    for item_distance in distances:
        if item_distance[1] > dist_limit:
            break
        set_initial_edges.add(item_distance[0])
    confirmed_faces = []
    list_edges = [set(i) for i in list(set_initial_edges)]
    for count_edge_1, item_edge_1 in enumerate(list_edges):
        for count_edge_2, item_edge_2 in enumerate(list_edges):
            if len(item_edge_1 & item_edge_2) != 1:
                continue
            full_face = item_edge_1 | item_edge_2
            if [True for i in confirmed_faces if not full_face - i]:
                continue
            vecs_face = [list_poly_vecs[i] for i in tuple(full_face)]
            vecs_face_tare = [i - vecs_face[0] for i in vecs_face]
            try:
                face_normal = vecs_face_tare[1].cross(vecs_face_tare[2]).unit()
            except ZeroDivisionError:
                continue
            minor_axis = face_normal.cross(vecs_face_tare[1]).unit()
            sign_cross = face_normal.dot(vecs_face[0])
            if abs(sign_cross) < 0.00001:
                continue
            outer_face = True
            for count_vec, item_vec in enumerate(list_poly_vecs):
                if count_vec in full_face:
                    continue
                if face_normal.dot(item_vec - vecs_face[0]) * sign_cross > convexicity_tolerance:
                    outer_face = False
                    break
                if abs(face_normal.dot(item_vec - vecs_face[0])) < 0.00001:
                    full_face.add(count_vec)
            if (len(full_face) > 3 or item_edge_1 ^ item_edge_2 in list_edges) and outer_face:
                confirmed_faces.append(full_face)
                indices_face = tuple(full_face)
                set_confirmed_edges.update(get_face_borders([list_poly_vecs[i] for i in indices_face], indices_face, face_normal, minor_axis))
    return set_confirmed_edges, confirmed_faces

def parse_symops(list_symops, dataset_atom, warnings):
    DICT_NUMS = {"x": dataset_atom[2], "y": dataset_atom[3], "z": dataset_atom[4], "1/6": 0.1666666666666666, "1/3": 0.3333333333333333, "2/3": 0.6666666666666666, "5/6": 0.8333333333333333, "1/2": 0.5, "1/4": 0.25, "3/4": 0.75, "0.5": 0.5, "0.25": 0.25, "0.75": 0.75}
    for item_symop_string in list_symops:
        symop_split = item_symop_string.split(",")
        new_coordinates = []
        for item_co in symop_split:
            plus_terms, minus_terms = [], []
            f_pluslist = item_co.split("+")
            for item_f_plus in f_pluslist:
                f_minuslist = item_f_plus.split("-")
                if f_minuslist[0]:
                    plus_terms.append(f_minuslist[0])
                minus_terms += f_minuslist[1:]
            result = 0
            for count_termlist, item_termlist in enumerate((plus_terms, minus_terms)):
                for item_string in item_termlist:
                    if item_string in DICT_NUMS:
                        result += DICT_NUMS.get(item_string) * (-1) ** count_termlist
                    else:
                        message = f"symop_parser didn't recognize all construction terms! Leftover string: '{item_string}'"
                        warnings.add(message)
            new_coordinates.append(result)
        yield new_coordinates

def build_coordinate(input_string):
    DICT_RATIONAL_NUMS = {"0.167": "1/6", "0.1667": "1/6", "0.16667": "1/6", "0.166667": "1/6", "0.67": "2/3", "0.667": "2/3", "0.6667": "2/3", "0.66667": "2/3", "0.666667": "2/3", "0.33": "1/3", "0.333": "1/3", "0.3333": "1/3", "0.33333": "1/3", "0.333333": "1/3", "0.833": "5/6", "0.8333": "5/6", "0.83333": "5/6", "0.833333": "5/6", "-0.167": "5/6", "-0.1667": "5/6", "-0.16667": "5/6", "-0.166667": "5/6", "-0.67": "1/3", "-0.667": "1/3", "-0.6667": "1/3", "-0.66667": "1/3", "-0.666667": "1/3", "-0.33": "2/3", "-0.333": "2/3", "-0.3333": "2/3", "-0.33333": "2/3", "-0.333333": "2/3", "-0.833": "1/6", "-0.8333": "1/6", "-0.83333": "1/6", "-0.833333": "1/6"}
    coord_builder = ""
    esd_given = False
    for item_char in input_string:
        if item_char in CHARS_FLOAT:
            coord_builder += item_char
        elif item_char == "(":
            esd_given = True
            break
    if coord_builder[-1] == ".":
        coord_builder.strip(".")
    if coord_builder[0] == ".":
        coord_builder = "0" + coord_builder
    elif coord_builder[0] == "-" and coord_builder[1] == ".":
        coord_builder = "-0" + coord_builder[1:]
    if not esd_given:
        if coord_builder in DICT_RATIONAL_NUMS:
            coord_builder = DICT_RATIONAL_NUMS.get(coord_builder)
    return coord_builder

def evaluate_redundancy(poly_name, finished_polyhedra): # checks if a polyhedron of higher symmetry is a perfect match, in which case the queried polyhedron is not evaluated.
    for item_poly in finished_polyhedra:
        if item_poly.csm < 0.0001:
            terminal_node = item_poly.env.cif.main.dict_upright_nodes.get(item_poly.name)
            for item_node in terminal_node.get_branches():
                try:
                    if poly_name == item_node.name:
                        return True
                except AttributeError:
                    pass
    return False

def prepare_belt_dicts(belt_dicts, direction_factor = -1):
    v0 = 0
    modified_belt_dicts = deepcopy(belt_dicts)
    for item_dict in modified_belt_dicts:
        v, phi_init = item_dict.get("v"), item_dict.get("phi_init")
        for item_factor in ("xy_factor", "z_factor"):
            if not item_factor in item_dict:
                item_dict[item_factor] = DICT_BELT_DEFAULTS.get(item_factor)
        for item_mod_type in DICT_SUFFIXES:
            for count_plus in range(100):
                mod_type = item_mod_type + count_plus * "+"
                suffixes = DICT_SUFFIXES.get(item_mod_type)
                if not mod_type in item_dict and not mod_type + suffixes[0] in item_dict:
                    break
                for item_extension in suffixes:
                    if not mod_type + item_extension in item_dict:
                        item_dict[mod_type + item_extension] = DICT_BELT_DEFAULTS.get(item_mod_type + item_extension)
        if item_dict.get("v") > 1 and not v0:
            v0 = item_dict.get("v")
        for item_param in ("xy", "z"):
            if item_dict.get(item_param) and item_dict.get(f"{item_param}_factor") == 0.0:
                item_dict.pop(item_param)
        if phi_init == "s":
            if v % 2 and not v0 % 2:
                item_dict["phi_init"] = pi / v0
            elif v0 % 2 and not v % 2:
                item_dict["phi_init"] = pi / v
            else:
                item_dict["phi_init"] = pi / max(v0, v)
        elif isinstance(phi_init, float) or isinstance(phi_init, int):
            item_dict["phi_init"] = phi_init
        else:
            item_dict["phi_init"] = 0.0
    return modified_belt_dicts

def prototype_polyhedron(belt_dicts):
    prototype = []
    for item_dict in belt_dicts:
        v, erase = item_dict.get("v"), item_dict.get("erase", ())
        xy_factor, z_factor = item_dict.get("xy_factor"), item_dict.get("z_factor")
        phi_init = item_dict.get("phi_init")
        w = [i for i in range(v) if not i in erase]
        generated_vecs = []
        for item_w in w:
            new_vec = Vec(xy_factor, 0, z_factor)
            for count_plus in range(100):
                if item_dict.get(f"z_mod{count_plus * '+'}_init"):
                    insert = count_plus * '+'
                    new_vec.z += item_dict.get(f"z_mod{insert}_init") * cos((item_w - item_dict.get(f"z_mod{insert}_off")) * item_dict.get(f"z_mod{insert}_frq") * 2 * pi / v)
                else:
                    break
            for count_plus in range(100):
                if item_dict.get(f"xy_mod{count_plus * '+'}_init"):
                    insert = count_plus * '+'
                    new_vec.x += item_dict.get(f"xy_mod{insert}_init") * cos((item_w - item_dict.get(f"xy_mod{insert}_off")) * item_dict.get(f"xy_mod{insert}_frq") * 2 * pi / v)
                else:
                    break
            for count_plus in range(100):
                if item_dict.get(f"phi_mod{count_plus * '+'}_init"):
                    insert = count_plus * '+'
                    new_vec = new_vec.rotate_z(-item_dict.get(f"phi_mod{insert}_init") * sin((item_w - item_dict.get(f"phi_mod{insert}_off")) * item_dict.get(f"phi_mod{insert}_frq") * 2 * pi / v))
                else:
                    break
            new_vec_rotated = new_vec.rotate_z(-(phi_init + (item_w) * (2 * pi / v)))
            generated_vecs.append(new_vec_rotated)
        prototype.append(generated_vecs)
    return prototype

def assign_belts(env, blueprint):
    dict_info, dict_input = blueprint[1], env.dict_input
    has_instructions = False
    for item_aristo in dict_info.get("aristohedra", {}):
        if dict_info.get("aristohedra").get(item_aristo):
            has_instructions = True
            aristo_node = env.dict_symmetree_nodes.get(item_aristo.removesuffix("*"))
            if not aristo_node: # this separation of if statements is intentional; all relevant aristohedra must be available to recycle somewhat reliably.
                has_instructions = False
                break
            if aristo_node.content.csm > dict_input.get("piggyback_cap"):
                has_instructions = False
                break
    if has_instructions and not dict_input.get("custom_axis"):
        return recycle_belt_assignment(env, blueprint)
    else:
        return assign_to_belts_from_scratch(env, blueprint)

def assign_to_belts_from_scratch(env, blueprint):
    HALF_SPHERE = (Vec(0.0, 0.0, 1.0), Vec(0.607, 0, 0.795), Vec(0.188, 0.577, 0.795), Vec(-0.491, 0.357, 0.795), Vec(-0.491, -0.357, 0.795), Vec(0.188, -0.577, 0.795), Vec(0.894, 0, 0.447), Vec(0.276, 0.851, 0.447), Vec(-0.724, 0.526, 0.447), Vec(-0.724, -0.526, 0.447),Vec(0.276, -0.851, 0.447), Vec(0.982, 0, 0.188), Vec(0.304, 0.934, 0.188), Vec(-0.795, 0.577, 0.188), Vec(-0.795, -0.577, 0.188), Vec(0.304, -0.934, 0.188), Vec(0.343, 0, 0.939), Vec(0.106, 0.3265, 0.939), Vec(-0.278, 0.202, 0.939), Vec(-0.278, -0.202, 0.939), Vec(0.106, -0.327, 0.939), Vec(0.687, 0, 0.727), Vec(0.212, 0.653, 0.727), Vec(-0.555, 0.404, 0.727), Vec(-0.555, -0.404, 0.727), Vec(0.212, -0.653, 0.727), Vec(0.793, 0.327, 0.515), Vec(0.555, 0.653, 0.515), Vec(-0.066, 0.855, 0.515), Vec(-0.449, 0.730, 0.515), Vec(-0.833, 0.202, 0.515), Vec(-0.833, -0.202, 0.515), Vec(-0.449, -0.730, 0.515), Vec(-0.066, -0.855, 0.515), Vec(0.555, -0.653, 0.515), Vec(0.793, -0.327, 0.515), Vec(0.964, 0.202, 0.172), Vec(0.490, 0.855, 0.172), Vec(0.106, 0.979, 0.172), Vec(-0.662, 0.730, 0.172), Vec(-0.899, 0.404, 0.172), Vec(-0.899, -0.404, 0.172), Vec(-0.662, -0.730, 0.172), Vec(0.106, -0.979, 0.172), Vec(0.490, -0.855, 0.172), Vec(0.964, -0.202, 0.172)) # the vertices on one hemisphere of a fully capped truncated icosahedron. high resolution, maybe unnecessarily high
    belt_lengths = []
    dict_info, belt_dicts = blueprint[1], blueprint[2]
    dict_input, ligand_sphere = env.dict_input, env.ligand_sphere
    for count_belt, item_dict in enumerate(belt_dicts):
        belt_lengths.append(item_dict.get("v") - len(item_dict.get("erase", ())))
    required_perspectives = ("obverse", "reverse") if len(dict_info.get("axis", "1")) == 1 or "+" in dict_info.get("axis", "1") else ("obverse",) # If the polyhedron symmetries in axis direction don't contain a rotoinversion or horizontal mirror plane, the probing vectors need to cover the full sphere to make sure the correct orientation is found.
    if dict_input.get("custom_axis"):
        tuple_probe_vecs = (dict_input.get("custom_axis").coordinate_transformation(env.cif.MATRIX_TO_CART).unit(),)
    else:
        tuple_probe_vecs = HALF_SPHERE
    all_maps, belt_assignments = [], []
    if ("#molecule" in dict_info.get("tags") and dict_input.get("center_method") != "force_centroid") or dict_input.get("center_method") == "force_central_atom":
        ligands_env = env.ligands_around_central_atom
    else:
        ligands_env = env.ligands_around_centroid
    for item_main_axis_candidate in tuple_probe_vecs:
        for item_perspective in required_perspectives:
            if item_perspective == "obverse":
                list_dotprod_tuples = []
                for count_dot, item_vec in enumerate(ligand_sphere):
                    dotprod = item_main_axis_candidate.dot(item_vec)
                    list_dotprod_tuples.append((dotprod, count_dot))
                list_dotprod_tuples.sort(key = lambda i: i[0], reverse = True)
            vecs_real_tuples = []
            belts_real, map_belts = [], []
            count_belt = 0
            for item_dot_tup in (list_dotprod_tuples if item_perspective == "obverse" else reversed(list_dotprod_tuples)):
                vecs_real_tuples.append(item_dot_tup)
                if len(vecs_real_tuples) == belt_lengths[count_belt]:
                    vecs_real_tuples.sort(key = lambda i: i[1]) # sort to identify redundant assignments
                    map_belts.append([i[1] for i in vecs_real_tuples])
                    belts_real.append([ligands_env[i[1]] for i in vecs_real_tuples])
                    vecs_real_tuples = []
                    count_belt += 1
            if not map_belts in all_maps and not map_belts[::-1] in all_maps:
                all_maps.append(map_belts)
                belt_assignments.append((deepcopy(belts_real), deepcopy(map_belts)))
    best_tuples = [(i, inf) for i in range(4)]
    unique_qualities = []
    for count_tuple, item_tuple in enumerate(belt_assignments):
        vecs_retare, belt_centroids = [], []
        for item_belt in item_tuple[0]:
            belt_centroid = get_centroid(item_belt)
            vecs_retare += [i - belt_centroid for i in item_belt]
            belt_centroids.append(belt_centroid)
        plane_normal = covariance_eigen(vecs_retare)[0][1]
        tuple_quality = 0
        for item_vec in vecs_retare:
            tuple_quality += item_vec.dot(plane_normal.unit()) ** 2
        for count_centroid, item_centroid in enumerate(belt_centroids):
            tuple_quality += len(item_tuple[0][count_centroid]) * item_centroid.cross(plane_normal).squared_length()
        for item_quality in unique_qualities:
            if item_quality - 0.0000001 <= tuple_quality <= item_quality + 0.0000001:
                break
        else: # else statement is executed unless for loop was broken.
            if tuple_quality < best_tuples[-1][1]:
                unique_qualities.append(tuple_quality)
                best_tuples.append((count_tuple, tuple_quality))
                best_tuples.sort(key = lambda i: i[1])
                best_tuples.pop()
    return [belt_assignments[i[0]] for i in best_tuples if i[1] != inf]

def recycle_belt_assignment(env, hetto_blueprint):
    dict_info, dict_input = hetto_blueprint[1], env.dict_input
    hetto_belt_lengths = [i.get("v") - len(i.get("erase", ())) for i in hetto_blueprint[2]]
    if ("#molecule" in dict_info.get("tags") and dict_input.get("center_method") != "force_centroid") or dict_input.get("center_method") == "force_central_atom":
        ligands_env = env.ligands_around_central_atom
    else:
        ligands_env = env.ligands_around_centroid
    belt_assignments = []
    for item_aristo_name in dict_info.get("aristohedra"):
        instructions = dict_info.get("aristohedra").get(item_aristo_name) 
        if not instructions:
            continue
        for item_poly in env.polyhedra:
            if item_poly.name == item_aristo_name.removesuffix("*"):
                aristohedron = item_poly
                break
        aristo_flat_mapped = []
        for count_belt, item_belt in enumerate(aristohedron.prototype):
            for count_vec, item_vec in enumerate(item_belt):
                aristo_flat_mapped.append(((aristohedron.map_belts[count_belt][count_vec]), item_vec)) # first tuple member maps the respective prototype vector onto the original, unbelted ligands_env vector
        for item_instruction in instructions:
            if isinstance(item_instruction, list):
                map_belts = [[aristo_flat_mapped[i][0] for i in j] for j in item_instruction]
                belt_assignments.append(([[ligands_env[i] for i in j] for j in map_belts], map_belts))
                continue
            if item_instruction == "same":
                belt_assignments.append(([[ligands_env[i] for i in j] for j in aristohedron.map_belts], aristohedron.map_belts))
                continue
            try:
                axis = get_centroid([aristo_flat_mapped[i][1] for i in item_instruction]).unit()
            except ZeroDivisionError:
                axis = Vec(0, 0, 1)
            map_belts, indices_hetto_belt = [], []
            hetto_belt_count = 0
            aristo_flat_mapped_enumerated = [(c, i) for c, i in enumerate(aristo_flat_mapped)]
            aristo_flat_mapped_enumerated.sort(key = lambda t: t[1][1].dot(axis), reverse = True)
            for item_vertex_tuple in aristo_flat_mapped_enumerated:
                indices_hetto_belt.append(item_vertex_tuple[1][0])
                if len(indices_hetto_belt) == hetto_belt_lengths[hetto_belt_count]:
                    map_belts.append(indices_hetto_belt)
                    indices_hetto_belt = []
                    hetto_belt_count += 1
            belt_assignments.append(([[ligands_env[i] for i in j] for j in map_belts], map_belts))
    return belt_assignments

def assemble_initial_parameters(belt_dicts, dict_bundles, dict_var_range, dict_bound_parameters):
    DICT_INITIAL_VALUES = {"len": 1.0, "z": 1.0, "xy": 1.0, "phi": 0.0}
    dict_parameters = {}
    set_types = set()
    for item_type in DICT_INITIAL_VALUES:
        dict_parameters[item_type] = {}
        for item_belt_dict in belt_dicts:
            if item_type in item_belt_dict:
                set_types.add(item_type)
                if item_belt_dict.get(item_type) not in dict_bound_parameters.get(item_type, {}):
                    dict_parameters.get(item_type)[item_belt_dict.get(item_type)] = DICT_INITIAL_VALUES.get(item_type)
    for item_type in ("phi_mod", "xy_mod", "z_mod"):
        for item_belt_dict in belt_dicts:
            for count_plus in range(100):
                plus_term = item_type + count_plus * "+"
                if plus_term in item_belt_dict:
                    set_types.add(plus_term)
                    if item_belt_dict.get(plus_term) in dict_bound_parameters.get(item_type, {}):
                        continue
                    if plus_term not in dict_parameters:
                        dict_parameters[plus_term] = {}
                    dict_parameters.get(plus_term)[item_belt_dict.get(plus_term)] = 0.0
                else:
                    break
    if dict_bundles:
        dict_parameters["parameter_bundles"] = {}
        for item_var in dict_bundles:
            dict_parameters.get("parameter_bundles")[item_var] = dict_var_range.get(item_var)[2]
    return dict_parameters

def register_bound_parameters(dict_parameters, dict_bundles, dict_bound):
    dict_vars = {}
    for item_type in dict_parameters:
        for item_param in dict_parameters.get(item_type):
            dict_vars[item_param] = dict_parameters.get(item_type).get(item_param)
    dict_full = deepcopy(dict_parameters)
    for item_var in dict_bundles:
        for item_type in dict_bound:
            for item_param in dict_bound.get(item_type):
                if item_param in dict_bundles.get(item_var):
                    dict_full.get("parameter_bundles")[item_param] = parse_math(dict_bundles.get(item_var).get(item_param), dict_variables = dict_vars)
    return dict_full

def find_coordination_spheres(atom_list_sorted, tuple_requested_spheres, max_size = 60):
    if len(atom_list_sorted) < 2:
        return [[atom_list_sorted[0]]]
    min_bond_length = atom_list_sorted[1][-1]
    predecessor = atom_list_sorted[0]
    coordination_spheres, sphere = [], []
    for item_ligand in atom_list_sorted[1:]:
        first_distance_sphere = sphere[0][-1] if sphere else item_ligand[-1]
        intervall = item_ligand[-1] - predecessor[-1]
        predecessor = item_ligand
        if (intervall > 0.2 * min_bond_length or len(sphere) == max_size or item_ligand[-1] / first_distance_sphere > 1.45) and sphere:
            coordination_spheres.append(sphere)
            sphere = []
        sphere.append(item_ligand)
    coordination_spheres.append(sphere)
    return [[atom_list_sorted[0]] + coordination_spheres[i - 1] for i in tuple_requested_spheres if 0 < i <= len(coordination_spheres)]

def evaluate_degeneracy(belt_dicts):
    # Some polyhedra with chiral point groups (e. g. twisted prisms) don't need chiral treatment, while some non-chiral polyhedra, such as the pyritohedral icosahedron, do.
    for item_belt in belt_dicts:
        if item_belt.get("phi_init"):
            if not 3.141 < item_belt.get("phi_init") * max(item_belt.get("v"), belt_dicts[0].get("v") if belt_dicts[0].get("v") != 1 else belt_dicts[1].get("v")) < 3.142:
                return 2
    if len(set([i.get("v") for i in belt_dicts if i.get("xy_factor")])) > 1:
        return 1
    degeneracy = 1
    for item_belt in belt_dicts:
        for item_type in ("phi_mod", "xy_mod", "z_mod"):
            if item_type in item_belt:
                try:
                    degeneracy_probe = int(item_belt.get("v") / item_belt.get(item_type + "_frq"))
                    if 2 < degeneracy_probe > degeneracy:
                        degeneracy = degeneracy_probe
                except ZeroDivisionError:
                    pass
    return degeneracy

def run_blueprint(blueprint, env, dict_input):
    dict_info = blueprint[1]
    print("start_new_polyhedron - central_atom:", env.env_fract[0][0], "| name:", blueprint[0], "| info:", dict_info)
    belt_dicts = prepare_belt_dicts(blueprint[2])
    belt_assignments = assign_belts(env, blueprint)
    degeneracy = evaluate_degeneracy(belt_dicts)
    best_poly = Polyhedron("placeholder", [">", {}, [], {}])
    for count_enantiomorph in range(degeneracy):
        belt_dicts_copy = deepcopy(belt_dicts)
        if "#chiral" in dict_info.get("tags"):
            addendum = "_clockwise" if count_enantiomorph == 0 else "_anticlockwise"
            for item_belt_dict in belt_dicts_copy:
                item_belt_dict["phi_init"] *= (-1) ** count_enantiomorph
                for item_param in ("phi", "xy", "z"):
                    if item_param + "_mod_off" in item_belt_dict:
                        item_belt_dict[item_param + "_mod_off"] *= (-1) ** count_enantiomorph
        else:
            addendum = ""
            for item_belt_dict in belt_dicts_copy:
                for item_param in ("phi", "xy", "z"):
                    if item_param + "_mod_off" in item_belt_dict:
                        item_belt_dict[item_param + "_mod_off"] += count_enantiomorph
        prototype = prototype_polyhedron(belt_dicts_copy)
        print("prototype:", [[(round(i.x, 4), round(i.y, 4), round(i.z, 4)) for i in j] for j in prototype], "| enantiomorph:", count_enantiomorph + 1, "/", degeneracy)
        for item_dist in belt_assignments:
            poly_object = Polyhedron(env, (blueprint[0] + addendum, blueprint[1], belt_dicts_copy), belts_model = deepcopy(prototype), belts_real = item_dist[0], map_belts = item_dist[1])
            poly_object.evaluate()
            print("poly_object csm:", round(poly_object.csm, 6), "| best_poly csm:", round(best_poly.csm, 6))
            print(f"{poly_object.name} post-refinement:")
            print(" real:", [[str(i) for i in j] for j in poly_object.belts_real])
            print("model:", [[str(i) for i in j] for j in poly_object.belts_model])
            if poly_object.csm_taxed < best_poly.csm_taxed:
                best_poly = poly_object.copy()
                if poly_object.csm < 0.00001:
                    best_poly.add_to_symmetree()
                    return best_poly
    best_poly.add_to_symmetree()
    return best_poly

def get_file(directory, filename, ending, overwrite = True):
    if overwrite:
        return open(f"{directory}{filename}{ending}", "w")
    list_files = os.listdir(directory)
    if not f"{filename}{ending}" in list_files:
        return open(f"{directory}{filename}{ending}", "w")
    for count in range(1, 1000):
        if not f"{filename}_{count}{ending}" in list_files:
            return open(f"{directory}{filename}_{count}{ending}", "w")

def rad(angle_deg):
    return pi * angle_deg / 180

def deg(angle_rad):
    return 180 * angle_rad / pi

def translate_to_unit_cell(coordinate):
    while coordinate < 0:
        coordinate += 1
    while coordinate > 1:
        coordinate -= 1
    return coordinate

def get_centroid(vectors):
    try:
        return Vec(sum([i.x for i in vectors]), sum([i.y for i in vectors]), sum([i.z for i in vectors])).scale(1 / len(vectors))
    except ZeroDivisionError:
        return Vec(0, 0, 0)

def covariance_eigen(list_of_cartesian_vecs):
    cen = get_centroid(list_of_cartesian_vecs)
    xdiff_squared, ydiff_squared, zdiff_squared, xydiff, zxdiff, yzdiff = 0, 0, 0, 0, 0, 0
    for item_vec in list_of_cartesian_vecs:
        xdiff_squared += (item_vec.x - cen.x)**2
        ydiff_squared += (item_vec.y - cen.y)**2
        zdiff_squared += (item_vec.z - cen.z)**2
        xydiff += (item_vec.x - cen.x) * (item_vec.y - cen.y)
        zxdiff += (item_vec.z - cen.z) * (item_vec.x - cen.x)
        yzdiff += (item_vec.y - cen.y) * (item_vec.z - cen.z)
    var_matrix = array([[xdiff_squared, xydiff, zxdiff], [xydiff, ydiff_squared, yzdiff], [zxdiff, yzdiff, zdiff_squared]])
    eigvals, np_eigvecs = linalg.eig(var_matrix)
    eigvecs = []
    for count_eigvec in range(len(np_eigvecs)):
        eigvecs.append(Vec(np_eigvecs[0][count_eigvec].real, np_eigvecs[1][count_eigvec].real, np_eigvecs[2][count_eigvec].real))
    eiglist = list(zip(eigvals, eigvecs))
    eiglist.sort(key = lambda i: i[0])
    return eiglist

def kabsch_algorithm(npvecs_real, npvecs_model):
    covariance_matrix = array(npvecs_model).T @ array(npvecs_real)
    svd_u, svd_sigma, svd_v_transposed = linalg.svd(covariance_matrix) #svd = singular value decomposition
    svd_v, svd_u_transposed = svd_v_transposed.T, svd_u.T
    sign_found = round(linalg.det(svd_v @ svd_u_transposed))
    diagonal_sign_matrix = array([[1, 0, 0], [0, 1, 0], [0, 0, sign_found]])
    rotation_matrix = svd_v @ diagonal_sign_matrix @ svd_u_transposed
    return rotation_matrix

def poly_search(list_candidates, dict_criteria, var_max):
    list_matches = []
    for item_candidate in list_candidates:
        if item_candidate[1].get("parameter_count") > var_max:
            continue
        match = False
        checklist = [item_candidate[0], item_candidate[1].get("point_group", "?"), DICT_POINT_GROUP_CONVERSION.get(item_candidate[1].get("point_group", "?"))] + list(item_candidate[1].get("tags", ())) + list(item_candidate[1].get("synonyms", ()))
        for item_qualifier_or in dict_criteria.get("or"):
            for item_pattern in checklist:
                if item_qualifier_or in item_pattern:
                    match = True
        for item_qualifier_and in dict_criteria.get("and"):
            for item_pattern in checklist:
                if item_qualifier_and in item_pattern:
                    match = True
                    break
            else: # for else applies if for loop was not broken.
                match = False
                break
        if not dict_criteria.get("and") and not dict_criteria.get("or"):
            match = True
        for item_disqualifier in dict_criteria.get("not"):
            for item_pattern in checklist:
                if item_disqualifier in item_pattern:
                    match = False
                    break
        if match:
            list_matches.append(item_candidate)
    return list_matches

def atom_search(list_candidates, dict_criteria_0, has_dummy):
    DICT_ELEMENT_WILDCARDS = {"*": ("H", "D", "T", "Li", "Na", "K", "Rb", "Cs", "Fr", "Be", "Mg", "Ca", "Sr", "Ba", "Ra", "B", "Al", "Ga", "In", "Tl", "C", "Si", "Ge", "Sn", "Pb", "N", "P", "As", "Sb", "Bi", "O", "S", "Se", "Te", "Po", "F", "Cl", "Br", "I", "At", "He", "Ne", "Ar", "Kr", "Xe", "Rn", "Sc", "Y", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Ti", "Zr", "Hf", "V", "Nb", "Ta", "Cr", "Mo", "W", "Mn", "Tc", "Re", "Fe", "Ru", "Os", "Co", "Rh", "Ir", "Ni", "Pd", "Pt", "Cu", "Ag", "Au", "Zn", "Cd", "Hg"), "*gr1": ("H", "D", "T", "Li", "Na", "K", "Rb", "Cs", "Fr", ), "*gr2": ("Be", "Mg", "Ca", "Sr", "Ba", "Ra"), "*gr3": ("Sc", "Y", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr"), "*gr4": ("Ti", "Zr", "Hf", "Rf"), "*gr5": ("V", "Nb", "Ta", "Db"), "*gr6": ("Cr", "Mo", "W", "Sg"), "*gr7": ("Mn", "Tc", "Re", "Bh"), "*gr8": ("Fe", "Ru", "Os", "Hs"), "*gr9": ("Co", "Rh", "Ir", "Mt"), "*gr10": ("Ni", "Pd", "Pt", "Ds"), "*gr11": ("Cu", "Ag", "Au", "Rg"), "*gr12": ("Zn", "Cd", "Hg", "Cn"), "*gr13": ("B", "Al", "Ga", "In", "Tl"), "*gr14": ("C", "Si", "Ge", "Sn", "Pb"), "*gr15": ("N", "P", "As", "Sb", "Bi"), "*gr16": ("O", "S", "Se", "Te", "Po"), "*gr17": ("F", "Cl", "Br", "I", "At"), "*gr18": ("He", "Ne", "Ar", "Kr", "Xe", "Rn"), "*m": ("Li", "Na", "K", "Rb", "Cs", "Fr", "Be", "Mg", "Ca", "Sr", "Ba", "Ra", "Al", "Ga", "In", "Tl", "Sn", "Pb", "Bi", "Po", "At", "Sc", "Y", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Ti", "Zr", "Hf", "V", "Nb", "Ta", "Cr", "Mo", "W", "Mn", "Tc", "Re", "Fe", "Ru", "Os", "Co", "Rh", "Ir", "Ni", "Pd", "Pt", "Cu", "Ag", "Au", "Zn", "Cd", "Hg"), "*e": ("H", "D", "T", "B", "Ga", "In", "Tl", "C", "Si", "Ge", "Sn", "Pb", "N", "P", "As", "Sb", "Bi", "O", "S", "Se", "Te", "Po", "F", "Cl", "Br", "I", "At", "He", "Ne", "Ar", "Kr", "Xe", "Rn"), "*x": ("N", "P", "O", "S", "Se", "F", "Cl", "Br", "I"), "*ln": ("La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu"), "*re": ("Sc", "Y", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu"), "*an": ("Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr"), "*tm": ("Ti", "Zr", "Hf", "V", "Nb", "Ta", "Cr", "Mo", "W", "Mn", "Tc", "Re", "Fe", "Ru", "Os", "Co", "Rh", "Ir", "Ni", "Pd", "Pt", "Cu", "Ag", "Au", "Zn", "Cd", "Hg", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn")}
    dict_matches = {}
    dict_criteria = deepcopy(dict_criteria_0)
    for item_logic in dict_criteria:
        for item_criterion in dict_criteria.get(item_logic).copy():
            if item_criterion.lower() in DICT_ELEMENT_WILDCARDS:
                dict_criteria.get(item_logic)[DICT_ELEMENT_WILDCARDS.get(item_criterion)] = dict_criteria.get(item_logic).pop(item_criterion)
    for item_candidate in list_candidates:
        match = False
        criterion_tuple = (item_candidate[0].lower(), item_candidate[1].lower())
        for item_qualifier_or in dict_criteria.get("or"):
            if isinstance(item_qualifier_or, tuple):
                for item_key in item_qualifier_or:
                    if item_key.lower() in criterion_tuple:
                        match = True
                        value = dict_criteria.get("or").get(item_qualifier_or)
            elif item_qualifier_or.lower() in criterion_tuple:
                match = True
                value = dict_criteria.get("or").get(item_qualifier_or)
        for item_qualifier_and in dict_criteria.get("and"):
            if isinstance(item_qualifier_and, tuple):
                match = False
                for item_key in item_qualifier_and:
                    if item_key.lower() in criterion_tuple:
                        match = True
                        value = dict_criteria.get("and").get(item_qualifier_and)
                if match == False:
                    break
            elif item_qualifier_and.lower() in criterion_tuple:
                match = True
                value = dict_criteria.get("and").get(item_qualifier_and)
            else:
                match = False
                break
        if not dict_criteria.get("and") and not dict_criteria.get("or") and not has_dummy:
            match, value = True, None
        for item_disqualifier in dict_criteria.get("not"):
            if isinstance(item_disqualifier, tuple):
                for item_key in item_disqualifier:
                    if item_key.lower() in criterion_tuple:
                        match = False
            elif item_disqualifier.lower() in criterion_tuple:
                match = False
        if match:
            dict_matches[item_candidate] = value
    return dict_matches

def construct_symmetrees(dict_blueprints_by_cn):
    dict_symmetrees, dict_nodes = {}, {}
    for item_cn in dict_blueprints_by_cn:
        dict_symmetrees[item_cn] = TreeNode(None, "root", [])
        for item_poly in dict_blueprints_by_cn.get(item_cn):
            poly_node = TreeNode(item_poly, item_poly[0], [])
            dict_nodes[item_poly[0]] = poly_node
            if not item_poly[1].get("aristohedra"):
                dict_symmetrees.get(item_cn).children.append(poly_node)
                continue
            for item_aristo_name in item_poly[1].get("aristohedra"):
                aristo_node = dict_nodes.get(item_aristo_name)
                if aristo_node:
                    aristo_node.children.append(poly_node)
    return dict_symmetrees, dict_nodes

def flip_symmetrees(dict_upright_trees):
    dict_upside_down_trees, dict_nodes = {}, {}
    for item_cn in dict_upright_trees:
        node_tuples = dict_upright_trees.get(item_cn).get_branch_depth_tuples(depth = 0)
        node_tuples.sort(key = lambda i: i[1], reverse = True)
        previous_depth = node_tuples[0][1]
        depth_layers, depth_list = [], []
        for item_node_tuple in node_tuples:
            if item_node_tuple[1] < previous_depth:
                previous_depth = item_node_tuple[1]
                depth_layers.append(depth_list)
                depth_list = []
            node = item_node_tuple[0].copy()
            dict_nodes[node.name] = node
            depth_list.append(node)
        depth_layers.append(depth_list)
        dict_upside_down_trees[item_cn] = TreeNode(None, "root", depth_layers[0])
        for count_depth, item_depth_list in enumerate(depth_layers):
            if count_depth < len(depth_layers) - 1:
                lower_level = depth_layers[count_depth + 1]
                candidate_indices = set(range(len(lower_level)))
            else:
                dict_upside_down_trees.get(item_cn).children += [lower_level[c] for c in list(candidate_indices) if lower_level[c].name != "root"]
                break
            for item_node in item_depth_list:
                item_node.children = []
                if not item_node.content:
                    continue
                for item_aristo_name in item_node.content[1].get("aristohedra", {}):
                    for count_candidate, item_candidate_node in enumerate(lower_level):
                        if item_candidate_node.name == item_aristo_name:
                            item_node.children.append(item_candidate_node)
                            candidate_indices.discard(count_candidate)
            dict_upside_down_trees.get(item_cn).children += [lower_level[c] for c in list(candidate_indices) if lower_level[c].name != "root"]
    return dict_upside_down_trees, dict_nodes

def stround(number, places):
    return f"%.{places}f" % round(number, places)

def get_transformation_matrix_to_fract(cell_parameters):
    a, b, c = cell_parameters[0], cell_parameters[1], cell_parameters[2]
    cos_alpha, cos_beta = cos(cell_parameters[3]), cos(cell_parameters[4])
    sin_gamma, cos_gamma = sin(cell_parameters[5]), cos(cell_parameters[5])
    sqrt_term = sqrt(1 - cos_alpha ** 2 - cos_beta ** 2 - cos_gamma ** 2 + 2 * cos_alpha * cos_beta * cos_gamma)
    row_1 = (1 / a, -cos_gamma / (a * sin_gamma), (cos_alpha * cos_gamma - cos_beta) / (a * sin_gamma * sqrt_term))
    row_2 = (0, 1 / (b * sin_gamma), (cos_beta * cos_gamma - cos_alpha) / (b * sin_gamma * sqrt_term))
    row_3 = (0, 0, sin_gamma / (c * sqrt_term))
    return (row_1, row_2, row_3)

def get_transformation_matrix_to_cart(cell_parameters):
    a, b, c = cell_parameters[0], cell_parameters[1], cell_parameters[2]
    cos_alpha, cos_beta = cos(cell_parameters[3]), cos(cell_parameters[4])
    sin_gamma, cos_gamma = sin(cell_parameters[5]), cos(cell_parameters[5])
    row_1 = (a, b * cos_gamma, c * cos_beta)
    row_2 = (0, b * sin_gamma, c * (cos_alpha - cos_beta * cos_gamma) / sin_gamma)
    row_3 = (0, 0, c * sqrt(1 - cos_alpha ** 2 - cos_beta ** 2 - cos_gamma ** 2 + 2 * cos_alpha * cos_beta * cos_gamma) / sin_gamma)
    return (row_1, row_2, row_3)

def generate_translated_atoms(d_max, central_vec_fract, ligand_coordinate_tuple_fract, range_cell_extension, transformation_matrix):
    lig_x, lig_y, lig_z = ligand_coordinate_tuple_fract[0], ligand_coordinate_tuple_fract[1], ligand_coordinate_tuple_fract[2]
    central_cart = central_vec_fract.coordinate_transformation(transformation_matrix)
    if not round(transformation_matrix[0][1], 6) and not round(transformation_matrix[0][2], 6) and not round(transformation_matrix[1][2], 6): # cubic / tetragonal / orthorhombic
        for item_x in range_cell_extension:
            if abs((lig_x + item_x) * transformation_matrix[0][0] - central_cart.x) > d_max:
                continue
            for item_y in range_cell_extension:
                if abs((lig_y + item_y) * transformation_matrix[1][1] - central_cart.y) > d_max:
                    continue
                for item_z in range_cell_extension:
                    if abs((lig_z + item_z) * transformation_matrix[2][2] - central_cart.z) > d_max:
                        continue
                    ligand_fract = Vec(lig_x + item_x, lig_y + item_y, lig_z + item_z)
                    yield ligand_fract.coordinate_transformation(transformation_matrix), ligand_fract
    elif not round(transformation_matrix[0][2], 6) and not round(transformation_matrix[1][2], 6): # trigonal / hexagonal
        for item_z in range_cell_extension:
            if abs((lig_z + item_z) * transformation_matrix[2][2] - central_cart.z) > d_max:
                continue
            for item_x in range_cell_extension:
                for item_y in range_cell_extension:
                    ligand_fract = Vec(lig_x + item_x, lig_y + item_y, lig_z + item_z)
                    ligand_cart = ligand_fract.coordinate_transformation(transformation_matrix)
                    if abs(ligand_cart.x - central_cart.x) < d_max and abs(ligand_cart.y - central_cart.y) < d_max:
                        yield ligand_cart, ligand_fract
    elif not round(transformation_matrix[0][1], 6) and not round(transformation_matrix[1][2], 6): # monoclinic b
        for item_y in range_cell_extension:
            if abs((lig_y + item_y) * transformation_matrix[1][1] - central_cart.y) > d_max:
                continue
            for item_x in range_cell_extension:
                for item_z in range_cell_extension:
                    ligand_fract = Vec(lig_x + item_x, lig_y + item_y, lig_z + item_z)
                    ligand_cart = ligand_fract.coordinate_transformation(transformation_matrix)
                    if abs(ligand_cart.x - central_cart.x) < d_max and abs(ligand_cart.z - central_cart.z) < d_max:
                        yield ligand_cart, ligand_fract
    else: # triclinic / rhombohedral
        for item_x in range_cell_extension:
            for item_y in range_cell_extension:
                for item_z in range_cell_extension:
                    ligand_fract = Vec(lig_x + item_x, lig_y + item_y, lig_z + item_z)
                    ligand_cart = ligand_fract.coordinate_transformation(transformation_matrix)
                    if abs(ligand_cart.x - central_cart.x) < d_max and abs(ligand_cart.y - central_cart.y) < d_max and abs(ligand_cart.z - central_cart.z) < d_max:
                        yield ligand_cart, ligand_fract

def parse_math(math_string, dict_variables = {}):
    def validate(math_string):
        if "()" in math_string:
            return validate(math_string.replace("()", ""))
        bra_tally = 0
        carved_string = math_string
        for item_operator in DICT_OPERATORS:
            carved_string = carved_string.replace(item_operator + "(", "(")
        for item_char in carved_string:
            if item_char == "(":
                bra_tally += 1
            elif item_char == ")":
                bra_tally -= 1
            elif item_char not in "0123456789+-*/^.e":
                raise NameError(f"bad_character '{item_char}' in '{math_string}'")
            if bra_tally < 0:
                raise SyntaxError(f"bad bracket nesting in '{math_string}': too many ')'")
        if bra_tally != 0:
            raise SyntaxError(f"bad bracket nesting in '{math_string}': too many '('")
        return math_string
    
    def resolve_brackets(math_string):
        index_bra, index_ket = -1, -1
        bra_tally = 0
        for count_char, item_char in enumerate(math_string):
            if item_char == "(":
                bra_tally += 1
                if index_bra == -1:
                    index_bra = count_char
            elif item_char == ")":
                bra_tally -= 1
                if bra_tally == 0:
                    index_ket = count_char
                    break
        if index_bra == -1:
            return add(math_string)
        prefix_length = 0
        if index_bra >= 4:
            if math_string[index_bra - 1].isalpha():
                string_prefix_test = math_string[index_bra - 4:index_bra]
                if string_prefix_test in DICT_OPERATORS:
                    try:
                        bracket, prefix_length = str(DICT_OPERATORS.get(string_prefix_test)(resolve_brackets(math_string[index_bra + 1:index_ket]))), 4
                    except ValueError: # deals with rounding errors
                        bracket, prefix_length = str(DICT_OPERATORS.get(string_prefix_test)(round(resolve_brackets(math_string[index_bra + 1:index_ket]), 8))), 4
        if index_bra >= 3 and prefix_length == 0:
            if math_string[index_bra - 1].isalpha():
                string_prefix_test = math_string[index_bra - 3:index_bra]
                if string_prefix_test in DICT_OPERATORS:
                    bracket, prefix_length = str(DICT_OPERATORS.get(string_prefix_test)(resolve_brackets(math_string[index_bra + 1:index_ket]))), 3
        if prefix_length == 0:
            bracket = str(resolve_brackets(math_string[index_bra + 1:index_ket]))
        return resolve_brackets(math_string[:index_bra - prefix_length] + bracket + math_string[index_ket + 1:])
    
    def add(math_string):
        try:
            if math_string[0] == "-" and math_string[1].isnumeric():
                math_string = "_" + math_string[1:]
        except IndexError:
            pass
        math_string = math_string.replace("*-", "~").replace("e+", "$^").replace("e", "$^").replace("^-", "§").replace("--", "-_")
        positives, negatives = [], []
        list_p = math_string.split("+")
        for item_p in list_p:
            list_n = item_p.split("-")
            if list_n[0]:
                positives.append(list_n[0])
            negatives += list_n[1:]
        result = 0
        for count_termlist, item_termlist in enumerate((positives, negatives)):
            for item_term in item_termlist:
                result += multiply(item_term.replace("~", "*-")) * (-1) ** count_termlist
        return result

    def multiply(math_string):
        factors, denominators = [], []
        list_d = math_string.split("/")
        for count_d, item_d in enumerate(list_d):
            list_f = item_d.split("*")
            for count_f, item_f in enumerate(list_f):
                if count_d % 2 and count_f == 0:
                    denominators.append(item_f)
                else:
                    factors.append(item_f)
        result = 1
        for count_termlist, item_termlist in enumerate((factors, denominators)):
            for item_term in item_termlist:
                result *= 1 / raise_to_power(item_term) if count_termlist % 2 else raise_to_power(item_term)
        return result

    def raise_to_power(math_string):
        list_exp = math_string.replace("$^", "e").replace("$§", "e-").replace("_", "-").split("^")
        exponents = []
        for item_exp in list_exp:
            exponents.append(float(item_exp))
        if len(exponents) == 1:
            return exponents[0]
        elif len(exponents) == 2:
            return exponents[0] ** exponents[1]
        else:
            result = exponents[-1]
            for item_base in exponents[len(exponents) - 2::-1]:
                result = item_base ** result
            return result
    
    DICT_OPERATORS = {"abs": abs, "sqrt": sqrt, "asin": asin, "acos": acos, "atan": atan, "sin": sin, "cos": cos, "tan": tan}
    math_string = math_string.replace("**", "^").replace(" ", "")
    for item_variable in dict_variables:
        math_string = math_string.replace(item_variable, str(dict_variables.get(item_variable)))
    math_string = validate(math_string)
    return resolve_brackets(math_string)
    
def polynator_main():
    icon_data = b64decode(ICON_STRING)
    icon_filename = "icon.ico"
    icon_file = open(icon_filename, "wb")
    icon_file.write(icon_data)
    icon_file.close()
    root = tk.Tk()
    style = ttk.Style(root)
    style.configure(".", background = "#ffffff")
    style.configure("Sunken.TButton", background = "#aaccff")
    root.configure(background = "#ffffff")
    root.wm_iconbitmap(icon_filename)
    root.title("Polynator")
    root.geometry("1140x700+100+40")
    MainWindow(root)
    root.mainloop()
    os.remove(icon_filename)


if __name__ == "__main__":
    polynator_main()
